// The "freEvolve System" is a platform supporting the reconfiguration of 
// component-based applications'
// Copyright (C) 1999  Ralph Hinken and Dr. Oliver Stiemerling
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Version: 1.0
// Date:    09.10.2001
//
// History:
//

package org.freevolve.server;

import java.util.*;
import java.io.*;
import java.rmi.server.*;
import java.rmi.*;
/*
class ComponentException extends Exception {
  public ComponentException () { super(); }
  public ComponentException (String s) { super(s); }
}
*/
//
//  Represents a type of a component
//
/** Eine Komponente, die aus Subkomponenten bestehen kann (komplexe Komponente). 
    Die Bltter sind FlexiBeans (implementierte Komponenten).<br>
    Jede Komponente hat Ports und Parameter. Jede (abstrakte - ein innerer Knoten
    der Hierarchie) Komponente kann Ports und Parameter der Subkomponenten zu 
    sich durchreichen und somit nach auen zur Verfgung stellen und kann Ports 
    der Subkomponenten verbinden.     <br>
    <br>
    Parameter knnen mit Werten belegt werden. <br>
    <br>
    Subkomponenten sind unter <b>Instanznamen</b> (local_name in CAT Spezifikation) 
    und ihrem Typnamen bei der 
    abstrakten Komponente angemeldet. Diese Instanznamen sind eindeutig innerhalb
    einer abstrakten Komponente; ebenso sind Port- und Parameternamen eindeutig 
    innerhalb einer komplexen Komponente (Die Ports und Parameter der Wurzel der
    komplexen Komponente, die wiederum in eine andere Subkomponente eingebaut werdem
    kann).

    @see Port
    @see PortBinding
    @see Param
    @see ParamBinding
*/
public class Component  extends UnicastRemoteObject implements IR_Component {

  /** Der Typ der (komplexen) Komponente, entspricht dem CAT-File Namen.
  */
  private String    typeName;            // unique for all components in appl.
  private String    name;                // same as typeName since only one instance per type is allowed ...
  private boolean   isAtomicComponent ; 
  /** name or predecessor.name + "." + name 
  */
  private String    fullName;           
  private Object    instanciatedProxy;   // Proxy -> only one proxy allowed ...

          Vector    ports;               // Vector of Port
          Vector    params;              // Vector of Param

          Hashtable subComponents;       // key is Name (String)
                                         // value  is of type IR_Component !!!!! *********************

          Vector    portBindings;        // Vector of PortBindings
          Vector    paramBindings;       // Vector of ParamBindings         
          Hashtable    proxies;             // Vector of IR_Proxy

  private IR_Component predecessor;
  private Server    server;
   
   static final char SEPERATOR = '.';

  // private static  Hashtable components;  // alle Komponenten  

  // ************************ END OF MEMBER FIELDS *************************
  //
  //     *** Constructors and save
  //


  public Component (Server s)  throws java.rmi.RemoteException {
    super();
    server = s;
    ports = new Vector();
    params = new Vector();
    subComponents = new Hashtable();
    portBindings = new Vector();
    paramBindings = new Vector();
    proxies       = new Hashtable();
    predecessor = null;
    instanciatedProxy = null;
    typeName = "";
    name = "";
    setAtomic(false);
  }
  

  public Component(String type, Server s)  throws java.rmi.RemoteException {
    this(s);
    this.typeName = type;
  }

  //   one constructor is for loading a CAT-file. Extension is cat.
  /** Mit diesem Konstruktor wird Verzeichnis und Dateiname der CAT-Datei bergeben,
      die die komplexe Komponente beschreibt.
      @param name Der Dateiname mit Endung.
  */
  public Component(String directory, String typName, Server s) throws MyFileException, RemoteException, IOException {
    this(s);
    //this.name = typName;              // spter neu zu setzen! Als Root unsinnig!
    this.typeName = typName;
    MyFile mf = new MyFile();
    comp_io cio = new comp_io(server);
    String CAT = mf.getFile(directory, typName+".CAT");
      //      System.out.println(CAT);
    Component result = null;
    result=cio.getComponent(CAT, typName, this);
    if (result==null) typeName="##";
    else {
       this.ports = result.ports;               
       this.params = result.params;             
       this.subComponents = result.subComponents;
       this.portBindings = result.portBindings;        // Vector of PortBindings
       this.paramBindings = result.paramBindings;       // Vector of ParamBindings         
    }
 }

  /** Gibt an, ob dieses Component Objekt fr eine implementierte Komponente steht.<br>
      Muss nach Erzeugen gesetzt werden. Default ist FALSE
  */
  public boolean isAtomic() {return isAtomicComponent;}

  /** Gibt an, ob dieses Component Objekt fr eine implementierte Komponente steht.<br>
      Muss nach Erzeugen gesetzt werden. Default ist FALSE
  */
  public void    setAtomic(boolean isAtomic) {isAtomicComponent = isAtomic;}

   /** Gibt CAT Reprsentation dieser komplexen Komponente zurck.
   */
  public String toText() {
    comp_io cio = new comp_io(server);
    return cio.getCompString(this);
  }

  /** Speichert die Beschreibung 
      der (komplexen oder implementierten) Komponente in ein CAT-File ab.
      @return true, falls erfolgreich.
  */
  public boolean save(String directory, String name) throws MyFileException, IOException {
      try {
         MyFile mf = new MyFile();
         comp_io cio = new comp_io(server);
         String CAT = cio.getCompString(this);
         mf.writeFile(directory, name, CAT);
         server.message("comp", 5, getTypeName() + " gespeichert unter " + directory +
         "/" + name);
         return true;
      } catch (RemoteException e) {
         server.message("error", 1, "Component.save: " + e.toString());
         return false;
      }
  } 
  
  //
  //     *** Access for the datastructure
  //
  /** Typname der komplexen oder implementierten Komponente
  */
  public String      getTypeName() {
    return typeName;
  }

   /** Der Name. Der Instanzname unter dem dieses Objekt bei seinem Vater angemeldet ist.
        @see Component#getTypeName
        @see Component#getSubComponent
   */
  public String      getName() {
    return name;
  }

   public String getFullName () {
      try {
         if (this.isRoot()) return typeName;
         return predecessor.getFullName() + SEPERATOR + this.name;
      } catch (RemoteException e) {
         server.message("error", 1, "Component.getFullName: " + e.toString());
      }
      return null;
   }

   /** Der Name. Der Instanzname unter dem dieses Objekt bei seinem Vater angemeldet ist.<br>
       VORSICHT: Diese Funktion sorgt nicht fr Konsistenz bei der Vaterkomponente.<br>
                 Statt dessen nur changeSubComponentName benutzen!!
        @see Component#changeSubComponentName
        @see Component#getTypeName
        @see Component#getSubComponent
   */
  public void setName(String name) { this.name = name; }

  /** Gibt die Vater/Vorgnger Komponente zurck, bei der wir als Subkomponente angemeldet sind.
      @return null, falls wir Rootkomponente sind.
  */
  public IR_Component   getPredecessor () { return predecessor; }

  /** Setzt die Vater/Vorgnger Komponente zurck, bei der wir als Subkomponente angemeldet sind.
      Bitte nicht aufrufen, wird nur von add/removeSubComponent aufgerufen.
  */
  public void setPredecessor(IR_Component comp) { 
      predecessor = comp;
  }

  /** Hier meldet sich das Proxy-Objekt an!
  */
  public void addIR_Proxy(String user, String ID, IR_Proxy proxy) {
      server.message("comp", 1, "addProxy "+name);
      proxies.put(user+":"+ID, proxy);
  }
 
   
  /** Hier meldet sich das Proxy-Objekt ab!
  */
  public void removeIR_Proxy(String user, String ID) {
      for (Enumeration e = proxies.keys(); e.hasMoreElements(); ) {
          String s = (String) e.nextElement();
      }
      server.message("comp", 1, "removeProxy "+name);
      proxies.remove(user+":"+ID);
      for (Enumeration e = proxies.keys(); e.hasMoreElements(); ) {
          String s = (String) e.nextElement();
      }

  }

  /** Gibt den Server zurck, der diese Komponente erzeugt hat.
      @see Server
  */
  public IR_Server getServer() { return server; }

  /** Die Namen der Ports der komplexen Komponente.
      Das sind alle Ports, die diese abstrakte Komponente anbietet.
      @return Vector      von String
      @see Port
      @see PortBinding
  */
  public Vector getPortNames() {
    Vector portNames = new Vector(this.ports.size());
    for( Enumeration p = this.ports.elements();
	    p.hasMoreElements();
	    portNames.addElement( ((Port) p.nextElement()).getName() )
    );
    return portNames;
  }

  /** Die Port Objekte der komplexen Komponente.
      Das sind alle Ports, die diese abstrakte Komponente anbietet.
      @return Vector      von IR_Port
      @see Port
      @see PortBinding
  */ 
  public Vector      getPorts() {
     return ports;
  }

  /** Das Port Objekte mit Namen name der komplexen Komponente.
      Das ist ein Port, den diese abstrakte Komponente anbietet.
      @return null, falls Port nicht existiert. 
      @see Port
      @see PortBinding
  */ 
  public IR_Port getIR_Port(String name) {
    Port port = null;
    for (Enumeration p = this.ports.elements(); p.hasMoreElements(); ) {
      port = (Port) p.nextElement();
      if (port.getName().equals(name)) {
	      return port;
      }
    }
    return null;
  }
	
  /** Wie getIR_Port, gibt aber Port zurck - spart einige try catch - Zeilen.
      @see #getIR_Port
      @see Port
      @see PortBinding
  */
   Port getPort(String _name) {
    Port port = null;
    for (Enumeration p = this.ports.elements(); p.hasMoreElements(); ) {
      port = (Port) p.nextElement();
      if (port.getName().equals(_name)) {
	      return port;
      }
    }
    return null;
  }

  /** Komfort: Gibt Typ des Ports name zurck.
      @return null, falls Port nicht existiert.
      @see Port
      @see Port#getType
      @see PortBinding
  */ 
  public String  getPortType(String name) {
    Port p = getPort(name);
    if (p != null) return p.getType();
    return null;
  }
  

  /** Komfort: Gibt Polaritt des Ports name zurck.
      @return null, falls Port nicht existiert.
      @see Port#getPolarity
      @see Port
      @see PortBinding
  */ 
  public String getPortPolarity(String name) {
    Port p = getPort(name);
    if (p != null) return p.getPolarity();
    return null;
  }

  // ENDE PORTS *************************************************************

   /** Die Parameter Objekte der komplexen Komponente.
      Das sind alle Parameter, die diese abstrakte Komponente anbietet und alle
      Parameterobjekte die Wertbelegungen von Subkomponentenparametern darstellen.
      @return Vector      von IR_Param
      @see Param
      @see ParamBinding
  */ 
  public Vector      getParams() {
     return params;
  }


  /** Das Parameterobjekt "name" dieser Komponente oder Belegungungsparameter einer Subkomponente . 
      Falls localName=null ist, ist dies ein Parameter, den diese abstrakte 
      Komponente anbietet.
      Ansonsten reprsentiert dieses Parameterobjekt die Wertbelegung eines 
      Parameters mit gleichem Namen der Subkomponente "localName".
      @return null, falls dieser Parameter nicht existiert. 
      @see Param
      @see ParamBinding
  */ 
  public IR_Param getIR_Param(String name, String localName) {
    Param result = null;
    for (Enumeration p = this.params.elements(); p.hasMoreElements();) {
      result = (Param) p.nextElement();
      if (result.getName().equals(name)) {
        if (localName != null && result.isSubCompParam()) {
            if (result.getSubCompName().equals(localName)) return result;
        } else if (localName == null && !result.isSubCompParam()) {
           return result;
        }
 
      }
    }
    return null;
  }
  
  /** wie getIR_Param, nur mit Param frs interne
      @see #getIR_Param
      @see Param
      @see ParamBinding
  */
   Param  getParam(String name, String localName) {
    Param result = null;
    for (Enumeration p = this.params.elements(); p.hasMoreElements();) {
      result = (Param) p.nextElement();
      if (result.getName().equals(name)) {
        if (localName != null && result.isSubCompParam()) {
            if (result.getSubCompName().equals(localName)) return result;
        } else if (localName == null && !result.isSubCompParam()) {
           return result;
        }
      }
    }
    return null;
  }


   /** Gibt die Namen aller Parameter der komplexen Komponente zurck. 
       Gibt keine Namen von Wertbelegungsparametern fr Subkomponentenparametern zurck!
       @return Vector      von String
      @see Param
      @see ParamBinding
   */
  public Vector      getParamNames() {
    Vector paramNames = new Vector(this.params.size());
    Param p = null;
    for( Enumeration e = this.params.elements(); e.hasMoreElements(); ) {
       p = (Param) e.nextElement();
       if (!p.isSubCompParam())
      	 paramNames.addElement(p.getName());
    }
    return paramNames;
  }


  /** Gibt den Typ des Parameters  zurck.
      Falls localName=null ist, ist dies ein Parameter, den diese abstrakte 
      Komponente anbietet.
      Ansonsten reprsentiert dieses Parameterobjekt die Wertbelegung eines 
      Parameters mit gleichem Namen der Subkomponente "localName".
      @return null, falls Parameter name nicht existiert.
      @see Param
      @see ParamBinding
  */
  public String      getParamType (String name, String localName) {
    Param p = getParam(name, localName);
    if (p != null) return p.getType();
    return null;
  }


 /** Testet, ob das IR_Param Objekt Wertbelegungsparameter eines Parameters
     einer Subkomponente ist.
 */
 public boolean     isSubCompParam (IR_Param p) {
      try {
         return p.isSubCompParam();
      } catch (RemoteException e) {
          server.message("error", 1, "Component.isSubCompParam: " + e.toString());
          return false;
     }
 }

  /** Der mgliche String-Wert (oder true, false, null) des Parameters, je nach Typ.
      Wenn dieser Parameter durch eine Wertbelegung innerhalb der Superkomponente
      belegt wurde, wird der Wert dieser Belegung zurckgegeben.<br>
      Falls localName=null ist, ist dies ein Parameter, den diese abstrakte 
      Komponente anbietet.
      Ansonsten reprsentiert dieses Parameterobjekt die Wertbelegung eines 
      Parameters mit gleichem Namen der Subkomponente "localName".
      @return null, falls Parameter nicht belegt oder Typ kein String-Typ.
      @see Param
      @see ParamBinding
  */ 
  public String      getParamSValue(String name, String localName) {
    Param p = getParam(name, localName);
    if (p == null) return null;
    try {
       if (localName==null && p.hasSuperValue()) return p.getSuperParam().getSValue();      
    } catch (RemoteException e) {System.err.println(e); return null;}
    return p.getSValue();
 }
  

  /** Der mgliche Double-Wert (short, byte, int, double) des Parameters, je nach Typ.
      Wenn dieser Parameter durch eine Wertbelegung innerhalb der Superkomponente
      belegt wurde, wird der Wert dieser Belegung zurckgegeben.<br>

        Falls localName=null ist, ist dies ein Parameter, den diese abstrakte 
      Komponente anbietet.
      Ansonsten reprsentiert dieses Parameterobjekt die Wertbelegung eines 
      Parameters mit gleichem Namen der Subkomponente "localName".
      @return 0, falls Parameter nicht belegt oder Typ kein Double-Typ.
      @see Param
      @see ParamBinding

  */ 
  public double      getParamNValue(String name, String localName) {
    Param p = getParam(name, localName);
    if (p == null) return 0.0;
    try {
       if (localName==null && p.hasSuperValue()) return p.getSuperParam().getNValue();      
    } catch (RemoteException e) {System.err.println(e); return 0.0;}
    return p.getNValue();
 }
  

  /** Der mgliche char-Wert.
      Wenn dieser Parameter durch eine Wertbelegung innerhalb der Superkomponente
      belegt wurde, wird der Wert dieser Belegung zurckgegeben.<br>
      Falls localName=null ist, ist dies ein Parameter, den diese abstrakte 
      Komponente anbietet.
      Ansonsten reprsentiert dieses Parameterobjekt die Wertbelegung eines 
      Parameters mit gleichem Namen der Subkomponente "localName".
      @return '?', falls Parameter nicht belegt oder Typ kein Char-Typ.
      @see Param
      @see ParamBinding
  */ 
  public char       getParamCValue(String name, String localName) {
    Param p = getParam(name, localName);
    if (p == null) return '?';
    try {
       if (localName==null && p.hasSuperValue()) return p.getSuperParam().getCValue();      
    } catch (RemoteException e) {System.err.println(e); return '?';}
    return p.getCValue();
 }

  /** Der mgliche Wert als Object.
      Wenn dieser Parameter durch eine Wertbelegung innerhalb der Superkomponente
      belegt wurde, wird der Wert dieser Belegung zurckgegeben.<br>
      Falls localName=null ist, ist dies ein Parameter, den diese abstrakte 
      Komponente anbietet.
      Ansonsten reprsentiert dieses Parameterobjekt die Wertbelegung eines 
      Parameters mit gleichem Namen der Subkomponente "localName".
      @return '?', falls Parameter nicht belegt oder Typ kein Char-Typ.
      @see Param#getValue
      @see ParamBinding
  */ 
  public Object       getParamValue(String name, String localName) {
    Param p = getParam(name, localName);
    if (p == null) return null;
    try {
       if (localName==null && p.hasSuperValue()) return p.getSuperParam().getValue();      
    } catch (RemoteException e) {System.err.println(" getParamValue: "+e); return null;}
    return p.getValue();
 }

  // ENDE PARAMETER *********************************************************
  
  /** Anzahl der Subkomponenten dieser abstrakten Komponente.
      Also nur die direkten Kinder, die ihrerseits wiederum Subkomponenten 
      haben knnen.
  */
  public int  getNumberOfSubComponents() {
    return subComponents.size();
  }

  
  /** Subkomponenten dieser abstrakten Komponente.
      Also nur die direkten Kinder, die ihrerseits wiederum Subkomponenten 
      haben knnen.
      @return  Hashtable von Names / IR_Component
      @see IR_Component
  */
  public Hashtable getSubComponents() {
    return subComponents;
  }

 
  /** Die Subkomponente, die unter dem Namen instanceName (local_name) angemeldet ist.
      @return null, falls ein solcher Name nicht gefunden wird.
  */
  public IR_Component   getSubComponent(String instanceName) {
    if (!subComponents.containsKey(instanceName)) { return null; }
    return (IR_Component) subComponents.get(instanceName);
  };


 /**  Gibt eine Portverbindung zurck.
      Es gibt nur Verbindungen zu einer Subkomponente oder zwischen zwei 
      unterschiedlichen Subkomponenten. Wenn ein Port zur abstrakten
      Komponente gehrt, ist compOne oder compTwo entsprechend "null" oder ".".
      @see Port
      @see PortBinding
      @return null, falls PortBinding nicht gefunden wurde.
 */
 public IR_PortBinding getPortBinding (String compOne, String portOne, 
                                       String compTwo, String portTwo) {
    // Nur Verbindungen zu einer Subkomponente oder zwischen Subkomponenten !;
    //System.out.println("teste "+compOne+"."+portOne+"-"+compTwo+"."+portTwo);
    if (compOne != null && compOne.equals(".")) compOne = null;
    if (compTwo != null && compTwo.equals(".")) compTwo = null;
    if (!existsSubComponent(compOne) && compOne != null) return null;
    if (!existsSubComponent(compTwo) && compTwo != null) return null;

    IR_Component cOne = this;
    IR_Component cTwo = this;
    if (compOne != null) cOne = (IR_Component) this.subComponents.get(compOne);
    if (compTwo != null) cTwo = (IR_Component) this.subComponents.get(compTwo);
    if (cOne.equals(cTwo)) return null;
    try {
       IR_Port pOne = cOne.getIR_Port(portOne);
       IR_Port pTwo = cTwo.getIR_Port(portTwo);
       PortBinding pb = null;
       for (Enumeration e = portBindings.elements() ; e.hasMoreElements() ; ) {
   	    pb = (PortBinding) e.nextElement();
          if (pb.isPortBinding(pOne, pTwo)) {
              return pb;
          }
       }
    } catch (RemoteException e) {
      server.message("error", 1, "getPortBinding: " + e.toString());
    }
    return null;
  }

 
  /** Textuelle Aufzhlung aller Portverbindungen, die uns gehren. 
      <br>
      Das sind Verbindungen zu Subkomponenten und zwischen Subkomponenten.
      Um Verbindungen zu 'Nachbar'-Komponenten oder unserer Vaterkomponente 
      zu erhalten, mu diese Funktion fr die Vaterkomponente aufgerufen werden.
      @return Vector      von String
      @see PortBinding#toTxt
  */
  public Vector      getPortBindingsString() {
    Vector result = new Vector(portBindings.size());
    for (Enumeration pb = this.portBindings.elements(); pb.hasMoreElements(); ) {
   	 result.addElement( ((PortBinding) pb.nextElement()).toTxt() );
	 }
    return result;
  }

   /** Alle Port Verbindungen, die dieser abstrakten Komponente gehren.
       <br>
      Das sind Verbindungen zu Subkomponenten und zwischen Subkomponenten.
      Um Verbindungen zu 'Nachbar'-Componenten oder unserer Vaterkomponente 
      zu erhalten, mu diese Funktion fr die Vaterkomponente aufrufen werden.
      @return Vector      von IR_PortBinding
      @see PortBinding
   */
   public Vector      getPortBindings() {
      return portBindings;
   }

     
 /** Alle verbundenen Ports mit dem Port port in textueller Form.
      Entweder ist der Port mit Namen port mit einem Port einer 
      Subkomponente verbunden: <b>instanceNameOfSubcomponent+"."+portName </b>
      oder mit einem Parameter der Vaterkomponente: <b> ".."+portName</b>
      oder mit einer Subkomponente der Vaterkomponente: <b> ".." + 
      InstaceNameofSubcomponent + "." + portName</b>
      @return Vector      von String
   */
  public Vector      getConnectedPortsString(String port) {
    try {
    Port p = getPort(port);
    if (p == null) return null;
    PortBinding cp = null;
    Vector result = new Vector();
    for (Enumeration pb = p.getBindings().elements(); pb.hasMoreElements(); ) {
       cp = (PortBinding) pb.nextElement();
       if (this.portBindings.contains(cp)) { 
	       if (p.equals(cp.getPortOne()) ) result.addElement(cp.getInstanceNameTwo() + 
                                             "." + cp.getPortTwo().getName());
          else                      result.addElement(cp.getInstanceNameOne() + 
                                             "." + cp.getPortOne().getName());
      } else {  // horizontale Bindung, cp muss in portBindings unseres Vaters sein 
      	if (p.equals(cp.getPortOne()) ) {
            if (predecessor.existsPort(cp.getPortTwo().getName())) {
               result.addElement(".." + cp.getPortTwo().getName());
            } else {
               result.addElement(".." + cp.getInstanceNameTwo() + "." + 
                                 cp.getPortTwo().getName());
            }
	      } else {
            if (predecessor.existsPort(cp.getPortOne().getName())) {
               result.addElement(".." + cp.getPortOne().getName());
            } else {
               result.addElement(".." + cp.getInstanceNameOne() + "." + 
                                       cp.getPortOne().getName());
            }
         } 
      } // endif my portBinding
    }
    return result;
    } catch (RemoteException e) {
       server.message("error", 1, e.toString());
       return null;
    }
  }


 /** Alle Ports, die mit dem Port mit Namen port unserer komplexen Komponente verbunden sind.
     Die verbundenen Ports sind entweder die von Subkomponenten, oder von der 
     Vaterkomponente oder von Subkomponenten der Vaterkomponente.
     @param port Der Namen des Ports.
     @return Vector      von IR_Port. Null, falls Port port nicht vorhanden.
     @see Port
 */
 public Vector      getConnectedPorts(String port) {
    Port p = getPort(port);
    if (p == null) return null;
    PortBinding cp = null;
    Vector result = new Vector();
    for (Enumeration pb = p.getBindings().elements(); pb.hasMoreElements(); ) {
       cp = (PortBinding) pb.nextElement();
       if (p.equals(cp.getPortOne()) ) result.addElement(cp.getPortTwo());
       else                      result.addElement(cp.getPortOne());
    }
    return result;
  }
   
   /** Textuelle Aufzhlung aller Parameterverbindungen. 
      <br>
      Das sind Verbindungen zu Parametern einer Subkomponenten.
      Um Verbindungen zu 'Nachbar'-Komponenten oder unserer Vaterkomponente 
      zu erhalten, mu diese Funktion fr die Vaterkomponente aufrufen werden.
      @return Vector      von String
      @see ParamBinding#toTxt
  */
  public Vector      getParamBindingsString() {
    Vector result = new Vector(paramBindings.size());
    for (Enumeration pb = this.paramBindings.elements(); pb.hasMoreElements(); ) {
   	 result.addElement( ((ParamBinding) pb.nextElement()).toTxt() );
	 }
    return result;
  }
  
   /** Alle Parameter Verbindungen, die dieser abstrakten Komponente gehren.
       <br>
      Das sind Verbindungen zu Subkomponenten.
      Um Verbindungen zu  unserer Vaterkomponente 
      zu erhalten, mu diese Funktion fr die Vaterkomponente aufgerufen werden.
      @return Vector      von IR_ParamBinding
      @see ParamBinding
   */
   public Vector      getParamBindings() {
      return paramBindings;
   }

   /**  Gibt eine Parameterverbindung zurck.
      Es gibt nur Verbindungen zu einer Subkomponente 
      @see Param
      @see ParamBinding
      @return null, falls ParamBinding nicht gefunden wurde.
 */
 public IR_ParamBinding getParamBinding (String paramOne, String compTwo, String paramTwo) {
    // Nur Verbindungen zu einer Subkomponente oder zwischen Subkomponenten !;
    if (!existsSubComponent(compTwo)) return null;
       
    IR_Component cOne = this;
    IR_Component cTwo = this;
    if (compTwo != null) cTwo = (IR_Component) this.subComponents.get(compTwo);
    if (cOne.equals(cTwo)) return null;
    try {
       IR_Param pOne = cOne.getIR_Param(paramOne, null);
       IR_Param pTwo = cTwo.getIR_Param(paramTwo, null);
       ParamBinding pb = null;
       for (Enumeration e = paramBindings.elements() ; e.hasMoreElements() ; ) {
   	    pb = (ParamBinding) e.nextElement();
          if (pb.isParamBinding(pOne, pTwo)) {
              return pb;
          }
       }
    } catch (RemoteException e) {
      server.message("error", 1, "getParamBinding: " + e.toString());
    }
    return null;
  }

  /** Alle verbundenen Parameter mit dem Parameter param in textueller Form.
      Entweder ist der Parameter mit Namen param mit einem Parameter einer 
      Subkomponente verbunden: <b>instanceNameOfSubcomponent+"."+paramName </b>
      oder mit einem Parameter der Vaterkomponente: <b> ".."+paramName</b>
      @return Vector      von String
      */
  public Vector          getConnectedParamsString(String param) {
  try
  {
  	 Param p = getParam(param, null);
    if (p == null) return null;
    ParamBinding pb = null;
    Vector result = new Vector();
    for (Enumeration b = p.getBindings().elements() ; b.hasMoreElements() ; ) {
      pb = (ParamBinding) b.nextElement();
      if (this.paramBindings.contains(pb)) { // unsere ParamBinding **********
	     if (pb.getParamOne().getName().equals(param)) 
             result.addElement(pb.getInstanceNameTwo() + "." + pb.getParamTwo().getName());
        else result.addElement(pb.getInstanceNameOne() + "." + pb.getParamOne().getName());
      } else {
	     if (pb.getParamOne().getName().equals(param)) 
            result.addElement(".." + pb.getParamTwo().getName());
        else result.addElement(".." + pb.getParamOne().getName());
      }
    }
    return result;
  }
  catch (RemoteException e) {
    server.message("error", 1, e.toString());
    return null;
  }
  } // end getConnectedParamString

  /** Alle verbundenen Parameter mit dem Parameter param.
      Entweder ist der Parameter mit Namen param mit einem Parameter einer 
      Subkomponente oder der Vaterkomponente verbunden.
      @return Vector      von IR_Param
      @see Param
      */
  public Vector          getConnectedParams(String param) {
    Param p = getParam(param, null);
    if (p == null) return null;
    ParamBinding pb = null;
    Vector result = new Vector();
    for (Enumeration b = p.getBindings().elements(); b.hasMoreElements();) {
      pb = (ParamBinding) b.nextElement();
      if (p.equals(pb.getParamOne())) result.addElement(pb.getParamTwo());
      else                     result.addElement(pb.getParamOne());
      } 
   return result;
  }
  
  // 
  //     *** For creation of the datastructure
  //

  /** Erzeugt einen neuen Port fr diese komplexe Komponente.
      Sollte der Port schon existieren, wird er nicht angelegt.
      @param name Der Name des Ports.
      @param type Der Typ des Ports.
      @param polarity "provided" oder "required" 
      @param protocol "event" oder "sharedObject"
      @return Der neue Port oder null, falls er nicht angelegt wurde
      @see Port
  */
  public IR_Port addPort(String name, String type, String polarity, String protocol) {
    if (name==null || type==null || polarity==null || protocol == null) return null;
    if (existsPort(name)) {
       server.message("error", 1, "Port " + name + " existiert schon bei der Komponente " + this.typeName + ":"+this.name);
       return null;
    }
    try {
      Port newPort = new Port(name, type, polarity, protocol, this);
      this.ports.addElement(newPort);
      server.message("comp", 5, "New "+newPort.print()+" fr "+getName()+":"+getTypeName());
       // An Proxy Objekte weitergeben!
       IR_Proxy proxy = null;
       for( Enumeration ep = this.proxies.elements(); ep.hasMoreElements(); ) {
    	    proxy = (IR_Proxy) ep.nextElement();
          try {
            proxy.addPortProxy(newPort);
          } catch (RemoteException proxyerror) {
            server.message("error",1,"addPort: Weitergabe an Proxy fr Component " 
                            + getFullName() + " fehlgeschlagen: " + proxyerror.toString());
          }
       }
       // Ende an Proxy Objekte weitergeben
    // Interessenten benachrichtigen.
    Object[] o = new Object[4];
    o[0] = name;
    o[1] = type;
    o[2] = polarity;
    o[3] = protocol;
    server.TailorActionPerformed(this, "addPort", o);

      return newPort;
    } catch (RemoteException e) {
      server.message("error", 1, e.toString());
    }
    return null;
  }


  /** Erzeugt einen neuen Parameter.
      Sollte der Parameter schon existieren, wird er nicht angelegt.
      Falls localName=null ist, ist dies ein Parameter, den diese abstrakte 
      Komponente anbietet.
      Ansonsten reprsentiert dieses Parameterobjekt die Wertbelegung eines 
      Parameters mit gleichem Namen der Subkomponente "localName".
      @param name Der Name des Parameters.
      @param type Der Typ des Parameters.
      @return Der neue Parameter oder null, falls er nicht angelegt wurde.
      @see Param
  */
  public IR_Param addParam (String name, String type, String localName) {
    if (name==null || type==null) return null;
    if (existsParam(name, localName)) {
       server.message("error", 1, "Param " + name + " existiert schon bei der Komponente " + this.typeName + ":"+this.name);
       return null;
    }
    try {
       Param nP = new Param(name, type, localName, this);
       this.params.addElement(nP);
       server.message("comp", 4, "New " + nP.print());
       if (localName != null) {
          IR_Component sc = getSubComponent(localName);
          if (sc == null) { 
               server.message("error", 1, "Komponente " + this.typeName + "("+this.name+") hat keine Subkomponente " + localName);
               return nP;
          }
          IR_Param sp = sc.getIR_Param(name, null);
          sp.setSuperParam(nP);
       } else {
           // An Proxy Objekte weitergeben!
           IR_Proxy proxy = null;
           for( Enumeration ep = this.proxies.elements(); ep.hasMoreElements(); ) {
              proxy = (IR_Proxy) ep.nextElement();
              try {
                proxy.addParamProxy(nP);
              } catch (RemoteException proxyerror) {
                server.message("error",1,"addParam: Weitergabe an Proxy fr Component " 
                                + getFullName() + " fehlgeschlagen: " + proxyerror.toString());
              }
           }
           // Ende an Proxy Objekte weitergeben
       }
    // Interessenten benachrichtigen.
    Object[] o = new Object[3];
    o[0] = name;
    o[1] = type;
    o[2] = localName;
    server.TailorActionPerformed(this, "addParam", o);

       return nP;
    } catch (RemoteException e) {
      server.message("error", 1, e.toString());
    }
    return null;
  }


  /** Setzt den Parameter name auf einen bestimmten Wert vom Typ type.
      Es wird nur ein Wert gesetzt (svalue, nvalue, cvalue).<br>
      Ein Parameter darf nicht gesetzt werden, wenn der Parameter nach oben
      weitergereicht wird.
      Falls localName=null ist, ist dies ein Parameter, den diese abstrakte 
      Komponente anbietet.
      Ansonsten reprsentiert dieses Parameterobjekt die Wertbelegung eines 
      Parameters mit gleichem Namen der Subkomponente "localName".
      @return false, falls der Parameter nicht auf den Wert gesetzt werden konnte.
      @see Param
      @see ParamBinding
  */
  public boolean setParam (String name, String localName, 
                           String svalue, double nvalue, char cvalue) {
    Param p = getParam(name, localName);
    if (p == null) return false;
    if (p.isConnected(this.predecessor)) return false;
    p.setValue(svalue);
    p.setValue(nvalue);
    p.setValue(cvalue);
    server.message("comp", 4, "New Value " + p.print());
    if (localName == null) {
       // An Proxy Objekte weitergeben!
       IR_Proxy proxy = null;
       for( Enumeration ep = this.proxies.elements(); ep.hasMoreElements(); ) {
    	    proxy = (IR_Proxy) ep.nextElement();
          try {
            proxy.setParam(name, p.getValue());
          } catch (RemoteException proxyerror) {
            server.message("error",1,"setParam: Weitergabe an Proxy fr Component " 
                            + getFullName() + " fehlgeschlagen: " + proxyerror.toString());
          }
       }

       // Ende an Proxy Objekte weitergeben
    } else {
        try {
            IR_Component sc = getSubComponent(localName);
            sc.setSubCompParam(name, p.getValue());
        } catch (RemoteException proxyerror) {
          server.message("error",1,"setParam:  " + proxyerror.toString());
        }
    }
    return true;
  }
  
 public void setSubCompParam (String name, Object value) {
       // An Proxy Objekte weitergeben!
       IR_Proxy proxy = null;
       for( Enumeration ep = this.proxies.elements(); ep.hasMoreElements(); ) {
    	    proxy = (IR_Proxy) ep.nextElement();
          try {
            proxy.setParam(name, value);
          } catch (RemoteException proxyerror) {
            server.message("error",1,"setParam: Weitergabe an Proxy fr Component " 
                            + getFullName() + " fehlgeschlagen: " + proxyerror.toString());
          }
       }
           // Interessenten benachrichtigen.
    Object[] o = new Object[2];
    o[0] = name;
    o[1] = value;
    server.TailorActionPerformed(this, "setSubCompParam", o);

       // Ende an Proxy Objekte weitergeben
 }
  /** Markiert den Parameter name als ungesetzt, damit er wieder verbunden werden kann.
      Macht nur Sinn bei einem realen Parameter und nicht bei einem Wertbelegungsparameter
      fr einen Subkomponentenparameter.
      @see Param#unSetValue
  */
  public void unSetParam (String name){
    Param p = getParam(name, null);
    if (p == null) return;
    p.unSetValue();
       // An Proxy Objekte weitergeben!
       IR_Proxy proxy = null;
       for( Enumeration ep = this.proxies.elements(); ep.hasMoreElements(); ) {
    	    proxy = (IR_Proxy) ep.nextElement();
          try {
            proxy.unsetParam(name);
          } catch (RemoteException proxyerror) {
            server.message("error",1,"unSetParam: Weitergabe an Proxy fr Component " 
                            + getFullName() + " fehlgeschlagen: " + proxyerror.toString());
          }
       }
       // Ende an Proxy Objekte weitergeben
      // Interessenten benachrichtigen.
      Object[] o = new Object[1];
      o[0] = name;
      server.TailorActionPerformed(this, "unSetParam", o);

  }
  

  // ersetzt ggf. alte Subkomponente !
  /** Neue Subkomponente unter dem (local) Namen <i>instanceName</i> anmelden.
      Es darf noch keine Subkomponente mit diesem Namen angemeldet sein.
      @return Die angemeldete Subkomponente oder null, falls sie nicht angemeldet werden konnte.
  */
  public IR_Component addSubComponent(IR_Component subComponent, String instanceName) {
    if (existsSubComponent(instanceName)) return null;
    try {
       this.subComponents.put(instanceName, subComponent); // gibt alte Subkomponente mit demselben Namen zurueck...;
       subComponent.setPredecessor(this);
       subComponent.setName(instanceName);
       server.message("comp", 4, "Neue Subkomponente " + instanceName + " an " + 
         typeName + " angemeldet ");
       // An Proxy Objekte weitergeben!
       IR_Proxy proxy = null;
       for( Enumeration ep = this.proxies.elements(); ep.hasMoreElements(); ) {
    	    proxy = (IR_Proxy) ep.nextElement();
          try {
            proxy.addSubProxy(subComponent);
          } catch (RemoteException proxyerror) {
            server.message("error",1,"addSubComponent: Weitergabe an Proxy fr Component " 
                            + getFullName() + " fehlgeschlagen: " + proxyerror.toString());
          }
       }
       // Ende an Proxy Objekte weitergeben
        // Interessenten benachrichtigen.
        Object[] o = new Object[2];
        o[0] = subComponent;
        o[1] = instanceName;
        server.TailorActionPerformed(this, "addSubComponent", o);

       return subComponent;
    } catch (RemoteException e) {
       server.message("error", 1, "addSubComponent: " + e.toString());
       return null;
    }
  }


  /** Zwei Ports verbinden.
      Es sind nur Verbindungen zu einer Subkomponente oder zwischen zwei 
      unterschiedlichen Subkomponenten moeglich. Wenn eine Port zur abstrakten
      Komponente gehrt, ist compOne oder compTwo entsprechend "null" oder ".".
      @see Port
      @see PortBinding
  */
  public IR_PortBinding addPortBinding (String compOne, String portOne,
			      String compTwo, String portTwo) {
    if (compOne != null && compOne.equals(".")) compOne = null;
    if (compTwo != null && compTwo.equals(".")) compTwo = null;
    // Nur Verbindungen zu einer Subkomponente oder zwischen Subkomponenten !;
    if (compOne != null && !existsSubComponent(compOne)) return null;
    if (compTwo != null && !existsSubComponent(compTwo)) return null;

    if (existsPortBinding(compOne, portOne, compTwo, portTwo)) return null;
    IR_Component cOne = this;
    IR_Component cTwo = this;
    if (compOne != null) cOne = (IR_Component) this.subComponents.get(compOne);
    else compOne = ".";
    if (compTwo != null) cTwo = (IR_Component) this.subComponents.get(compTwo);
    else compTwo = ".";
    if (cOne.equals( cTwo)) return null;
    if (cOne == null) { server.message("error", 1, "addPortBinding: Komponente" + compOne + " nicht gefunden bei Komponente "+this.typeName); return null;}
    if (cTwo == null) { server.message("error", 1, "addPortBinding: Komponente" + compTwo + " nicht gefunden bei Komponente "+this.typeName); return null;}
    try {
       IR_Port pOne = cOne.getIR_Port(portOne);
       IR_Port pTwo = cTwo.getIR_Port(portTwo);
       if (pOne == null)  {server.message("error", 1, "addPortBinding: Port " + compOne + "." + portOne + " nicht gefunden bei Komponente "+this.typeName); return null; } 
       if (pTwo == null)  {server.message("error", 1, "addPortBinding: Port " + compTwo + "." + portTwo + " nicht gefunden bei Komponente "+this.typeName); return null; } 
       PortBinding pb = new PortBinding(pOne, pTwo, this);
       this.portBindings.addElement(pb);
       pOne.addBinding(pb);
       pTwo.addBinding(pb);
       String s = pb.toTxt();
       server.message("comp", 5, "addPortBinding: " + s);
       // An Proxy Objekte weitergeben!
       IR_Proxy proxy = null;
       for( Enumeration ep = this.proxies.elements(); ep.hasMoreElements(); ) {
    	    proxy = (IR_Proxy) ep.nextElement();
          try {
            proxy.bindPorts(compOne, portOne, compTwo, portTwo);
          } catch (RemoteException proxyerror) {
            server.message("error",1,"addPortBinding: Weitergabe an Proxy fr Component " 
                            + getFullName() + " fehlgeschlagen: " + proxyerror.toString());
          }
       }
       // Ende an Proxy Objekte weitergeben
       // Interessenten benachrichtigen.
        Object[] o = new Object[4];
        o[0] = compOne;
        o[1] = portOne;
        o[2] = compTwo;
        o[3] = portTwo;
        server.TailorActionPerformed(this, "addPortBinding", o);

       return pb;
     } catch (RemoteException e) {
       server.message("error",1 , "addPortBinding: " + e.toString());
     }
     return null;
  }
  
  

  // Nur Bindungen von unserem Parameter zu einer Subkomponente moeglich !;
  /** Zwei Parameter verbinden.
      Es sind nur Verbindungen zu einer Subkomponente moeglich. Und nur zu einem
      Parameter, der noch keinen Wert gesetzt hat.<br>
      todo: selben typ sicherstellen
      Parameter <i>paramOne</i> mu ein lokaler Parameter sein.
      @see Param
      @see ParamBinding
      @see #unSetParam
  */
  public IR_ParamBinding addParamBinding(String paramOne, String compTwo, String paramTwo) {
    if (!existsSubComponent(compTwo) || compTwo==null) return null;
    IR_Component cTwo = (IR_Component) this.subComponents.get(compTwo);
    if (cTwo == null) return null;
    if (existsParamBinding(paramOne, compTwo, paramTwo)) return null;    
    try {
       IR_Param pOne = this.getParam(paramOne, null); // Keine subcompParams!
       if (pOne == null) return null; 
       IR_Param pTwo = cTwo.getIR_Param(paramTwo, null); // Keine subcompParams!
       if (pTwo == null)  return null;
       if (pTwo.isValueSet() || pTwo.hasSuperValue()) {
           server.message("error", 4, "addParamBinding versucht an einen gesetzten  Param anzubinden: " + compTwo + "." + paramTwo+ " bei Komponente "+this.typeName);
           return null;
       }
   
       ParamBinding pb = new ParamBinding(pOne, pTwo, this);
       this.paramBindings.addElement(pb);
       pOne.addBinding(pb);
       pTwo.addBinding(pb);
       server.message("comp", 5, "addParamBinding: " + pb.toTxt());
       // An Proxy Objekte weitergeben!
       IR_Proxy proxy = null;
       for( Enumeration ep = this.proxies.elements(); ep.hasMoreElements(); ) {
    	    proxy = (IR_Proxy) ep.nextElement();
          try {
            proxy.bindParams(paramOne, compTwo, paramTwo);
          } catch (RemoteException proxyerror) {
            server.message("error",1,"addParamBinding: Weitergabe an Proxy fr Component " 
                            + getFullName() + " fehlgeschlagen: " + proxyerror.toString());
          }
       }
       // Ende an Proxy Objekte weitergeben
       // Interessenten benachrichtigen.
        Object[] o = new Object[3];
        o[0] = paramOne;
        o[1] = compTwo;
        o[2] = paramTwo;
        server.TailorActionPerformed(this, "addParamBinding", o);

       return pb;
    } catch (RemoteException e) {
       server.message("error",1 , "addParamBinding: " + e.toString());
    }
    return null;
}


  /** Kopiert die komplexe Komponente. 
  */
  public IR_Component copy(String newTypeName) {
    Component result = null;
    try {
      comp_io cio = new comp_io(server);
      String CAT = cio.getCompString(this);
      //server.message("comp",3,"copy CAT: " + CAT);
      Component c = new Component(server);
      result = cio.getComponent(CAT, this.typeName, c);
      result.typeName = newTypeName;
      return result;
    } catch (RemoteException e) {
      server.message("error", 1, "Component.copy: "+ e.getClass().getName() + ": " + e.getMessage());
      return null;
    }

  }  // copies datastructure


  /** Kopiert die komplexe Komponente. 
  */
  public Component local_copy(String newTypeName) {
    Component result = null;
    try {
      comp_io cio = new comp_io(server);
      String CAT = cio.getCompString(this);
      //server.message("comp",3,"copy CAT: " + CAT);
      Component c = new Component(server);
      result = cio.getComponent(CAT, this.typeName, c);
      result.typeName = newTypeName;
      return result;

    } catch (RemoteException e) {
      server.message("error", 1, "Component.copy: "+ e.getClass().getName() + ": " + e.getMessage());
      return null;
    }

  }  // copies datastructure
  //
  //     *** destruction of the datastructure
  //;

  /** Entfernt den Port p und alle Verbindungen an ihn.
      Entfernt keine Verbindugen rekursiv.
  */
  void deletePort(Port p) {
    if (p == null) return; 
    Vector v = (Vector) p.getBindings().clone();
    for (Enumeration e =  v.elements() ; e.hasMoreElements() ; ) {
	    deletePortBinding( (PortBinding) e.nextElement()) ;
    }
    this.ports.removeElement(p);
    server.message("comp", 5, "Port " + p.getName() + " von Component " +
         typeName + " gelscht.");
     // An Proxy Objekte weitergeben!
     IR_Proxy proxy = null;
     for( Enumeration ep = this.proxies.elements(); ep.hasMoreElements(); ) {
        proxy = (IR_Proxy) ep.nextElement();
        try {
          proxy.removePortProxy(p.getName());
        } catch (RemoteException proxyerror) {
          server.message("error",1,"deletePort: Weitergabe an Proxy fr Component " 
                          + getFullName() + " fehlgeschlagen: " + proxyerror.toString());
        }
     }
     // Ende an Proxy Objekte weitergeben
       // Interessenten benachrichtigen.
        Object[] o = new Object[1];
        o[0] = p;
        server.TailorActionPerformed(this, "deletePort", o);

  }

/**  Ruft deletePort(String portName) auf.
 */
 public void deletePort(IR_Port p) {
    if (p==null) return;
    if (!ports.contains(p)) {
          server.message("error",1,"Component.deletePort: bergebener Port gehrt nicht zur Komponente " + this.typeName);
          return;
    }
    try  {
      String p1 = p.getName();
      deletePort(p1);                    
    }
    catch (RemoteException ex) {
        server.message("error",1,"Component.deletePort: " +ex);
        return;
    }
 }

  /** Entfernt den Port mit Namen name und alle Verbindungen an ihn.
      Entfernt keine Verbindugen rekursiv.
  */
  public void deletePort(String name) {
    Port p = this.getPort(name);
    if (p == null) return; // i don't need to delete this port';
    Vector v = (Vector) p.getBindings().clone();
     for (Enumeration e = v.elements() ; e.hasMoreElements() ; ) {
	    deletePortBinding( (PortBinding) e.nextElement());
    }
    this.ports.removeElement(p);
     // An Proxy Objekte weitergeben!
     IR_Proxy proxy = null;
     for( Enumeration ep = this.proxies.elements(); ep.hasMoreElements(); ) {
        proxy = (IR_Proxy) ep.nextElement();
        try {
          proxy.removePortProxy(p.getName());
        } catch (RemoteException proxyerror) {
          server.message("error",1,"deletePort: Weitergabe an Proxy fr Component " 
                          + getFullName() + " fehlgeschlagen: " + proxyerror.toString());
        }
     }
     // Ende an Proxy Objekte weitergeben
        // Interessenten benachrichtigen.
        Object[] o = new Object[1];
        o[0] = name;
        server.TailorActionPerformed(this, "deletePort", o);

    server.message("comp", 5, "Port " + p.getName() + " von Component " +
         typeName + " gelscht.");
  }
  
/**  Ruft deleteParam(String Name, String Subkomponentenname) auf.
 */
 public void deleteParam(IR_Param p) {
    if (p==null) return;
    if (!params.contains(p)) {
          server.message("error",1,"Component.deleteParam: bergebener Parameter gehrt nicht zur Komponente " + this.typeName);
          return;
    }
    try  {
      String p1 = p.getName();
      String c1 = p.getSubCompName();
      deleteParam(p1,c1);                    
    }
    catch (RemoteException ex) {
        server.message("error",1,"Component.deleteParam: " +ex);
        return;
    }
 }


  /** Entfernt den Parameter p und alle Verbindungen an ihn.
      Entfernt keine Verbindugen rekursiv.
  */
  void deleteParam (Param p) {
    if (p == null) return; // i don't need to delete this param';
    Vector v = (Vector) p.getBindings().clone();
    for (Enumeration e = v.elements() ; e.hasMoreElements() ; ) {
   	 deleteParamBinding( (ParamBinding) e.nextElement());
    }
    this.params.removeElement(p);
    if (!p.isSubCompParam()) {
       // An Proxy Objekte weitergeben!
       IR_Proxy proxy = null;
       for( Enumeration ep = this.proxies.elements(); ep.hasMoreElements(); ) {
          proxy = (IR_Proxy) ep.nextElement();
          try {
            proxy.removeParamProxy(p.getName());
          } catch (RemoteException proxyerror) {
            server.message("error",1,"deleteParam: Weitergabe an Proxy fr Component " 
                            + getFullName() + " fehlgeschlagen: " + proxyerror.toString());
          }
       }
       // Ende an Proxy Objekte weitergeben
       // Interessenten benachrichtigen.
        Object[] o = new Object[1];
        o[0] = p;
        server.TailorActionPerformed(this, "deleteParam", o);

    }
    server.message("comp", 5, "Param " + p.getName() + " von Component " +
         typeName + " gelscht.");
}


 
  /** Entfernt den Parameter p und alle Verbindungen an ihn.
      Entfernt keine Verbindugen rekursiv.
  */
  public void deleteParam (String name, String localName) {
    Param p = this.getParam(name, localName);
    if (p == null) return; // i don't need to delete this param';
    Vector v = (Vector) p.getBindings().clone();
    for (Enumeration e = v.elements() ; e.hasMoreElements() ; ) {
   	 deleteParamBinding( (ParamBinding) e.nextElement());
    }
    this.params.removeElement(p);
    if (localName != null) {
          IR_Component sc = getSubComponent(localName);
          if (sc == null) { 
               server.message("error", 1, "Komponente " + this.name + " hat keine Subkomponente " + localName);
               return ;
          }
          try {
             IR_Param sp = sc.getIR_Param(name, null);
             sp.setSuperParam(null);
          } catch (RemoteException e) { System.err.println(e); }
    } else {
       // An Proxy Objekte weitergeben!
       IR_Proxy proxy = null;
       for( Enumeration ep = this.proxies.elements(); ep.hasMoreElements(); ) {
          proxy = (IR_Proxy) ep.nextElement();
          try {
            proxy.removeParamProxy(p.getName());
          } catch (RemoteException proxyerror) {
            server.message("error",1,"deleteParam: Weitergabe an Proxy fr Component " 
                            + getFullName() + " fehlgeschlagen: " + proxyerror.toString());
          }
       }
       // Ende an Proxy Objekte weitergeben
       // Interessenten benachrichtigen.
        Object[] o = new Object[2];
        o[0] = name;
        o[1] = localName;
        server.TailorActionPerformed(this, "deleteParam", o);

    }
    server.message("comp", 5, "Param " + p.getName() + " von Component " +
         typeName + " gelscht.");
 }

   /** Abmelden einer Subkomponente, die unter dem Instanznamen local_name angemeldet war.
       @return Die entfernte Subkomponente oder null, falls sie nicht angemeldet war.
   */
   public IR_Component removeSubComponent(String local_name) {
      IR_Component cOld = null;
      if (!existsSubComponent(local_name)) return null;
      server.message("comp", 5, "Subkomponente " + local_name + " von " + name +
         " gelscht.");
      deletePortBindingsWithComponent(local_name);
      deleteParamBindingsWithComponent(local_name);
      try {
         ((IR_Component) subComponents.get(local_name)).setPredecessor(null);
      } catch (RemoteException e) {
         server.message("error", 1, "Component.removeSubComponent: " + e.toString());
      }
      cOld = (IR_Component) subComponents.remove(local_name);

       // An Proxy Objekte weitergeben!
       IR_Proxy proxy = null;
       for( Enumeration ep = this.proxies.elements(); ep.hasMoreElements(); ) {
//        System.out.println("3333");
          proxy = (IR_Proxy) ep.nextElement();
          try {
            proxy.removeSubProxy(local_name);
          } catch (RemoteException proxyerror) {
            server.message("error",1,"removeSubComponent: Weitergabe an Proxy fr Component " 
                            + getFullName() + " fehlgeschlagen: " + proxyerror.toString());
          }
       }
       // Ende an Proxy Objekte weitergeben
       // Interessenten benachrichtigen.
        Object[] o = new Object[1];
        o[0] = local_name;
        server.TailorActionPerformed(this, "removeSubComponent", o);

      return (IR_Component) cOld;
   }

  /** Entfernt eine Port Verbindung und meldet diese von beiden Ports ab.
  */
  void deletePortBinding(PortBinding pb) {
    try {
       if (pb == null) return;
       if (pb.getPortOne() != null) pb.getPortOne().removeBinding(pb);
       if (pb.getPortTwo() != null) pb.getPortTwo().removeBinding(pb);
       pb.owner.portBindings.removeElement(pb);
   } catch (RemoteException e) {
         server.message("error", 1, "Component.deletePortBinding: " + e.toString());
   }
  }

 /**  Ruft deletePortBinding(String compOne, String portOne, String compTwo, String portTwo) auf.
 */
 public IR_PortBinding deletePortBinding(IR_PortBinding pb) {
    if (pb==null) return null;
    if (!portBindings.contains(pb)) {
          server.message("error",1,"Component.deletePortBinding: bergebenes Portverbindungsobjekt gehrt nicht zur Komponente " + this.typeName);
          return null;
    }
    try  {
    	String c1 = pb.getInstanceNameOne();
      String p1 = pb.getPortNameOne();
      String c2 = pb.getInstanceNameTwo();
      String p2 = pb.getPortNameTwo();
      return deletePortBinding(c1, p1, c2, p2);                    
    }
    catch (RemoteException ex) {
        server.message("error",1,"Component.deletePortBinding: " +ex);
        return null;
    }
 }
 
 /**  Eine Portverbindung lschen.
      Es gibt nur Verbindungen zu einer Subkomponente oder zwischen zwei 
      unterschiedlichen Subkomponenten moeglich. Wenn ein Port zur abstrakten
      Komponente gehrt, ist compOne oder compTwo entsprechend "null" oder ".".
      @see Port
      @see PortBinding
 */
 public IR_PortBinding deletePortBinding (String compOne, String portOne, 
                                          String compTwo, String portTwo) {
    // Nur Verbindungen zu einer Subkomponente oder zwischen Subkomponenten !;
    if (!existsSubComponent(compOne) && compOne != null && !compOne.equals(".")) return null;
    if (!existsSubComponent(compTwo) && compTwo != null && !compTwo.equals(".")) return null;
       
    IR_Component cOne = this;
    IR_Component cTwo = this;
    if (compOne != null) cOne = (IR_Component) this.subComponents.get(compOne);
    else compOne = ".";
    if (compTwo != null) cTwo = (IR_Component) this.subComponents.get(compTwo);
    else compTwo = ".";
    if (cOne == cTwo) return null;
    try {
       IR_Port pOne = cOne.getIR_Port(portOne);
       IR_Port pTwo = cTwo.getIR_Port(portTwo);
       PortBinding pb = null;
       Vector v = (Vector) portBindings.clone();
       for (Enumeration e = v.elements() ; e.hasMoreElements() ; ) {
   	    pb = (PortBinding) e.nextElement();
          if (pb.isPortBinding(pOne, pTwo)) {
              this.deletePortBinding(pb);
              server.message("comp", 4, "Lsche PortBinding " +pb.toTxt());
               // An Proxy Objekte weitergeben!
               IR_Proxy proxy = null;
               for( Enumeration ep = this.proxies.elements(); ep.hasMoreElements(); ) {
                  proxy = (IR_Proxy) ep.nextElement();
                  try {
                    proxy.unbindPorts(compOne, portOne, compTwo, portTwo);
                  } catch (RemoteException proxyerror) {
                    server.message("error",1,"removePortBinding: Weitergabe an Proxy fr Component " 
                                    + getFullName() + " fehlgeschlagen: " + proxyerror.toString());
                  }
               }
               // Ende an Proxy Objekte weitergeben
       // Interessenten benachrichtigen.
        Object[] o = new Object[4];
        o[0] = compOne;
        o[1] = portOne;
        o[2] = compTwo;
        o[3] = portTwo;
        server.TailorActionPerformed(this, "deletePortBinding", o);

              return pb;
          }
       }
    } catch (RemoteException e) {
      server.message("error", 1, "deletePortBinding: " + e.toString());
    }
    return null;
  }
  
  /** Lscht alle Verbindungen zum Port mit Namen aPort, der zu unserer komplexen Komponente gehrt.
      Dazu gehren auch Verbindungen, die unserer Vaterkomponente gehren und 
      damit zu einer Nachbarkomponente gehen. 
  */
  public void deletePortBindingsWithPort (String aPort) {
    Port p = getPort(aPort);
    PortBinding pb = null;
    if (p == null) return; // sorry ...;
    Vector v = (Vector) p.getBindings().clone();
    for (Enumeration e= v.elements(); e.hasMoreElements(); ) {
        pb = (PortBinding) e.nextElement();
      	this.deletePortBinding(pb);
        // An Proxy Objekte weitergeben!
         IR_Proxy proxy = null;
         for( Enumeration ep = this.proxies.elements(); ep.hasMoreElements(); ) {
            proxy = (IR_Proxy) ep.nextElement();
            try {
              proxy.unbindPorts(pb.getInstanceNameOne(), pb.getPortNameOne(),
                                pb.getInstanceNameTwo(), pb.getPortNameTwo());
            } catch (RemoteException proxyerror) {
              server.message("error",1,"deletePortBindingWithPort: Weitergabe an Proxy fr Component " 
                              + getFullName() + " fehlgeschlagen: " + proxyerror.toString());
            }
         }
         // Ende an Proxy Objekte weitergeben
        // Interessenten benachrichtigen.
        Object[] o = new Object[1];
        o[0] = aPort;
        server.TailorActionPerformed(this, "deletePortBindingsWithPort", o);
    }
  }
  
  /** Lscht alle Verbindungen unserer Ports zu Ports unserer Subkomponente mit dem Instanznamen local_name.
   */
  public void deletePortBindingsWithComponent (String local_name) {
    if (local_name == null) return ;
    PortBinding pb = null;
    Vector v = (Vector) portBindings.clone();
    for (Enumeration e = portBindings.elements() ; e.hasMoreElements() ; ) {
   	    pb = (PortBinding) e.nextElement();
        if (pb.connectsComp(local_name)) {
              this.deletePortBinding(pb);
               // An Proxy Objekte weitergeben!
               IR_Proxy proxy = null;
               for( Enumeration ep = this.proxies.elements(); ep.hasMoreElements(); ) {
                  proxy = (IR_Proxy) ep.nextElement();
                  try {
                    proxy.unbindPorts(pb.getInstanceNameOne(), pb.getPortNameOne(),
                                      pb.getInstanceNameTwo(), pb.getPortNameTwo());
                  } catch (RemoteException proxyerror) {
                    server.message("error",1,"deletePortBindingWithComponent: Weitergabe an Proxy fr Component " 
                                    + getFullName() + " fehlgeschlagen: " + proxyerror.toString());
                  }
               }
               // Ende an Proxy Objekte weitergeben
       // Interessenten benachrichtigen.
        Object[] o = new Object[1];
        o[0] = local_name;
        server.TailorActionPerformed(this, "deletePortBindingsWithComponent", o);

        }
    }          
  }; 

  
  /** Lscht eine Parameter Verbindung.
  */
  void deleteParamBinding(ParamBinding pb) {
    try {
       if (pb == null) return;
       if (pb.paramOne != null) pb.getParamOne().removeBinding(pb);
       if (pb.paramTwo != null) pb.getParamTwo().removeBinding(pb);
       pb.owner.paramBindings.removeElement(pb);
    }  catch (RemoteException e) {
       server.message("error", 1, "deletePortBinding: " + e.toString());
    }
  }

  

 /**  Ruft deleteParamBinding(String paramOne, String compTwo, String paramTwo) auf.
 */
 public IR_ParamBinding deleteParamBinding(IR_ParamBinding pb) {
    if (pb==null) return null;
    if (!paramBindings.contains(pb)) {
          server.message("error",1,"Component.deleteParamBinding: bergebenes Param-Verbindungsobjekt gehrt nicht zur Komponente " + this.typeName);
          return null;
    }
    try  {
      String p1 = pb.getParamNameOne();
      String c2 = pb.getInstanceNameTwo();
      String p2 = pb.getParamNameTwo();
      return deleteParamBinding(p1, c2, p2);                    
    }
    catch (RemoteException ex) {
        server.message("error",1,"Component.deleteParamBinding: " +ex);
        return null;
    }
 }

  
  /** Lscht alle Verbindungen zum Parameter mit Namen aParam, der zu unserer komplexen Komponente gehrt.
      Dazu gehren auch Verbindungen, die unserer Vaterkomponente gehen und nicht
      von uns erzeugt wurde.
  */
  public void deleteParamBindingsWithParam(String aParam) {
    Param p = getParam(aParam, null);
    ParamBinding pb =  null;
    if (p == null) return; // sorry ...;
    Vector v = (Vector) p.getBindings().clone();
    for (Enumeration e= v.elements(); e.hasMoreElements(); ) {
        pb = (ParamBinding) e.nextElement();
   	    deleteParamBinding(pb);
         // An Proxy Objekte weitergeben!
         IR_Proxy proxy = null;
         for( Enumeration ep = this.proxies.elements(); ep.hasMoreElements(); ) {
            proxy = (IR_Proxy) ep.nextElement();
            try {
              proxy.unbindParams(pb.getParamNameOne(), pb.getComponentOne().getName(), pb.getParamNameTwo());
            } catch (RemoteException proxyerror) {
              server.message("error",1,"deleteParamBindingWithParam: Weitergabe an Proxy fr Component " 
                              + getFullName() + " fehlgeschlagen: " + proxyerror.toString());
            }
         }
         // Ende an Proxy Objekte weitergeben
       // Interessenten benachrichtigen.
        Object[] o = new Object[1];
        o[0] = aParam;
        server.TailorActionPerformed(this, "deleteParamBindingWithParam", o);

    }
  }


  /** Lscht alle Verbindungen unserer Parameter zu Parametern unserer Subkomponente mit dem Instanznamen local_name.
  */
  public void deleteParamBindingsWithComponent (String local_name) {
    if (local_name == null) return ;
    ParamBinding pb = null;
    Vector v = (Vector) paramBindings.clone();
    for (Enumeration e = v.elements() ; e.hasMoreElements() ; ) {
   	    pb = (ParamBinding) e.nextElement();
          if (pb.connectsComp(local_name)) {
              this.deleteParamBinding(pb);
              // An Proxy Objekte weitergeben!
               IR_Proxy proxy = null;
               for( Enumeration ep = this.proxies.elements(); ep.hasMoreElements(); ) {
                  proxy = (IR_Proxy) ep.nextElement();
                  try {
                    proxy.unbindParams(pb.getParamNameOne(), pb.getComponentOne().getName(), pb.getParamNameTwo());
                  } catch (RemoteException proxyerror) {
                    server.message("error",1,"deleteParamBindingWithComponent: Weitergabe an Proxy fr Component " 
                                    + getFullName() + " fehlgeschlagen: " + proxyerror.toString());
                  }
               }
               // Ende an Proxy Objekte weitergeben
       // Interessenten benachrichtigen.
        Object[] o = new Object[1];
        o[0] = local_name;
        server.TailorActionPerformed(this, "deleteParamBindingsWithComponent", o);

        }
    }          
  }; 

  /**  Eine Parameterverbindung lschen.
      Es gibt nur Verbindungen zu einer Subkomponente
      @see Param
      @see ParamBinding
 */
 public IR_ParamBinding deleteParamBinding (String paramOne, String compTwo, String paramTwo) {
    // Nur Verbindungen zu einer Subkomponente;
    if (!existsSubComponent(compTwo)) return null;
       
    IR_Component cOne = this;
    IR_Component cTwo = this;
    if (compTwo != null) cTwo = (IR_Component) this.subComponents.get(compTwo);
    try {
       IR_Param pOne = cOne.getIR_Param(paramOne, null);
       IR_Param pTwo = cTwo.getIR_Param(paramTwo, null);
       ParamBinding pb = null;
       Vector v = (Vector) paramBindings.clone();
       for (Enumeration e = v.elements() ; e.hasMoreElements() ; ) {
   	    pb = (ParamBinding) e.nextElement();
          if (pb.isParamBinding(pOne, pTwo)) {
              this.deleteParamBinding(pb);
              server.message("comp", 4, "Lsche ParameterBinding " +pb.toTxt());
              // An Proxy Objekte weitergeben!
               IR_Proxy proxy = null;
               for( Enumeration ep = this.proxies.elements(); ep.hasMoreElements(); ) {
                  proxy = (IR_Proxy) ep.nextElement();
                  try {
                    proxy.unbindParams(paramOne, compTwo, paramTwo);
                  } catch (RemoteException proxyerror) {
                    server.message("error",1,"deleteParamBinding: Weitergabe an Proxy fr Component " 
                                    + getFullName() + " fehlgeschlagen: " + proxyerror.toString());
                  }
               }
               // Ende an Proxy Objekte weitergeben
       // Interessenten benachrichtigen.
        Object[] o = new Object[3];
        o[0] = paramOne;
        o[1] = compTwo;
        o[2] = paramTwo;
        server.TailorActionPerformed(this, "deleteParamBinding", o);

              return pb;
          }
       }
    } catch (RemoteException e) {
      server.message("error", 1, "deleteParameterBinding: " + e.toString());
    }
    return null;
  }
  
  /** Selbstmord. Und zwar als komplexe Komponente (rekursiv). 
      Lscht alle Ports (incl. Verbindungen) und Parameter (incl. Verbindungen).
      Fr alle Subkomponenten wird delete() aufgerufen und diese abgemeldet.
      Die Komponente wird nicht bei ihrem Vorgnger abgemeldet! Schlielich 
      mu jemand sie begraben.
  */
  public void delete() {
    Vector v = (Vector) params.clone();
    for (Enumeration e=v.elements(); e.hasMoreElements(); ) {
         deleteParam( (Param) e.nextElement()) ;
    }
    v = (Vector) ports.clone();
    for (Enumeration e=v.elements(); e.hasMoreElements(); ) {
         deletePort( (Port) e.nextElement())  ;
    }
    Hashtable subC = (Hashtable) subComponents.clone();
    IR_Component c = null;
    String cname = null;
    try {
       for (Enumeration e=subC.keys(); e.hasMoreElements(); ) { // Namen
         cname = (String) e.nextElement();
         c = (IR_Component) subC.get(cname);
         c.delete();
         this.removeSubComponent(cname);
       }
    } catch (RemoteException e) {
      server.message("error", 1, "Component.delete: " + e.toString());
    }
  }
  
  
  //
  //     *** For testing the existence
  //

  /** Test, ob diese Komponente eine implementierte Komponente ist.
      Also, ob sie eine FlexiBean darstellt. Getestet wird ob 
      Subkomponenten vorhanden sind.
  */
  public boolean isImplComp() {
    return isAtomic();
    //return !((this.subComponents.elements()).hasMoreElements());
  }

   /** Test, ob wir Rootkomponente sind. Getestet wird, ob wir bei jemand
       anderem als Subkomponente angemeldet sind.<br>
       Ist  predecessor == null?
   */
  public boolean isRoot() {
    return (this.predecessor == null);
  }
  
  /** Test, ob wir einen Port mit Namen name haben.
  */
  public boolean existsPort(String name) {
    if (getPort(name) == null) return false;
    else return true;
  }
  
  /** Test, ob wir einen Parameter mit Namen name haben.
  */
  public boolean existsParam (String name, String localName) {
    if (getParam(name, localName) == null) return false;
    else return true;
  }

  /** Test, ob wir eine Subkomponente mit Instanznamen subComp haben.
  */
  public boolean existsSubComponent(String subComp) {
    if (subComp == null) return true;
    if (this.subComponents.containsKey(subComp)) return true;
    else return false;
  }


 /**  Testet ob es  eine Portverbindung gibt.
      Es gibt nur Verbindungen zu einer Subkomponente oder zwischen zwei 
      unterschiedlichen Subkomponenten. Wenn ein Port zur abstrakten
      Komponente gehrt, ist compOne oder compTwo entsprechend "null" oder ".".
      @see Port
      @see PortBinding
 */
  public boolean existsPortBinding (String compOne, String portOne,
                                    String compTwo, String portTwo) {
      if (getPortBinding(compOne, portOne, compTwo, portTwo) == null) return false;
      return true;
  }
  
  /** Testet, ob der Port portName verbunden ist.
  */
  public boolean existsPortBindingWith (String portName) {
    Port a = getPort(portName);
    Enumeration pb=a.getBindings().elements();
    if (pb.hasMoreElements()) return true;
    else return false;
  }
  
  /** Testet, ob die Subkomponente componentName mit uns verbunden ist.
  */
  public boolean existsPortBindingWithComponent (String componentName) {
    PortBinding b = null;
    if (!existsSubComponent(componentName)) return false;
    try {
       for (Enumeration pb=this.portBindings.elements(); pb.hasMoreElements(); ) {
          b = (PortBinding) pb.nextElement();
          if ((b.getComponentOne().getName().equals(componentName)) ||
	          (b.getComponentTwo().getName().equals(componentName))) {
            	return true;
          }
       }
    } catch (RemoteException e) {
       server.message("error", 1, "Component.existsPortBindingWithComponent: " + e.toString());
    }
    return false;
  }


  
 /**  Testet ob es  eine Parameterverbindung gibt.
      Es gibt nur Verbindungen zu einer Subkomponente.
      @see Param
      @see ParamBinding
 */
  public boolean existsParamBinding (String paramOne, String compTwo, String paramTwo) {
      if (getParamBinding(paramOne, compTwo, paramTwo) == null) return false;
      return true;
  }
  
  /** Testet, ob der Parameter paramName verbunden ist.
  */
  public boolean existsParamBindingWith (String paramName) {
    Param a = getParam(paramName, null);
    Enumeration pb=a.getBindings().elements();
    if (pb.hasMoreElements()) return true;
    else return false;
  }
  
  /** Testet, ob die Subkomponente componentName ber Parameter mit uns verbunden ist.
  */
  public boolean existsParamBindingWithComponent (String componentName) {
    ParamBinding b = null;
    if (!existsSubComponent(componentName)) return false;
    try {
       for (Enumeration pb=this.paramBindings.elements(); pb.hasMoreElements(); ) {
          b = (ParamBinding) pb.nextElement();
          if ((b.getComponentOne().getName().equals(componentName)) ||
	          (b.getComponentTwo().getName().equals(componentName))) {
            	return true;
          }
       }
    } catch (RemoteException e) {
       server.message("error", 1, "Component.existsParamBindingWithComponent: " + e.toString());
    }
    return false;
  }





  //
  //     *** For changing   
  //
  /** ndert den Typname der komplexen Komponente
  */
  public void changeTypeName(String newTypeName) {
    typeName = newTypeName;
  }

  /** ndert den Portnamen. newName darf nicht schon existieren.
      @return Den neuen Namen; null, falls Port nicht existiert.
  */
  public String changePortName(String oldName, String newName) {
    Port a = getPort(oldName);
    if (a==null) return null;
    if (getPort(newName) != null) return a.getName();
    a.setName(newName);				   
    return a.getName();
  }
  
  /** ndert den Parameternamen eines lokalen Parameters. newName darf nicht schon existieren.
      @return Den neuen Namen; null, falls Parameter nicht existiert.
  */
  public String changeParamName (String oldName, String newName) {
    Param a = getParam(oldName, null);
    if (a==null) return null;
    if (getParam(newName, null) != null) return a.getName();
    a.setName(newName);
    return a.getName();
  }

  /** ndert den Instanznamen der Subkomponente, 
      die unter dem local_name oldName bei uns abgemeldet ist.  <br>
      Eine Subkomponente, die unter dem Instanznamen newName angemeldet ist,
      darf nicht schon existieren.
      @return Den neuen lokalen Namen; null, falls Subkomponente nicht existiert.
  */
  public String changeSubComponentName(String oldName, String newOne) {
    if (!this.subComponents.contains(oldName)) return null;
    if (this.subComponents.contains(newOne)) return oldName;
    IR_Component c = (IR_Component) this.subComponents.remove(oldName);
    this.subComponents.put(newOne, c);
    return newOne;
  }

   /** Zum Durchschleifen.
   */
   void message(String typ, int level, String message) {
      server.message(typ, level, message);
   }

  public String  getProxyDef()   {
    Port po;
    Param pa;
    PortBinding pob;
    ParamBinding pab;
    String sc;
    Enumeration e;

    String result = "component " + typeName;
    if (name == null || name.equals("")) result += " .";
    else              result += " " + name;         
    if (isAtomicComponent) result += " A;\n";
    else                   result += " N;\n";
    
    // PORTS:
    for ( e = ports.elements(); e.hasMoreElements(); ) {
          po = (Port) e.nextElement()  ;
          result += po.printShort() + ";\n";
    }
    //PARAMS
    for ( e = params.elements(); e.hasMoreElements(); ) {
          pa = (Param) e.nextElement()  ;
          if (!pa.isSubCompParam()) result += pa.printShort() + ";\n";
    }
    //SUBCOMPONENTS
    for ( e = subComponents.keys(); e.hasMoreElements(); ) {
          sc = (String) e.nextElement()  ;
          result += "s " + sc + " ;\n";
    }
    //PORTBINDINGS
    for ( e = portBindings.elements(); e.hasMoreElements(); ) {
          pob = (PortBinding) e.nextElement()  ;
          result += "bpo " + pob.printShort() + ";\n";
    }
    //PARAMBINDINGS
    for ( e = paramBindings.elements(); e.hasMoreElements(); ) {
          pab = (ParamBinding) e.nextElement()  ;
          result += "bpa " + pab.printShort() + ";\n";
    }
    result += "ende";
    
    return result;
  }

  /** Alle Verweise auf Proxy Objekte der Benutzer im Vektor users werden abgemeldet und bei der Komponente toComp angemeldet.
  */
  public void changeScope(Vector users, IR_Component toComp) {

     try {
         for (Enumeration e = proxies.keys(); e.hasMoreElements() ; ) {
             String userid = (String) e.nextElement();
             String us     = userid.substring(0, userid.indexOf(':'));
             String id     = userid.substring(userid.indexOf(':'));
             if (id.length()>1) id = id.substring(1);
             else id = "";
             if (users.contains(us)) {
                  IR_Proxy proxy = (IR_Proxy) proxies.get(userid);
                  removeIR_Proxy(us, id);
                  toComp.addIR_Proxy(us, id, proxy);
             }
         }   
         for (Enumeration e = subComponents.keys(); e.hasMoreElements(); ) {
              String local_name = (String) e.nextElement();
              IR_Component sc = (IR_Component) subComponents.get(local_name);
              IR_Component to_sc = toComp.getSubComponent(local_name);
              sc.changeScope(users, to_sc);
         }
    } catch (RemoteException ex) {
         server.message("error", 1 , "changeScope: " + ex + ". Dieser Fehler kam unerwartet. Die Meta Struktur kann nun inkonsistent sein.");
         return;
    }
    Object[] o = new Object[2];
    o[0] = users;
    o[1] = toComp;
    server.TailorActionPerformed(this, "changeScope", o);
  }
} // end class Component;


