// The "freEvolve System" is a platform supporting the reconfiguration of 
// component-based applications'
// Copyright (C) 1999  Ralph Hinken and Dr. Oliver Stiemerling
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Version: 1.0
// Date:    09.10.2001
//
// History:
//

package org.freevolve.server;

import java.util.*;
import java.io.*;
import java.rmi.RemoteException;
import java.rmi.server.*;
import java.lang.reflect.Method;

// bei addRBD -> verdrahten !!
// bei removeRBD -> entdrahten !!


/**   CLASS DistComponent
 *
 *    Referenziert von: Server          - Alle  verteileten Komponenten<br>
 *                      DistComponent   - Andere DistComponent, die sich ange-<br>
 *		                         bunden haben (wirkennen nur deren RBDs.)<br>
 *		       RBD             - Die eigenen Remote Binds<br>
 *<br>
 *	   <br>    
 *   Ein DistComp Objekt stellt dar:<br>
 *   1.) Eine Beschreibung fr eine lokale Komponente, die der Benutzer auf<br>
 *       seinem Host bald starten moechte, dann<br>
 *   2.) ein Proxy-Objekt, dass nach dem Aufbauen hier angemeldet werden muss!<br>
 *       Und wenn diese Komponente mit anderen entfernten Komponenten verbunden<br>
 *       werden soll<br>
 *   3.) eine verteilte Anwendung, mit verbundenen DistComponents und den<br>
 *       remote Binds (RBD) Diese werden hier angefordert oder erstellt und<br>
 *        hier werden dann auch die Proxies verdrahtet.<br>
 *<br>
 *    Ein DistComp-Objekt wird vom Server angelegt! Dort wird ueber<br>
 *    den Dateinamen ein DCAT-Beschreibungsfile ausgewaehlt und dem Konstruktor
 *    uebergeben<br>
 *    Wird ein Proxy-Objekt fr ein Client aufgebaut,  muss dieses Objekt hier
 *    angemeldet werden! Dies sollte die entsprechende Funktion auf dem Host
 *    sicherstellen, bei der  der Benutzer eine  Komponente starten will.<br>
 *  <br>
 *     Falls dieses Objekt eine Komponente auf dem Server darstellt, ist
 *     Username gleich "server" und DCAT-Name gleich Instanzname+"."+TypName der
       Komponente<br>
 *   <br>
 *     Jedes DistComp Objekt wird am Server angemeldet.<br>
 *<br>
 *    I.Allg. startet der Benutzer direkt seinen Teil der verteilten
 *    Anwendung, wie es im  DCAT-String uebergeben wird.
 *    <br><br>
 *    Ansonsten hat der Benutzer nur eine lokale Komponente mit
 *    Beschreibung auf dem Server. Wenn er sich an eine oder mehrere
 *    andere Komponenten anbinden moechte, meldet er hier die anderen
 *    DistComponent-Objekte an und verbindet sie.
 *<br><br>
 *    Da ein DistComponent-Objekt eine lokale Komponente bzw eine
 *    Serverkomponente darstellt, kann von dem Objekt aus im Hinblick
 *    auf "seine" verteilte Anwendung  nur Verbindungen von diesem Objekt
 *    zu anderen DistComponent-Objekten aufgebaut werden.
 *    <br><br>
 *    Am Server sind DistComponent Objekte angemeldet. Diese verwiesen auf (locale)
 *    Component Objekte (Komponentenbeschreibung, Klasse). Die DistComp.Objekte 
 *    stehen fr Exemplare dieser Comp.Objekte. Der Exemplarname (local Name)
 *    ist der Name, unter dem die Komponenten somit am Server angemeldet werden.
 *    Im DCAT-File wird genau dieses Exemplar ber Exemplarname und Typ (CAT-File-Name)
 *    eindeutig referenziert.
 *    <br><br>
 *    (((
 *    Zusaetzlich koennen Rbinds (Verbindungen innerhalb eines RBD-Objektes)
 *    ergaenzt oder entfernt werden. Die Verbindungsaenderungen werden dann
 *    direkt an die Proxy-Objekte weitergegeben, die dann auch real die
 *    Verbidnung anpasst.
 *    Ebenso koennen DistComponents und RBD's hinzugefgt und entfernt werden.
 *    )))
 *    <br><br>
 *    Wennn eine RBD Angemeldet wird, wird diese auch in der anderen
 *    DistComponente, die sie verbindet angemeldet. Dann weiss die
 *    DistComponente, die beendet wird, welche Remote  Binds (RBDs) sie
 *    abbauen muss.
 *<br><br>
 *    Eine RBD, die uns gehoert, hat als erste DistComponent uns eingetragen.
 *    Also alle RBDs in RemoteBinds mit DistCompOne == this gehoerten uns,
 *    alle anderen sind von anderen DistComponents, die uns anbinden.
 *    <br><br>
 *    DATEI-FORMAT DistComponent (DCAT)<br>
 *<br>
 *  COMPONENT CompInstanzName CompTypName LOCAL ";"            // Name der lokalen Komponente<br>
 * (COMPONENT CompInstanzName CompTypName SERVER ";" )*        // Verbundene K. auf dem Server <br>
 * (COMPONENT CompInstanzName CompTypName USER userName ";" )* // Verbundende K. bei anderen Usern<br>
 * (Rbind CompInstanzOne CompInstanzTwo RBD-Name)*        // Welche RBD verbindet welche Komp.<br>
 * // Eine der beiden Comp. in der letzten Zeile muss local sein!<br>
 *   
 */
public class DistComponent  extends UnicastRemoteObject implements IR_DistComponent {
  private String  name;        // Name of Distributed Component;
  private String  Id;          // Nummer;
  private String  userName;    // Benutzer;
  private String  localName;   // ExemplarName der lokalen Komponente, wie am Server angemeldet
  private Component component; // lokale Komponente der verteilten Anwendung;
  private IR_Proxy   local_proxy;       // Proxy der lokalen Komponente;
  private Vector  remoteBinds; // ALLE RBDs, die diese Komponente verbinden!;
  private Server  server;      // Evolve Server;
  private boolean firstBind;
  // Vector of IR_DistComponents, key is local name;
  private Vector  linkedDistComps;//verbundene verteilte Komponenten, die zur;
                                  //verteilten Anwendung des Benutzers gehoeren.;
  
  // Keywords 
  public final static String COMPONENT = "subcomponent";
  public final static String LOCAL     = "LOCAL";
  public final static String SERVER    = "SERVER";
  public final static String USER      = "USER";
  public final static String RBIND     = "BIND";

  // Nur zum deserialisiern . Noetig?;
  //  DistComponent() throws RemoteException {
    //  super();
    // }
  

  /** Erzeugt eine Verwaltungsobjekt fr Lokale Komponenten, verteilte Komponenten und Proxies.
      Zu bergeben sind Benutzername, Rechnername des Benutzers, Name des Verwaltungsobjektes
      und einen String, der ein <b>DCAT</b> File beinhaltet.<br>
      Dieser DCAT String mu angegeben werden und enthlt mindestens eine lokale 
      Komponente, also ein Objekt der Klasse Component. Das heit auch fr Anwendungen,
      die nur lokal sind mu es ein CAT File und ein DCAT File geben. Das DCAT File 
      enthlt nur einen  Verweis auf das CAT File.<br><br>

      In diesem Konstruktor wird die  lokale Komponente und alle Serverkomponenten
      angelegt und geprft, ob die bentigten USER-Komponenten vorhanden sind.<br><br>

      @see Component      
  */
  DistComponent(String user, String Id, String DcompName,
		String DCompString, Server s) throws RemoteException {
    super();
    userName       = user;
    name           = DcompName;
    server         = s;
    local_proxy    = null;
    component      = null;
    linkedDistComps = new Vector();
    remoteBinds    = new Vector();
    this.Id        = Id;
    firstBind      = false;  
    this.readDistComponent(DCompString, DcompName);   // u.a. component setzen!;
  }


  /** Erzeugt eine Verwaltungsobjekt fr eine Serverkomponente und Proxies.
      Zu bergeben sind Benutzername, Rechnername des Benutzers und der Name des Verwaltungsobjektes 
      und die lokale Serverkomponente.
<br><br>
      @see Component      
  */
  DistComponent(String user, String Id, String DcompName, String localName,
		Component ServerComp, Server s) throws RemoteException {
    super();
    userName       = user;
    name           = DcompName;
    this.localName = localName;
    server         = s;
    local_proxy    = null;
    component      = ServerComp;
    linkedDistComps = new Vector();
    remoteBinds    = new Vector();
    firstBind      = false;
    this.Id        = Id;
  }


  /** Jedes DistComponent Objekt erhlt eine eindeutige ID vom Server.
      Das ist wichtig, weil der Benutzer auch zwei Exemplare des selben
      DCAT Files anfordern kann.
  */
  public    String    getId()       { return Id;          }

  /** Der Name des Benutzers, der dieses Objekt angemeldet hat.
  */
  public    String    getUserName() { return userName;    }

  /** Der Exemplarname der  lokalen Komponente. 
      Damit ist die lokale Komponente am Server angemeldet.
  */
  public    String    getLocalName() { return localName;    }

  /** Die lokale Komponente, die immer vorhanden sein mu.
  */
  public    IR_Component getComponent(){ 
      return component;   
  }

  /** Fr den Scope Control
  */
  public Component local_getComponent(){ 
      return component;   
  }

  /** Fr den Scope Control
  */
  void setComponent(Component c) { component = c; }

  //  public    Proxy     getProxy()    { return remoteProxy; }

  /** Der Name des Objektes. Kann beliebig sein.
      Der Name mu der  Name des DCAT Files sein. Das DCAT File wird 
      vom Server angelegt.
  */
  public    String    getName()     { return name;        }

  /** Fr Scope Control
  */
  void setName(String name) { this.name = name;}
   /** Gibt das Proxy-Objekt der lokalen Komponente zurck.
   */
   public IR_Proxy  getProxy() { return local_proxy; }

   /** Meldet das Proxy-Objekt der lokalen Komponente an.
       Die Funktion wird vom Server oder Client Objekt aufgerufen.
   */
   public void      setProxy(IR_Proxy proxy) { this.local_proxy = proxy; }
   
   /** Alle verbundenen Objekte der Klasse DistComponent. 
       Der Schlssel ist der Exemplarname, unter dem die Komponente im DCAT 
       File angemeldet ist.
       @return Vector      von IR_DistComponent
       @see IR_DistComponent
   */
  public  Vector      getLinkedDistComps() { return linkedDistComps;}
  
  /** Testet, ob das Objekt IR_DistComponent in diesem Objekt benutzt wird.
      Entweder ob es dieses Objekt selber ist oder es mit diesem verbunden ist.
  */
  public boolean usedDistComponent(IR_DistComponent c) {
    if (c.equals(this)) return true;
    if (linkedDistComps.contains(c)) return true;
    return false;
  }

  /** Test, ob das bergebene Objekt mit unserem verbunden ist.
      Es reicht nicht, als verbunden angemeldet zu sein. Es wird nach einem
      RBD Objekt gesucht, da nicht unbedingt ein RBind Objekt entalten mu.<br>
      Das heit, wenn alle RBinds vom RBD Objekt abgemeldet sind, sollte das 
      RBD Objekt entfernt werden, weil sonst angenommen wird, da beide DistComponent
      Objekte noch verbunden sind.<br>
      Das sicherzustellen geschieht innerhalb der Klasse DistComponent.
      @see RBD
      @see RBind
  */
  public boolean isConnected(IR_DistComponent c) {
    RBD r = null;
    for(Enumeration e = remoteBinds.elements(); e.hasMoreElements(); ) {
      r = (RBD) e.nextElement();
      if (r.links(this, c)) return true;
    }
    return false;
  }

  /** Hat meine Komponente diesen Port p?; 
  */ 
  public boolean hasPort(String p) {
    return component.existsPort(p);
  }
  
  //
  // THE ULTIMATIVE I/O PART ***************************************************
  //
  
  private boolean readDistComponent(String apl, String compName) {
    RBind r = null;

    String          value = null;
    String          cName = null;
    String          c2Name= null;
    String          instanz = null;
    String          rbd   = null;
    DistComponent   dc    = null;
    IR_DistComponent rdc    = null;
    Component       c     = null;
    StringTokenizer token = new StringTokenizer(apl);
    if (token.hasMoreTokens()) {
      if (!token.nextToken().equalsIgnoreCase("s_component"))
         error("DCAT: Keine s_component gefunden in DCAT-Datei " + compName);
    }
    if (token.hasMoreTokens()) {
      if (!token.nextToken().equals(compName))
         error("DCAT: s_component " + compName + " nicht gefunden in DCAT_Datei " + compName);
    }
    if (token.hasMoreTokens()) {
      if (!token.nextToken().equalsIgnoreCase("{")) {
         error("DCAT: Keine ffnende Klammer { gefunden in DCAT_Datei " + compName);
         return false;
      }
    }
    int last = apl.lastIndexOf("};");
    int first = apl.indexOf('{') + 1;
    String rest = null;
    if (last == -1) {
        error("DCAT: Scliessende Klammer }; fr s_component nicht gefunden in DCAT_Datei " + compName);
        rest = apl.substring(first);
    } else {
        rest = apl.substring(first, last);
    }
     
 
    token = new StringTokenizer(rest, ";");    
    // First read all components
    while (token.hasMoreTokens()) {
      StringTokenizer st = new StringTokenizer(token.nextToken().trim());
      if (!st.hasMoreTokens()) continue;
      value = st.nextToken();
      // EINLESEN DER KOMPONENTEN DEKLARATIONEN *******************************;
//      if (value==null) continue;
      if (value.equalsIgnoreCase(COMPONENT)) {
      	if (st.hasMoreTokens()) instanz = st.nextToken();  // InstanzName;
      	else error("DCAT: Fehlender InstanzName in DCAT-Datei " + compName +" "+ st);
      	if (st.hasMoreTokens()) cName = st.nextToken();  // compTypName;
      	else error("DCAT: Fehlender TypName fr Instanz " + instanz + " in DCAT-Datei " + compName + " "+st);
      	if (st.hasMoreTokens()) value = st.nextToken();
      	else error("DCAT: Fehlendes Schlsselwort LOKAL|SERVER|USER nach Komponente "+instanz+" in DCAT-Datei " + compName + " "+st);
      	// EINLESEN LOKALE KOMPONENTEN DEKLARATION ***************************;
      	if (value.equalsIgnoreCase(LOCAL)) {
      	  c = null;
      	  c = server.giveComp(cName);
      	  if (c==null) {
      	    error("Konnte lokale Komponente " + instanz + " vom Typ " + cName + " nicht anlegen.");
      	    return false;
      	  }
           //c.setName(instanz);
           this.localName = instanz;
      	  this.component = c;
      	  server.message("dcomp", 2, "DistComponent " + getName() + " von " +
      			 getUserName() + " Nr " + getId() +
      			 " wurde lokale Komponente " + getLocalName() + " vom Typ " + cName +
      			 " hinzugefgt.");
      	} // END LOCAL;
	// EINLESEN EINER SERVER KOMPONENTEN DEKLARATION *********************;
	      else if (value.equalsIgnoreCase(SERVER)) {
      	  dc = null;
      	  dc = server.giveDistComp(instanz, cName, "server", null);
      	  if (dc==null) {
      	    error("Konnte Server-Komponente "+ instanz + " vom Typ " + cName + " nicht anlegen.");
      	  } else {
      	    this.linkedDistComps.addElement(dc);
             server.message("dcomp", 2, "DistComponent " + getName() + " von " +
	         	   getUserName() + " Nr " + getId() +
			         " wurde Server-Komponente " + instanz + " vom Typ " + cName +
      			   " hinzugefgt.");
      	  }
      	} // END SERVER;

      	// EINLESEN EINER USER KOMPONENTEN DEKLARATION **********************;
      	else if (value.equalsIgnoreCase(USER)) {
      	  dc = null;
      	  if (st.hasMoreTokens()) value = st.nextToken();  // UserName;
      	  else error("DCAT: Fehlender UserName in DCAT-Datei " + compName + " "+st);
      	  dc = server.giveDistComp(instanz, cName, value, null);  // wo kommt ID her
      	  if (dc==null) {
      	    error("Konnte User-Komponente " + cName + " von " +
      		  value + " nicht anlegen.");
      	  } else {
      	    linkedDistComps.addElement(c);
      	    server.message("dcomp", 2, "DistComponent " + getName() + " von " +
      			   getUserName() + " Nr " + getId() +
      			   " wurde User-Komponente " + dc.getName() + " vom Typ " + cName +
      			   " von " + value + " hinzugefgt.");
      	  }
         } else error("DCAT: wrong keyword " + value + " in " + st);
     	if (st.hasMoreTokens()) error("DCAT: In s_component sind (noch) keine Parameter erlaubt: " + compName);
	   // ENDE KOMPONENTEN DEKLARATION *************************************;
   
      
      // EINLESEN DER REMOTE BIND DEFINITIONEN ******************************;
      } else if (value.equalsIgnoreCase(RBIND)) {
         //server.message("dcomp", 6, "RBIND beginnt");
	      if (st.hasMoreTokens()) cName = st.nextToken();
      	else error("DCAT: Bei Verbindungsdefinition fehlt der erste Komponentenname  in DCAT-Datei " + compName + " "+st);
      	if (st.hasMoreTokens()) c2Name = st.nextToken();
      	else error("DCAT: Bei Verbindungsdefinition fehlt der zweite Komponentenname  in DCAT-Datei " + compName + " "+st);
      	if (st.hasMoreTokens()) rbd    = st.nextToken();
      	else error("DCAT: Bei Verbindungsdefinition fehlt der RBD-Datei-Name in DCAT-Datei " + compName + " "+st);
      	// Ist lokale Komponente schon da zum anbinden ? *************;
      	rdc = null;
         if (this.component == null) {
      	     error("DCAT: Erst Lokale Komponente definieren in DCAT-Datei " + compName + " "+st);
      	} // Hole die anzubindende Komponente *************************;
      	else {
            if (cName.equals(getLocalName())) {
      	       rdc = getOtherDistComp(c2Name);
         	} else if (c2Name.equals(getLocalName())) {
         	    rdc = getOtherDistComp(cName);
         	} else {
         	    error("DCAT: Eine Komponenten kann nur an die lokale Komponente gebunden werden / in DCAT-Datei " + compName + " "+st);
                error("c = "+getLocalName());
         	}
         }
      	if (rdc != null) {
      	    // RBD-Beschreibungsdatei laden *****************************;
      	    cName = server.getFile(rbd+".RBD");
      	    if (cName != null) {
      	      if (addRBD(rdc, rbd, cName)==null) {
      	      	error("DCAT: Kann RBD " + rbd + " nicht anlegen.");
	            } //else server.message("dcomp", 6, "RBD "+rbd+" gefunden.");
      	    } else
	               error("DCAT: Kann RBD " + rbd + " nicht lesen.");
      	} else error("DCAT: Finde an " + component.getName() +
                        " anzubindende Komponente nicht. (" + cName + " oder "+
                         c2Name+")");
       } else error("DCAT: wrong keyword " + value + " in " + st.toString());
          // ENDE REMOTE BIND DEFINITIONEN *****************************;
      } // naechstes Token;
      return true;
}

  public String writeDistComponent() {
    String        s = "";
    IR_DistComponent c = null;
    RBD           r = null;

   s+="s_component " + this.name + "  {\n";
    // Local Component ********************************************************;
    s += "    "+ COMPONENT + " " + this.getLocalName() + " " + component.getTypeName()+ " " + LOCAL + " ;\n";

    // linked Components  *****************************************************;
    for(Enumeration e = linkedDistComps.elements(); e.hasMoreElements(); ) {
      try {
      	c = (IR_DistComponent) e.nextElement();
      	if (c.getUserName().equals("server"))
      	  s += "    "+ COMPONENT + " " + c.getLocalName() + " " + c.getComponent().getTypeName() + " " + SERVER + " ;\n";
      	else
	        s += "    "+ COMPONENT + " " + c.getLocalName() + " " + c.getComponent().getTypeName()+ " " + USER +
	             " " + c.getUserName() + " ;\n";
      } catch (RemoteException d) {
	      server.message("error", 1, "DistComp.writeDistComponent(): RMI-Fehler: "+ d);
      }
    } // End for all linked Distributed Components ***************************;

    // Remote Binds **********************************************************;
    s+="\n\n";
    for(Enumeration e = remoteBinds.elements(); e.hasMoreElements(); ) {
      r = (RBD) e.nextElement();
      try {
         s += "    bind " + r.getDistComponentOne().getLocalName() + " " 
            + r.getDistComponentTwo().getLocalName() + " " + r.getName()+" ;\n";
      } catch (RemoteException d) {
	      server.message("error", 1, "DistComp.writeDistComponent(): RMI-Fehler: "+ d);
      }
    }  // End Remote Binds ***************************************************;
    s += "};\n";
    return s;
  }

  //
  // END OF THE ULTIMATIVE I/O PART  *******************************************
  //

  // erst anmelden, dann verbinden ...;
  /** Anmelden eines DistComponent Objektes mit dem man sich verbinden mchte.
      Erst mu das Objekt angemdeldet sein, damit man sich bei ihm anbinden kann.
      @param local_name Der Exemplarname, unter dem die Komponente am Server 
      angemeldet ist.
      @see #bind
  */
  public void addLinkedDistComponent(IR_DistComponent c) {
    if (c==null) return;
    try {
       if (!usedDistComponent(c)) {
         linkedDistComps.addElement(c);
         server.message("dcomp", 4, "DistComponent " + c.getName() +
	   	     " wurde angemeldet an " + this.name + ".");
       } else {
         server.message("error", 5, "DistComponent " + c.getName() +
		        " ist schon angemeldet an " + this.name + ".");
       }
    } catch (RemoteException e) {
         server.message("error", 1, "DistComponent.addLinkedDistComponent: " + 
               e.toString());
    }
  }
  
  // erst Verbindungen trennen, dann abmelden
  /** Abmelden eines DistComponent Objektes, mit dem man nicht mehr verbunden ist.
      Man mu erst alle Verbindungen lschen.
  */
  public void removeLinkedDistComponent(IR_DistComponent c) {
    if (c==null || c==this) return;
    try {
       if (!isConnected(c)) {
         linkedDistComps.removeElement(c);
         server.message("dcomp", 4, "DistComponent " + c.getName() +
   		     " wurde abgemeldet von " + this.name + ".");
          } else {
         server.message("error", 1, "DistComponent " + c.getName() +
	   	     " ist noch verbunden mit " + this.name + ".");
      }
    } catch (RemoteException e) {
         server.message("error", 1, "DistComponent.removeLnkedDistComponent: " + e.toString());
         return ;
    }
  }

  // Neue RBD, die moeglichst noch nicht als Datei existiert, hinzufuegen;
  /** Hinzufgen eines neuen RBD Objektes, das noch nicht als Datei existiert.
      Alle RBD Dateien werden gespeichert, wenn der Server das DistComponent
      Objekt persistent machen will. Eventuell bestehende RBD Dateien werden 
      berschrieben, der Name kann jedoch ber das RBD Objekt selbst gendert
      werden.
      @param otherDistComp Das andere IR_DistComponent Objekt, mit dem 
      sich verbunden werden soll.
      @param rbdName Der Name der RBD Datei, die dieses Verbindungsobjekt
      enthlt.
  */
  public IR_RBD addNewRBD(IR_DistComponent otherDistComp, String rbdName) {
    if (otherDistComp==null || rbdName==null) return null;
    try {
      if (getRBD(otherDistComp, rbdName) == null) { // new RBD ..;
      RBD r = new RBD(rbdName, this, otherDistComp, server);
      otherDistComp.addRBD(r);
      this.addRBD(r);     
           if (firstBind) {
             // Entfernte Verbindungen an Proxy Objekten durchfhren:
             IR_RBind rb = null;
             for(Enumeration e = r.getRBinds().elements(); e.hasMoreElements(); ) {
                 rb = (IR_RBind) e.nextElement();
                 bind(rb.getPortNameOne(), rb.getPortNameTwo(), r);
             }
           }
      
      server.message("rbind", 2, "Neue RBD " + rbdName + " zwischen " +
		     this.getName() + " und " +  otherDistComp.getName() +
		     "zugefgt.");
      return r;
      } else {
      	server.message("error", 1, "RBD "+rbdName+ " zwischen " + this.getName() +
		       " und " +  otherDistComp.getName() + " existiert bereits.");
      }
    } catch (RemoteException e) {
      server.message("error", 1, "DistComponent.addNewRBD: RMI-Fehler: "+e);
      return null;
    }
    return null;
  }


  // RBD, die als Datei existiert, hinzufuegen;
  /** Hinzufgen eines neuen RBD Objektes, das als Datei existiert.
      Die Datei wird als String mit bergeben.<br>
      Andere DistComponent Objekte, die dieses RBD Objekt instanziieren,
      bekommen eigene RBD Objekte, soda nderungen hier nicht bei anderen
      RBD Objekten wirksam werden. Es lt sich aber automatisch fr alle 
      betroffenen RBD Objekte nderungen durchfhren, ohne die Anwendung 
      stoppen zu mssen.<br><br>
      Alle RBD Dateien werden gespeichert, wenn der Server das DistComponent
      Objekt persistent machen will. Eventuell bestehende RBD Dateien werden 
      berschrieben, der Name kann jedoch ber das RBD Objekt selbst gendert
      werden.
      @param otherDistComp Das andere IR_DistComponent Objekt, mit dem 
      sich verbunden werden soll.
      @param rbdName Der Name der RBD Datei, die dieses Verbindungsobjekt
      enthlt.
  */
public IR_RBD addRBD(IR_DistComponent otherDistComp, String rbdName, String rbdString) {
   //server.message("dcomp", 6, "addRBD beginnt");
   if (otherDistComp==null || rbdName==null || rbdString==null) return null;
   try {
      RBD r = null;
      if (getRBD(otherDistComp, rbdName) == null) {
         try {
         	r = new RBD(rbdName, this, otherDistComp, rbdString, server);
         } catch (Exception e) {
         	server.message("error", 1, "RBD " + rbdName + " zwischen " +
		          this.getName() + " und " +  otherDistComp.getName() +
		          " konnte nicht angelegt werden." + e);
         	return null;
         }
    	   otherDistComp.addRBD(r);
         this.addRBD(r);
         // PROXIES VERBINDEN !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!;
           if (firstBind) {
             // Entfernte Verbindungen an Proxy Objekten durchfhren:
             IR_RBind rb = null;
             for(Enumeration e = r.getRBinds().elements(); e.hasMoreElements(); ) {
                 rb = (IR_RBind) e.nextElement();
                 bind(rb.getPortNameOne(), rb.getPortNameTwo(), r);
             }
           }
         server.message("rbind", 2, "RBD " +rbdName + " zwischen " +
	   	     this.getName() +  " und " +   otherDistComp.getName() +
		        " zugefgt.");
         return r;
       } else {
            server.message("error", 1, "RBD " + rbdName + " zwischen " +
		      this.getName() +  " und " +  otherDistComp.getName() +
		      " existiert bereits.");
       }
   } catch (RemoteException e) {
    	server.message("error", 1, "DistComponent.addRBD: RMI-Fehler: "+e);
     	return null;
   }

    return null;
 }

  
  // Zum Loeschen der Beschreibung der Verbidung (RBD) genuegt das Austragen aus
  // den Listen remoteBinds beider DistComponent;
  // WO WIRD TATSAECHLICH ENTBUNDEN? Hier, spezielle FKT ueber RBinds!!;

  
/** Alle RBDs entfernen, die uns gehoeren und die uns mit otherDistComp verbinden.
*/
public void removeRBD(IR_DistComponent otherDistComp) {
    if (otherDistComp==null) return;
    try {
       RBD r = null;
       Vector v = (Vector) remoteBinds.clone();
       for(Enumeration e = v.elements(); e.hasMoreElements(); ) {
         r = (RBD) e.nextElement();
         if (r.links(this, otherDistComp)) {
   	      // PROXIES entBINDEN !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!;
           // Entfernte Verbindungen an Proxy Objekten durchfhren:
           if (firstBind) {
               IR_RBind rb = null;
               for(Enumeration e2 = r.getRBinds().elements(); e2.hasMoreElements(); ) {
                   rb = (IR_RBind) e2.nextElement();
                   unbind(rb.getPortNameOne(), rb.getPortNameTwo(), r);
               }
          }
          otherDistComp.removeRBD(r);
          this.removeRBD(r);
  	      server.message("rbind", 3, "RBD " + r.getName() + " zwischen " +
		          this.getComponent().getName() + " und " +  otherDistComp.getName() +
		          " entfernt.");
         }
       } // END FOR all remoteBinds
     } catch (RemoteException e2) {
	         server.message("error", 1, "DistComponent.removeRBD: RMI-Fehler: " +e2);
   	      return;
	  }
  }

  /** Spezielle RBD entfernen, die uns mit otherDistComp verbindet und 
      den Namen rbdName trgt.
  */
  public void  removeRBD(IR_DistComponent otherDistComp, String rbdName) {
    if (otherDistComp==null || rbdName==null) return;
    try {
      IR_RBD r = getRBD(otherDistComp, rbdName);
      if (r!= null) {
      	// PROXIES entBINDEN !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!;
         // Entfernte Verbindungen an Proxy Objekten durchfhren:
           if (firstBind) {
               IR_RBind rb = null;
               for(Enumeration e2 = r.getRBinds().elements(); e2.hasMoreElements(); ) {
                   rb = (IR_RBind) e2.nextElement();
                   unbind(rb.getPortNameOne(), rb.getPortNameTwo(), r);
               }
          }
        	otherDistComp.removeRBD(r);
        	this.removeRBD(r);
        	server.message("rbind", 3, "RBD " + rbdName + " zwischen " + this.getName() +
		       " und " +   otherDistComp.getName() + " entfernt.");
      } else {
             server.message("error", 1, "Zu entfernende RBD " + rbdName + " zwischen " +
		       this.getName() + " und " + otherDistComp.getName() +
		       "existiert nicht.");
      }
    } catch (RemoteException e) {
      server.message("error", 1, "DistComponent.removeRBD: RMI-Fehler: "+e);
      return;
    }
  }

  /**   
   very dangerous low function!!
   nur von anderen remove-Funktionen aufzurufen!
   Keine Integritaetskontrolle
   Ist public, weil dies auf IR_DistComponent Objekte angewendet werden 
   mu und Interface keine protected Funktion beinhalten darf.
  */
  public void removeRBD(IR_RBD r) {
   if (r==null) return;
   //try {
      remoteBinds.removeElement(r);
   //} catch (RemoteException e2) {
	 //    server.message("error", 1, "DistComponent.removeRBD: RMI-Fehler: " +e2);
   //     return;
	//}
 }

  /** very dangerous low function!!
    nur von anderen remove-Funktionen aufzurufen!
    Keine Integritaetskontrolle
   Ist public, weil dies auf IR_DistComponent Objekte angewendet werden 
   mu und Interface keine protected Funktion beinhalten darf.

   */
public void addRBD(IR_RBD r) {
   if (r==null) return;
   try {
       // server.message("dcomp",5,"dcomp.addRBD(RBD) BEGINNT");
       if (r.links(this))  {
           remoteBinds.addElement(r);
           //server.message("dcomp",5,"dcomp.addRBD(R) meldet an!");
       }
   } catch (RemoteException e2) {
	     server.message("error", 1, "DistComponent.addRBD: RMI-Fehler: " +e2);
        return;
	}
}

  
  // port1 soll mir gehoeren...;
  /** Verbindet ber das IR_RBD Objekt die Ports port1 und port2, wobei port1 uns gehrt!
  */
  public void bind (String port1, String port2, IR_RBD r) {
    if (r==null) return;
    try {
      IR_DistComponent other = r.getDistComponentTwo(); // Sollte unsere RBD sein;
      if (other==null) return;
      // Typ-Integritaet wird in RBD ! Ports koennten dort auch getestet werden ...;
      if (r.owns(this) && this.hasPort(port1) && other.hasPort(port2))  {
          try {
            r.rbind(port1, port2);
          } catch (RBD_Exception e) {
              server.message("error", 2, "Verbindung unter " + r.getName() + "  " +
                  this.getName() + "." + port1 + " mit " + other.getName() +
                 "." + port2 + " konnte nicht erstellt werden.");
            return;
          }
          // PROXIES VERBINDEN !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!;
        try {
              // test auf null!
              IR_Proxy proxy1           = this.getProxy();
              IR_Proxy proxy2           = other.getProxy();
              // test auf null!
              IR_PortProxy pport1 = null;
              IR_PortProxy pport2 = null;
              RBind        rbind = null;

              //server.message("dcomp",5,"DistComp.bind:  "+ this.getName()+" " + other.getName()+
              //       " wird verbunden" );
             if (proxy1 == null) {
                   server.message("error",1,"DistComp.bind: CompOne "+ this.getName()+" hat kein Proxy" );
                  return ;
             }
             if (proxy2 == null) {
                   server.message("error",1,"DistComp.bind: CompTwO "+ other.getName()+" hat kein Proxy" );
                return ;
             }
                 pport1 = proxy1.getPort(port1);
                 if (pport1 == null) {
                     server.message("error",1,"DistComp.bind: Port " + port1 + 
                                    " existiert nicht im Proxy Objekt.");
                     return;
                 }
                 pport2 = proxy2.getPort(port2);
                 if (pport2 == null) {
                     server.message("error",1,"DistComp.bind: Port " + port2 + 
                                    " existiert nicht im Proxy Objekt.");
                      return;
                 }
                 bind(pport1, pport2) ;
        } catch (Exception ex) {
          server.message("error",1,"DistComp.bind: " +ex);
          return;
        }
        // ENDE PROXY VERBINDEN *************************************************

          server.message("rbind", 2, "Verbinde unter " + r.getName() + "  " +
                 this.getName() + "." + port1 + " mit " +  other.getName() +
                 "." + port2);
      } else {
            String error = new String("Verbindung unter " + r.getName() + "  " +
              this.getName()+ "." + port1 + " mit " +
               other.getName()+"."+port2 + " fehlgeschlagen:\n" );
          if       (!this.hasPort(port1)) error+="Port " +port1+ " existiert nicht.";
          else if (!other.hasPort(port2)) error += "Port "+port2+" existiert nicht.";
          else                   error += "RBD " + r.getName() + " existiert nicht.";
             server.message("error", 1, error);
      }
    } catch (RemoteException e) {
      server.message("error", 1, "DistComponent.bind: RMI-Fehler: "+e);
      return ;
    }
  }
    
    
  // port1 soll mir gehoeren...;
 /** Lscht Verbindung des IR_RBD Objekts zwischen den Ports port1 und port2, wobei port1 uns gehrt!
     Wenn das IR_RBD Objekt dann leer ist, wird es gelscht.
     Wahrscheinlich sollte man das extra aufrufen knnen (z.B. clean()), damit
     nicht ein neues RBD angelegt werden mu, wenn ein RBind gelscht wird...
  */
  public void unbind (String port1, String port2, IR_RBD r) {
    if (r==null) return;
    //server.message("dcomp", 3, "RBD UNBIND" +port1 + "  " + port2);
    try {
      IR_DistComponent other = r.getDistComponentTwo();     // Sollte unsere RBD sein;
      if (other==null) return;
      if (r.owns(this) &&  this.hasPort(port1) && other.hasPort(port2))  {
      	// PROXIES entBINDEN !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!;
                  // PROXIES VERBINDEN !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!;
        try {
              // test auf null!
              IR_Proxy proxy1           = this.getProxy();
              IR_Proxy proxy2           = other.getProxy();
              // test auf null!
              IR_PortProxy pport1 = null;
              IR_PortProxy pport2 = null;
              RBind        rbind = null;

              server.message("dcomp",5,"DistComp.unbind:  "+ this.getName()+" " + other.getName()+
                     " wird entbunden" );
             if (proxy1 == null) {
                   server.message("error",1,"DistComp.unbind: CompOne "+ this.getName()+" hat kein Proxy" );
                  return ;
             }
             if (proxy2 == null) {
                   server.message("error",1,"DistComp.unbind: CompTwO "+ other.getName()+" hat kein Proxy" );
                return ;
             }
                 pport1 = proxy1.getPort(port1);
                 if (pport1 == null) {
                     server.message("error",1,"DistComp.unbind: Port " + port1 + 
                                    " existiert nicht im Proxy Objekt.");
                     return;
                 }
                 pport2 = proxy2.getPort(port2);
                 if (pport2 == null) {
                     server.message("error",1,"DistComp.unbind: Port " + port2 + 
                                    " existiert nicht im Proxy Objekt.");
                      return;
                 }
                 unbind(pport1, pport2) ;
        } catch (Exception ex) {
          server.message("error",1,"DistComp.unbind: " +ex);
          return;
        }
        // ENDE PROXY ENTBINDEN *************************************************

        r.unbind(port1, port2);
      	//server.message("rbind", 2, "Entferne Verbindung unter " +r.getName()+"  " +
		    //   this.getName() + "." + port1+ " mit " +other.getName()+"." +
		    //   port2);
      } else {
      	String error = new String("Entfernen der Verbindung unter " +r.getName() +
				  "  " +this.getName() + "." + port1 + " mit " +
				  other.getName()+"."+port2 + " fehlgeschlagen." );
      	if (!this.hasPort(port1))   error += "Port " + port1 + " existiert nicht.";
      	else if (!other.hasPort(port2)) error+="Port " +port2+ " existiert nicht.";
      	else                   error += " RBD  " + r.getName() + " gehrt mir nicht.";
      	server.message("error", 1, error);
     }
     if (r.isEmpty()) {removeRBD(r); }

    } catch (RemoteException e) {
      server.message("error", 1, "DistComponent.bind: RMI-Fehler: "+e);
      return;
    }        
  }


  // Nur unsere RBDs, die wir erzeugt haben. Bei denen ist DistCompOne==this;
  /** Gibt alle RBD Objekte zurck, die wir angelegt haben. 
  */
  public Vector      getMyRBDs () {
    Vector v = new Vector();
    RBD r = null;
    for(Enumeration e = remoteBinds.elements(); e.hasMoreElements(); ) {
      r = (RBD) e.nextElement();
      if (r.owns(this)) v.addElement(r);
    }
    return v;
  }

  // Alle anderen RBDs. Die die wir nicht erzeugt haben, also andere
  // die sich bei uns angebunden haben, wie interessan..;
  /** Gibt alle RBD Objekte zurck, die hier angemeldet sind, aber von anderen DistComponent
  Objekten angelegt wurden - die sich bei uns angebunden haben.
  */
  public Vector      getOtherRBDs () {
    Vector v = new Vector();
    RBD r = null;
    for(Enumeration e = remoteBinds.elements(); e.hasMoreElements(); ) {
      r = (RBD) e.nextElement();
      if (!r.owns(this)) v.addElement(r);
    }
    return v;
  }

  /** Alle RBDs, die uns gehoeren und uns mit der DistComponent c verbinden;
  */
  public Vector      getRBDs (IR_DistComponent c) {
    if (c==null) return null;
    RBD r = null;
    Vector v = new Vector();
    for(Enumeration e = remoteBinds.elements(); e.hasMoreElements(); ) {
      r = (RBD) e.nextElement();
      if (r.links(this, c))  v.addElement(r);
    }
    return v;
  }


  // Eindeutige Indentifizierung von RBD !;
  // RBD  von this mit otherDistComp und Name rbdName zurueckgeben;
  // rbdName == null dann Name egal...;
  /** Gibt IR_RBD Objekt zurck, falls es existiert.
      @param otherDistComponent Das DistComponent Objekt, mit dem uns das RBD
      Objekt verbindet.
  */
  public IR_RBD getRBD(IR_DistComponent otherDistComp, String rbdName) {
    if (otherDistComp==null) return null;
    RBD r = null;
    for(Enumeration e = remoteBinds.elements(); e.hasMoreElements(); ) {
      r = (RBD) e.nextElement();
      if (r.links(this, otherDistComp)) {
	if ((rbdName==null) || (r.getName().equals(rbdName))) return r;
      }
    }
    return null;
  }

  /** Gibt IR_DistComponent mit Exemplarname <b> name </b> der lokalen Komponente zurck .
  */
 public IR_DistComponent getOtherDistComp(String name) {
   IR_DistComponent c = null;
   try {
      for(Enumeration e = linkedDistComps.elements(); e.hasMoreElements(); ) {
         c = (IR_DistComponent) e.nextElement();
         if (c.getLocalName().equals(name)) return c;
      }
   } catch (RemoteException e) {
      server.message("error", 1, "DistComponent.getOtherDistComp: RMI-Fehler: "+e);
      return null;
   }        
   return null;
}

   /** Lscht dieses Objekt.
      Soll alle Verbindungen zu anderen Objekten abmelden und abbauen und
      das Proxy Objekt beenden.
   */
  public void kill() {
  // noch proxies beim local Component abmelden! und wo sonst angemeldet-
    DistComponent c = null;
    Vector v = (Vector) linkedDistComps.clone();
    for(Enumeration e = v.elements(); e.hasMoreElements(); ) {
      c = (DistComponent) e.nextElement();
      removeRBD(c); // dort werden alle RBDs this mit c abgebaut, auch real;
    }
    try {
       if (getProxy() == null) {
           server.message("error", 1, "DistComponent.kill: no proxy, sorry");          
       } else {
          getProxy().kill();
       }
        // Proxy.hide();
    } catch (Exception ex) {
        server.message("error", 1, "DistComponent.kill: RMI-Fehler: "+ex);
    }
  } // END KILL

  void error(String e) {server.message("error", 1, e);}

 public boolean bindAll() {
    RBD r = null;
    //server.message("dcomp",5,"dcomp:IN BINDALL !!!!!");
    boolean result = true;
    for(Enumeration e = remoteBinds.elements(); e.hasMoreElements(); ) {
         //server.message("dcomp",5,"dcomp:IN BINDALL-next RBD");
         r = (RBD) e.nextElement();
         if (!bindAll(r)) result = false;
    }
    return result;
 }
  
 public boolean unbindAll() {
    RBD r = null;
    //server.message("dcomp",5,"dcomp:IN UNBINDALL !!!!!");
    boolean result = true;
    for(Enumeration e = remoteBinds.elements(); e.hasMoreElements(); ) {
         //server.message("dcomp",5,"dcomp:IN UNBINDALL-next RBD");
         r = (RBD) e.nextElement();
         if (!unbindAll(r)) result = false;
    }
    return result;
 }

  private boolean  bindAll (RBD r) {
    boolean result = true;
    try {
        IR_DistComponent distCompOne = r.getDistComponentOne();
        IR_DistComponent distCompTwo = r.getDistComponentTwo();
        // test auf null!
        IR_Proxy proxy1           = distCompOne.getProxy();
        IR_Proxy proxy2           = distCompTwo.getProxy();
        // test auf null!
        IR_PortProxy port1 = null;
        IR_PortProxy port2 = null;
        RBind        rbind = null;

        //server.message("dcomp",5,"DistComp.bindAll:  "+ distCompOne.getName()+" " +distCompTwo.getName()+
//               " wird verbunden" );
       if (distCompOne.getProxy()==null) {
             server.message("error",1,"DistComp.bindAll: CompOne "+ distCompOne.getName()+" hat kein Proxy" );
            return false;
       }
       if (distCompTwo.getProxy()==null) {
             server.message("error",1,"DistComp.bindAll: CompTwO "+ distCompTwo.getName()+" hat kein Proxy" );
          return false;
       }
       for (Enumeration e = r.getRBinds().elements() ; e.hasMoreElements() ; ) {
           rbind = (RBind) e.nextElement();
           port1 = proxy1.getPort(rbind.getPortNameOne());
           if (port1 == null) {
               server.message("error",1,"DistComp.bindAll: Port " + rbind.getPortNameOne() + 
                              " existiert nicht im Proxy Objekt.");
               continue;
           }
           port2 = proxy2.getPort(rbind.getPortNameTwo());
           if (port2 == null) {
               server.message("error",1,"DistComp.bindAll: Port " + rbind.getPortNameOne() + 
                              " existiert nicht im Proxy Objekt.");
               continue;
           }
           if (bind(port1, port2) == false) result = false;
           //unbind(port1.getName(), port2.getName(), r);
           //bind(port1.getName(), port2.getName(), r);
        }
    } catch (Exception ex) {
      server.message("error",1,"DistComp.bindAll: " +ex);
      result = false;
    }
    firstBind = true;
    return result;
  }

  private boolean  unbindAll (RBD r) {
    boolean result = true;
    try {
        IR_DistComponent distCompOne = r.getDistComponentOne();
        IR_DistComponent distCompTwo = r.getDistComponentTwo();
        // test auf null!
        IR_Proxy proxy1           = distCompOne.getProxy();
        IR_Proxy proxy2           = distCompTwo.getProxy();
        // test auf null!
        IR_PortProxy port1 = null;
        IR_PortProxy port2 = null;
        RBind        rbind = null;

        //server.message("dcomp",5,"DistComp.unbindAll:  "+ distCompOne.getName()+" " +distCompTwo.getName()+
        //       " wird entbunden" );
       if (distCompOne.getProxy()==null) {
             server.message("error",1,"DistComp.unbindAll: CompOne "+ distCompOne.getName()+" hat kein Proxy" );
            return false;
       }
       if (distCompTwo.getProxy()==null) {
             server.message("error",1,"DistComp.unbindAll: CompTwO "+ distCompTwo.getName()+" hat kein Proxy" );
          return false;
       }
       for (Enumeration e = r.getRBinds().elements() ; e.hasMoreElements() ; ) {
           rbind = (RBind) e.nextElement();
           port1 = proxy1.getPort(rbind.getPortNameOne());
           if (port1 == null) {
               server.message("error",1,"DistComp.unbindAll: Port " + rbind.getPortNameOne() + 
                              " existiert nicht im Proxy Objekt.");
               continue;
           }
           port2 = proxy2.getPort(rbind.getPortNameTwo());
           if (port2 == null) {
               server.message("error",1,"DistComp.unbindAll: Port " + rbind.getPortNameOne() + 
                              " existiert nicht im Proxy Objekt.");
               continue;
           }
           if (unbind(port1, port2) == false) result = false;
        }
    } catch (Exception ex) {
      server.message("error",1,"DistComp.unbindAll: " +ex);
      result = false;
    }
    return result;
  } // end Unbind
   

  private boolean bind(IR_PortProxy port1, IR_PortProxy port2) {
      try {
         //server.message("proxy", 4, "Verbinde remote: " + port1.toTxt() + "\nmit " + port2.toTxt());
         port1.addHorizontal(port2);
         port2.addHorizontal(port1);
         // von beidden nach unten FlexiBeansPorts sammeln
         Vector ports1 = port1.getLeafPorts();
         Vector ports2 = port2.getLeafPorts();
         /*
         for (Enumeration a=ports1.elements(); a.hasMoreElements(); ) {
            port1 = (IR_PortProxy) a.nextElement();
            //server.message("proxy",4,"PREMOTEH: " + port1.toTxt());
         }
         for (Enumeration a=ports2.elements(); a.hasMoreElements(); ) {
            port2 = (IR_PortProxy) a.nextElement();
            server.message("proxy",4,"PREMOTEH: " + port2.toTxt());
         }
         */
         return bindAllPorts(ports1, ports2);
      } catch (Exception ex) {
         server.message("error",1,"DistComp.bind: " +ex);
         return false;
      }
  }

  private boolean unbind(IR_PortProxy port1, IR_PortProxy port2) {
      try {
         //server.message("proxy", 4, "Entbinde remote: " + port1.toTxt() + "\nmit " + port2.toTxt());
         port1.removeHorizontal(port2);
         port2.removeHorizontal(port1);
         // von beidden nach unten FlexiBeansPorts sammeln
         Vector ports1 = port1.getLeafPorts();
         Vector ports2 = port2.getLeafPorts();
         /*
         for (Enumeration a=ports1.elements(); a.hasMoreElements(); ) {
            port1 = (IR_PortProxy) a.nextElement();
            server.message("proxy",4,"PREMOTEH: " + port1.toTxt());
         }
         for (Enumeration a=ports2.elements(); a.hasMoreElements(); ) {
            port2 = (IR_PortProxy) a.nextElement();
            server.message("proxy",4,"PREMOTEH: " + port2.toTxt());
         }
         */
         return unbindAllPorts(ports1, ports2);
      } catch (Exception ex) {
         server.message("error",1,"DistComp.unbind: " +ex);
         return false;
      }
  }

  /** Verbindet alle Ports aus ports1 mit allen aus ports2.
    Dazu mssen alle Ports von selben Typ sein, alle in ports1 dieselbe Polaritt haben 
    und alle in ports2 die entgegengesetzte Polaritt.<br>
    In ports1 und ports2 sind nur Ports von FlexiBean Proxy Objekten enthalten!
  */
public boolean bindAllPorts(Vector ports1, Vector ports2) {

    Enumeration e1      = ports1.elements();
    Enumeration e2      = ports2.elements();
    Vector help         = null;
    boolean     isEvent = false;
    Object      o       = null;
    Method      get     = null;
    Method      add     = null;
    try {
       if (!e1.hasMoreElements()) return true;
       if (!e2.hasMoreElements()) return true;
       IR_PortProxy pport1    = (IR_PortProxy) e1.nextElement();
       IR_PortProxy pport2    = null;
       //server.message("proxy",4,"Bind All Remote geht weiter");
       if (pport1.getPolarity().equals("required")) {
           help = ports1;
           ports1 = ports2;
           ports2 = help;
           e1=ports1.elements();
           e2=ports2.elements();
       }
       // Nun sind die Ports in e1 Provider !!!
       isEvent = pport1.isEvent();
       // Nun wissen wir alles...
       if (isEvent) {
           for (e1 = ports1.elements() ; e1.hasMoreElements() ; ) {
               pport1 = (IR_PortProxy) e1.nextElement();  // Provider: Eventsource
               //add    = pport1.getAddMethod();         // addEventListener
               for (e2 = ports2.elements() ; e2.hasMoreElements() ; ) {
                   pport2 = (IR_PortProxy) e2.nextElement(); // Consumer: EventListener
                   server.message("proxy", 4, "Verbinde (Event): " + pport1.toTxt() + 
                                    "\nmit " + pport2.toTxt());
                   //get    = pport2.getGetMethod();        // getEventListener
                   // Und, jetzt ....
                   try {
                       //o[0] = get.invoke(pport2.getOwner().getObject(), null);
                       o = pport2.invokeGet();
                       pport1.invokeAdd(o);
                       //server.message("proxy",4,"habe Fkt aufgerufen !!");
                   } catch (Exception x) {
                       server.message("error",1, "DistComp.bindAllPorts: Event: " + x.toString() + 
                                      " beim Verbinden von " + pport2.toTxt() + "\n mit " +pport1.toTxt());
                   }
               } // end for e2
           } // end for e1
        } else {
           for (e1 = ports1.elements() ; e1.hasMoreElements() ; ) {
               pport1 = (IR_PortProxy) e1.nextElement();  // Provider: SharedObject owner
               //get    = pport1.getGetMethod();         // getSharedObject
               for (e2 = ports2.elements() ; e2.hasMoreElements() ; ) {
                   pport2 = (IR_PortProxy) e2.nextElement(); // Consumer: SharedObject user2
                   server.message("proxy", 4, "Verbinde (Shared Object): " + pport2.toTxt() 
                                    + "\nmit" + pport1.toTxt());
                   //add    = pport2.getAddMethod();        // setSharedObject
                   // Und, jetzt ....
                   try {
                       o = pport1.invokeGet();
                       //server.message("proxy",4,"habe object");
                       //if (o instanceof Lists.RemoteList) server.message("proxy",4,"instance RemoteList");             
                       pport2.invokeAdd(o);
                       //server.message("proxy",4,"habe Fkt aufgerufen !!");
                   } catch (Exception x) {
                       server.message("error",1,"DistComp.bindAllPorts: Event: " + x.toString()+ 
                                      " beim Verbinden von " + pport1.toTxt() + "\n mit " +pport2.toTxt());
                   }
               } // end for e2
             } // end for e1
         } // end else (!isEvent)
         return true;
     } catch (Exception ex) {
         server.message("error",1,"RBD.bindAllPorts: " +ex);
         return false;
     }
  } // end bindAll 


 /** Entbindet alle Ports aus ports1 von allen aus ports2.
    Dazu mssen alle Ports von selben Typ sein, alle in ports1 dieselbe Polaritt haben 
    und alle in ports2 die entgegengesetzte Polaritt.<br>
    In ports1 und ports2 sind nur Ports von FlexiBean Proxy Objekten enthalten!
  */
public boolean unbindAllPorts(Vector ports1, Vector ports2) {

    Enumeration e1      = ports1.elements();
    Enumeration e2      = ports2.elements();
    Vector help         = null;
    boolean     isEvent = false;
    Object      o       = null;
    Method      get     = null;
    Method      remove  = null;
    try {
       if (!e1.hasMoreElements()) return true;
       if (!e2.hasMoreElements()) return true;
       IR_PortProxy pport1    = (IR_PortProxy) e1.nextElement();
       IR_PortProxy pport2    = null;
       //server.message("proxy",4,"Unbind all Remote geht weiter");
       if (pport1.getPolarity().equals("required")) {
           help = ports1;
           ports1 = ports2;
           ports2 = help;
//*           e1=ports1.elements();
//*           e2=ports2.elements();
       }
       // Nun sind die Ports in e1 Provider !!!
       isEvent = pport1.isEvent();
       // Nun wissen wir alles...
       if (isEvent) {
           for (e1 = ports1.elements() ; e1.hasMoreElements() ; ) {
               pport1 = (IR_PortProxy) e1.nextElement();  // Provider: Eventsource
               for (e2 = ports2.elements() ; e2.hasMoreElements() ; ) {
                   pport2 = (IR_PortProxy) e2.nextElement(); // Consumer: EventListener
                   server.message("proxy", 4, "Entbinde (Event): " + pport1.toTxt() + 
                                    "\nvon " + pport2.toTxt());
                   // Und, jetzt ....
                   try {
                       o = pport2.invokeGet();
                       pport1.invokeRemove(o);
                       //server.message("proxy",4,"habe Fkt aufgerufen !!");
                   } catch (Exception x) {
                       server.message("error",1, "DistComp.unbindAllPorts: Event: " + x.toString());
                   }
               } // end for e2
           } // end for e1
        } else {
           for (e1 = ports1.elements() ; e1.hasMoreElements() ; ) {
               pport1 = (IR_PortProxy) e1.nextElement();  // Provider: SharedObject owner
               for (e2 = ports2.elements() ; e2.hasMoreElements() ; ) {
                   pport2 = (IR_PortProxy) e2.nextElement(); // Consumer: SharedObject user2
                   server.message("proxy", 4, "Entbinde (Shared Object): " + pport2.toTxt() 
                                    + "\nvon" + pport1.toTxt());
                   // Und, jetzt ....
                   try {
                       //o = pport1.invokeGet();
                       //server.message("proxy",4,"habe object");
                       //if (o instanceof Lists.RemoteList) server.message("proxy",4,"instance RemoteList");             
                       pport2.invokeRemove(null);
                       //server.message("proxy",4,"habe Fkt aufgerufen !!");
                   } catch (Exception x) {
                       server.message("error",1,"DistComp.unbindAllPorts: Event: " + x.toString());
                   }
               } // end for e2
             } // end for e1
         } // end else (!isEvent)
         return true;
     } catch (Exception ex) {
         server.message("error",1,"RBD.unbindAllPorts: " +ex);
         return false;
     }
  } // end unbindAll 

} // END CLASS DistComponent ;

