// The "freEvolve System" is a platform supporting the reconfiguration of 
// component-based applications'
// Copyright (C) 1999  Ralph Hinken and Dr. Oliver Stiemerling
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Version: 1.0
// Date:    09.10.2001
//
// History:
//

package org.freevolve.server;

import java.util.*;
import java.io.*;
import java.rmi.*;

/**
  Remote Interface, da nur von der Klasse Component instanziiert wird.
   
  @see Component
  */
public interface IR_Component extends Remote  {

  public boolean save(String directory, String name) throws RemoteException, IOException, MyFileException;
  public String toText() throws RemoteException;

  //
  //     *** Access for the datastructure
  //
  public String      getTypeName()                 throws RemoteException;
  public String      getName()                     throws RemoteException;
  public void        setName(String name)          throws RemoteException;
  public String      getFullName()                 throws RemoteException;
  public IR_Server   getServer()                   throws RemoteException;
  public IR_Component   getPredecessor ()          throws RemoteException;
  public void setPredecessor(IR_Component comp)    throws RemoteException;
  public String  getProxyDef()                 throws RemoteException;
  public boolean isAtomic()                        throws RemoteException;
  public void    setAtomic(boolean isAtomic)       throws RemoteException;

  public Vector      getPortNames   ()             throws RemoteException;
  public Vector      getPorts       ()             throws RemoteException;
  public IR_Port     getIR_Port     (String name) throws RemoteException;  
  public String      getPortType    (String name)  throws RemoteException;
  public String      getPortPolarity(String name)  throws RemoteException;

  public Vector      getParams      ()             throws RemoteException;
  public Vector      getParamNames  ()             throws RemoteException;
  public IR_Param    getIR_Param    (String name, String localName)  throws RemoteException;
  public String      getParamType   (String name, String localName)  throws RemoteException;
  public String      getParamSValue (String name, String localName)  throws RemoteException;
  public double      getParamNValue (String name, String localName)  throws RemoteException;
  public char        getParamCValue (String name, String localName)  throws RemoteException;
  public Object      getParamValue  (String name, String localName)  throws RemoteException;
  public boolean     isSubCompParam (IR_Param p)   throws RemoteException;
  
  public int         getNumberOfSubComponents()    throws RemoteException;
  public Hashtable   getSubComponents()            throws RemoteException;
  public IR_Component getSubComponent(String instanceName)  throws RemoteException;

  // Bindungen zu Subcomponenten und zwischen Subcomponenten
  // Um Bindungen zu 'Nachbar'-Componenten zu erhalten, diese Funktion fuer den Vorgaenger aufrufen 
  public Vector      getPortBindingsString()         throws RemoteException;
  public Vector      getPortBindings()               throws RemoteException;
  public IR_PortBinding getPortBinding (String compOne, String portOne, String compTwo, String portTwo) throws RemoteException; 
  public Vector      getConnectedPorts(String port)  throws RemoteException;
  public Vector      getConnectedPortsString(String port)  throws RemoteException;
  public Vector      getParamBindingsString()        throws RemoteException;
  public Vector      getParamBindings()              throws RemoteException;
  public IR_ParamBinding getParamBinding (String paramOne, String compTwo, String paramTwo) throws RemoteException;
  public Vector      getConnectedParamsString(String param) throws RemoteException;  
  public Vector      getConnectedParams(String param) throws RemoteException; 
  // 
  //     *** For creation of the datastructure
  //
  public IR_Port  addPort  (String name, String type, String polarity, String protocol)  throws RemoteException;  
  public IR_Param addParam (String name, String type, String local_name)  throws RemoteException;  
  public boolean  setParam (String name, String localName, String svalue, double nvalue, char cvalue)  throws RemoteException;  
  public void   unSetParam (String name)  throws RemoteException;  
  public void setSubCompParam (String name, Object value) throws RemoteException;  

  public IR_Component    addSubComponent(IR_Component subComponent, String instanceName)  throws RemoteException;  
  public IR_PortBinding  addPortBinding (String compOne, String portOne, String compTwo, String portTwo)  throws RemoteException;  
  public IR_ParamBinding addParamBinding(String paramOne, String compTwo, String paramTwo)                throws RemoteException;  
  public void addIR_Proxy   (String user, String ID, IR_Proxy proxy)       throws RemoteException;  
  public void removeIR_Proxy( String user, String ID)       throws RemoteException;  
  public IR_Component copy   (String newTypeName)  throws RemoteException;  

  public void deletePort     (IR_Port p)    throws RemoteException; 
  public void deletePort     (String name)  throws RemoteException;  
  public void deleteParam    (IR_Param p)     throws RemoteException; 
  public void deleteParam    (String name, String localName)  throws RemoteException;  
  public IR_Component removeSubComponent(String local_name)  throws RemoteException; 
  
  //public void deletePortBinding(PortBinding pb) throws RemoteException;  
  public IR_PortBinding deletePortBinding(IR_PortBinding pb) throws RemoteException; 
  public IR_PortBinding deletePortBinding (String compOne, String portOne, String compTwo, String portTwo)  throws RemoteException;
  public void deletePortBindingsWithPort (String aPort)  throws RemoteException;  
  public void deletePortBindingsWithComponent (String comp)  throws RemoteException;  

//  public void deleteParamBinding(ParamBinding pb) throws RemoteException;  
  public IR_ParamBinding deleteParamBinding(IR_ParamBinding pb) throws RemoteException; 
  public IR_ParamBinding deleteParamBinding (String paramOne, String compTwo, String paramTwo)  throws RemoteException; 
  public void deleteParamBindingsWithParam (String aParam)  throws RemoteException;  
  public void deleteParamBindingsWithComponent (String local_name)  throws RemoteException;  
  
  public void delete()  throws RemoteException;  
	  
  
  //
  //     *** For testing the existence
  //

  public boolean isImplComp()  throws RemoteException;  
  public boolean isRoot()  throws RemoteException;  
  public boolean existsPort(String name)  throws RemoteException;  
  public boolean existsParam (String name, String localName)  throws RemoteException;  
  public boolean existsSubComponent(String subComp)  throws RemoteException;  
  
  public boolean existsPortBinding (String compOne, String portOne, String compTwo, String portTwo)  throws RemoteException;  
  public boolean existsPortBindingWith (String portName)  throws RemoteException;  
  public boolean existsPortBindingWithComponent (String componentName)  throws RemoteException;  
  public boolean existsParamBinding (String paramOne, String compTwo, String paramTwo)  throws RemoteException;  
  public boolean existsParamBindingWith (String paramName)  throws RemoteException;  
  public boolean existsParamBindingWithComponent (String componentName)  throws RemoteException;  
  
  //
  //     *** For changing   
  //
  public void    changeTypeName(String newTypeName) throws RemoteException;
  public String  changePortName(String oldName, String newName) throws RemoteException;
  public String  changeParamName (String oldName, String newName) throws RemoteException;
  public String  changeSubComponentName(String oldInstanceName, String newOne) throws RemoteException;

  public void changeScope(Vector users, IR_Component toComp) throws RemoteException;
} // end Interface RemoteComponent;



