// The "freEvolve System" is a platform supporting the reconfiguration of 
// component-based applications'
// Copyright (C) 1999  Ralph Hinken and Dr. Oliver Stiemerling
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Version: 1.0
// Date:    09.10.2001
//
// History:
//

package org.freevolve.server;
import java.io.*;
import java.rmi.*;
import java.rmi.server.*;


/** Komfortable Klasse um Verzeichnisse zu lesen und sich Dateien in einem langen String geben zu lassen.
*/
public class MyFile extends UnicastRemoteObject implements IR_File {

  public MyFile() throws java.rmi.RemoteException {
    super();
  }

/** Gibt ein String-Array zurck mit allen Unterverzeichnissen im Verzeichnis path.
   @param path Das Verzeichnis.
   @return Ein String-Array mit allen Unterverzeichnissen im bergebenen Pfad.
*/
  public String[] getDir(String path) throws MyFileException {
    File f = new File(path);
    if (!f.isDirectory()) throw new MyFileException("getDir: there is no directory " + path);
    String[] inhalt = f.list();
    int number=0;
    if (inhalt != null) {
    for(int i=0; i < inhalt.length ; i++) {
	   File fj = new File(path, inhalt[i]);
	   if (fj.isFile()) {
	    inhalt[i]=null;
	   } else {
	   number++;
	 }
      }
      String[] result = new String[number];
      number = 0;
      for(int i=0; i < inhalt.length ; i++) {
	if (inhalt[i] != null) {
	  result[number++] = inhalt[i];
	}
      }
      return result;
    }
    return null;
  }

/** Gibt ein String-Array zurck mit allen Dateinamen im Verzeichnis path.
   @param path Das Verzeichnis.
   @param extension Endung der auzuwhlenden Dateien oder *.
   @return Ein String-Array mit allen Dateinamen im bergebenen Pfad.
*/
  public String[] getFiles(String path, String extension) throws MyFileException {
    File f = new File(path);
    if (!f.isDirectory())  throw new MyFileException("getFiles: there is no directory " + path);
    if (extension.length() == 0 || extension == null) throw new MyFileException("getFiles: i need an extension or *");
    
    String[] inhalt = f.list();
    int number=0;
    if (inhalt != null) {
      for(int i=0; i < inhalt.length ; i++) {
	File fj = new File(path, inhalt[i]);
	if (fj.isFile() && (extension.compareTo("*")==0 || inhalt[i].endsWith(extension))) {
	  number++;
	} else {
	  inhalt[i]=null;
	}
      }
      String[] result = new String[number];
      number = 0;
      for(int i=0; i < inhalt.length ; i++) {
	if (inhalt[i] != null) {
	  result[number++] = inhalt[i];
	}
      }
      return result;
    }
    return null;
  }

/** Gibt einen String zurck der die Datei enthlt.
*/
  public String getFile(String directory, String fileName)  throws MyFileException, IOException {
    String result = null;
    File f = new File(directory, fileName); 
    if (!f.isFile()) throw new MyFileException("MyFile.getFile: there is no such file " + f.getAbsolutePath());
    FileReader in = null;
    // Read and display the file contents.  Since we're reading text, we
    // use a FileReader instead of a FileInputStream.
    try {
      in = new FileReader(f);            // Create a char stream to read  it
      int size = (int) f.length();       // Check file size
      char[] data = new char[size];      // Allocate an array big enough for it
      int chars_read = 0;                // How many chars read so far?
      while(chars_read < size) {           // Loop until we've read it all
        chars_read += in.read(data, chars_read, size-chars_read);
      }
      result = new String(data);
      //System.out.println(result);
    }
    // Display messages if something goes wrong
    catch (IOException e) { 
      System.out.println(e.getClass().getName() + ": " + e.getMessage());
    }
    // Always be sure to close the input stream!
    finally { try { if (in != null) in.close(); } catch (IOException e) {} }
    return result;
  }
  
  /** Schreibt einen String in eine Datei.
     Sehr wahrscheinlich berschreibt die Methode bestehende Dateien.
     @param directory Das Verzeichnis, in das die Datei geschrieben werden soll.
     @param fileName Der Dateiname der anzulegenden oder zu berschreibenden Datei.
     @param inhalt Der String, der in die Datei geschrieben werden soll.
  */
  public void writeFile(String directory, String fileName, String inhalt )  throws MyFileException, IOException {
    String result = null;
    File f = new File(directory, fileName); 
    File test = new File(directory);
    if (!test.isDirectory()) throw new MyFileException("MyFile.getFile: there is no such directory " + test.getAbsolutePath());
    FileWriter out = null;
    try {
      out = new FileWriter(f);            // Create a char stream to read  it
      int size = (int) inhalt.length();       // Check file size
      out.write(inhalt, 0, size);
      out.flush();
    }
    // Display messages if something goes wrong
    catch (IOException e) { 
      System.out.println(e.getClass().getName() + ": " + e.getMessage());
    }
    // Always be sure to close the input stream!
    finally { try { if (out != null) out.close(); } catch (IOException e) {} }
  }
  
   /** Schreibt einen String in eine Datei und erzeugt eine Sicherheitskopie.
     Diese Methode berschreibt bestehende Dateien.
     @param directory Das Verzeichnis, in das die Datei geschrieben werden soll.
     @param fileName Der Dateiname der anzulegenden oder zu berschreibenden Datei.
     @param ext Die Endung, die an den Filenamen gehngt wird, um eine Sicherheitskopie zu erstellen.
     @param inhalt Der String, der in die Datei geschrieben werden soll.
  */
  public synchronized void writeFileWithBak(String directory, String fileName, String ext, String inhalt )  throws MyFileException, IOException {
    String result = null;
    File f = new File(directory, fileName); 
    File test = new File(directory);
    File bak = new File(directory, fileName + "." + ext);
    if (!test.isDirectory()) throw new MyFileException("MyFile.writeFileWithBak: there is no such directory " + test.getAbsolutePath());
    // Lsche File.ext;
    if (bak.exists()) bak.delete();
    f.renameTo(bak);
    f = new File(directory, fileName); 
    FileWriter out = null;
  try {
      out = new FileWriter(f);            // Create a char stream to read  it
      int size = (int) inhalt.length();       // Check file size
      out.write(inhalt, 0, size);
      out.flush();
      //f.renameTo(fileName);
    // Always be sure to close the input stream!
    } finally { 
        try { 
            if (out != null) out.close(); 
            
        } catch (IOException e) {System.out.println(e);} 
        }
  }
   
}








