// The "freEvolve System" is a platform supporting the reconfiguration of 
// component-based applications'
// Copyright (C) 1999  Ralph Hinken and Dr. Oliver Stiemerling
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Version: 1.0
// Date:    09.10.2001
//
// History:
//

package org.freevolve.server;
import java.util.*;
import java.rmi.*;
import java.rmi.server.*;

/** Verbindet zwei Parameter und merkt sich deren Komponenten, sowie deren Instanznamen in der Komponentenhierarchie. <br>
  Der Instanzname ist der Name unter dem eine Komponente bei ihrem Vorgnger angemeldet ist.<br>
  Weiterhin wird der Besitzer/Owner der Verbindung gemerkt, der diese erzeugt hat. Eine
  Verbindung reicht somit einen Parameter von einer Subkomponente zur Owner-Komponente hoch.
  @see Component
  @see Param
  */
public class ParamBinding  extends UnicastRemoteObject implements IR_ParamBinding {
  IR_Param      paramOne;
  IR_Param      paramTwo;
  Component     owner;
  boolean   isSymm;

  public ParamBinding(IR_Param p_one, IR_Param p_two, Component owner) throws RemoteException {
     paramOne = p_one;
     paramTwo = p_two;
     this.owner = owner;
     isSymm = false; // noch nicht benutzt;
     if (owner==null || paramOne == null || paramTwo == null) {
        owner.message("error", 1, "ParamBinding hat null Eintrge!" + toTxt());
     }
  }

  /** Gibt einen Parameter der Verbidung zurck. */
  public IR_Param getParamOne() { return paramOne; }

  /** Gibt einen Parameter der Verbidung zurck. */
  public IR_Param getParamTwo() { return  paramTwo; }
  
    /** Gibt die Komponente von Parameter One zurck. */
  public IR_Component getComponentOne() { 
      try {
         return paramOne.getOwner();
      } catch (RemoteException e) {
         owner.message("error", 1, "ParamBinding.getComponentOne() " + e.toString());
         return null;
      }
  }

  /** Gibt die Komponente von Parameter Two zurck. */
  public IR_Component getComponentTwo() { 
    try {
         return paramTwo.getOwner();
      } catch (RemoteException e) {
         owner.message("error", 1, "ParamBinding.getComponentTwo() " + e.toString());
         return null;
      }
  }

  /** Gibt von der Komponente, der Parameter One gehrt, den Namen unter dem diese
    Komponente bei seiner Vorgngerkomponente angemeldet ist, zurck. Falls Parameter one
    zur selben Komponente gehrt, die auch dem Verbidungsobjekt gehrt, wird "." zurckgegeben.*/
  public String getInstanceNameOne() { 
    try {
      IR_Component c = paramOne.getOwner();
      if (c.equals(owner)) return "."; 
      if (c==null) return null;
      else return c.getName();
   } catch (RemoteException e) {
      owner.message("error", 1, "ParamBinding.getInstanceNameOne() " + e.toString());
      return null;
   }
}
 

 /** Gibt von der Komponente, der Parameter Two gehrt, den Namen unter dem diese
    Komponente bei seiner Vorgngerkomponente angemeldet ist, zurck. Falls Parameter one
    zur selben Komponente gehrt, die auch dem Verbidungsobjekt gehrt, wird "." zurckgegeben.*/
  public String getInstanceNameTwo() { 
    try {
      IR_Component c = paramTwo.getOwner();
      if (c.equals(owner)) return "."; 
      if (c==null) return null;
      else return c.getName();
   } catch (RemoteException e) {
      owner.message("error", 1, "ParamBinding.getInstanceNameTwo() " + e.toString());
      return null;
   }
}

  /** Gibt die Komponente zurck, die diese Parameter Verbindung erzeugt hat. */
  public IR_Component getOwner() { return owner; }


  /** Hat noch keinen Sinn...*/
  public boolean isSymmetric() { return isSymm; }
  
    // ComponentName.ParamName-ComponentName.ParamName 
/** Gibt einen String zurck, der einer bind-Zeile (ohne bind Schlsselwort) 
   in der CAT Datei entspricht.
   <b>    ParamName - ComponentTwoName.ParamName </b>*/
  public String   toTxt() {
    try {
      String i1 = getInstanceNameOne();
      String i2 = getInstanceNameTwo();
      String result = "";
      if (i1.equals(".")) result += paramOne.getName();
      else result += i1 + "." + paramOne.getName();

      result += " - ";

      if (i2.equals(".")) result += paramTwo.getName();
      else result += i2 + "." + paramTwo.getName();

      return result;

    } catch (RemoteException e) {
      owner.message("error", 1, "ParamBinding.toTxt: " + e.toString() ); 
      return null;
    }
  }

/** Gibt einen String zurck, dieses Objekt darstellt.
   <b>    ParamName ComponentTwoName ParamName </b>*/
  public String   printShort() {
    try {
      String i1 = getInstanceNameOne();
      String i2 = getInstanceNameTwo();
      String result = "";
      if (i1.equals(".")) result += paramOne.getName();
      else result += i1 + " " + paramOne.getName();

      result += " ";

      if (i2.equals(".")) result += "null " + paramTwo.getName();
      else result += i2 + " " + paramTwo.getName();

      return result;

    } catch (RemoteException e) {
      owner.message("error", 1, "ParamBinding.printShort: " + e.toString() ); 
      return null;
    }
  }
 
  /** Testet, ob dieses Objekt  Parameter beider Komponenten verbindet.
  */
  public boolean connects(IR_Component cOne, IR_Component cTwo) {
    try {
       if (cOne.equals(paramOne.getOwner())  && cTwo.equals(paramTwo.getOwner())) return true;
       if (cTwo.equals(paramOne.getOwner()) && cOne.equals(paramTwo.getOwner())) return true;
    } catch (RemoteException e) {
      owner.message("error", 1, "ParamBinding.connects() " + e.toString());
    }
    return false;
  }
    
  /** Testet, ob dieses Objekt die Parameter verbindet.
  */
  public boolean isParamBinding(IR_Param oneParam, IR_Param twoParam) {
      if (oneParam.equals(paramOne) && twoParam.equals(paramTwo) ) return true;
      if (oneParam.equals(paramTwo) && twoParam.equals(paramOne) ) return true;
      return false;
  }

   /** Testet, ob dieses Objekt die Params verbindet. Wenn eine der beiden Komponenten
       die lokale Komponente ist, der das Verbindungsobjekt gehrt ist der Parameter 
       gleich ".".
  */
  public boolean isParamBinding(String oneComp, String oneParam, String twoComp, String twoParam) {
      try {
         if (oneComp.equals(getInstanceNameOne()) && oneParam.equals(paramOne.getName()) 	&&
             twoComp.equals(getInstanceNameTwo()) && twoParam.equals(paramTwo.getName())) return true;
         if (twoComp.equals(getInstanceNameOne()) && twoParam.equals(paramOne.getName()) 	&&
             oneComp.equals(getInstanceNameTwo()) && oneParam.equals(paramTwo.getName())) return true;
      }
      catch (RemoteException e) {
         owner.message("error", 1, "ParamBinding.isParamBinding: " + e.toString() ); 
      }
      return false;
  }

   /** Testet, ob dieses ParamBinding Objekt eine Subkomponente des owners mit dem Namen local_name anbindet.
   */
   public boolean connectsComp(String local_name) {
      if (local_name.equals(getInstanceNameOne()) || local_name.equals(getInstanceNameTwo())) return true;
      return false;
   }

 /** Gibt den Namen des Params One zurck, unter dem der Param bei seiner Komponente angemeldet ist.
 */
  public String getParamNameOne() { 
   try {
     return paramOne.getName();
    } catch (Exception e) {
      owner.message("error", 1, "ParamBinding.getParamNameOne() " + e.toString());
      return null;
   }
}

 /** Gibt den Namen des Params Two zurck, unter dem der Param bei seiner Komponente angemeldet ist.
 */
  public String getParamNameTwo() { 
    try {
      return paramTwo.getName();
    } catch (Exception e) {
      owner.message("error", 1, "ParamBinding.getParamNameTwo() " + e.toString());
      return null;
   }
} 

}  
