// The "freEvolve System" is a platform supporting the reconfiguration of 
// component-based applications'
// Copyright (C) 1999  Ralph Hinken and Dr. Oliver Stiemerling
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Version: 1.0
// Date:    09.10.2001
//
// History:
//

package org.freevolve.server;
import java.util.*;
import java.lang.reflect.*;
import java.rmi.server.*;
import java.rmi.*;

public class ParamProxy { // extends UnicastRemoteObject implements IR_Proxy {

  private String       type;            
  private String       name;            
  private Proxy        owner;
  private Vector       up;
  private Vector       down;
  private boolean      beanParam;
  private Method       get;
  private Method       set; 
  private Object       value;


  public ParamProxy (String name, String type,  Proxy owner) {
      this.name = name;
      this.type = type;
      this.owner = owner;
      this.up = new Vector();
      this.down = new Vector();
      this.beanParam = owner.isFlexiBean();
      this.value = null;
  }

  public ParamProxy (IR_Param p, Proxy owner) throws java.rmi.RemoteException {
      this.name = p.getName();
      this.type = p.getType();
      this.owner = owner;
      this.up = new Vector();
      this.down = new Vector();
      this.beanParam = owner.isFlexiBean();
      this.value = null;
  }

  public String getName() {return name;}
  public void   setName(String n) { name = n; }

  public void   setValue(Object v) {value = v;}
  public void   unsetValue() {value = null;}
  public Object getValue () { return value;}
  public boolean hasValue() { return (value != null); }
  public Method getSetMethod() {return set; }
  public Proxy  getOwner() {return owner;}

  public boolean readFunctions(Class klasse) {
    try {
        boolean weiter = true;
        get = null;
        set = null;
        while (weiter) {
            Method m[] = klasse.getDeclaredMethods();
            //owner.client.message("i",1,toTxt());                                                
            if (beanParam && ! name.equals("X-Pos") && ! name.equals("Y-Pos") 
                && ! name.equals("Height") && ! name.equals("Width") ) {
               String setFunc = "set" + this.name;
               String getFunc = "get" + this.name;
               String methodName = null;
            //    owner.client.message("i",1,addListener+"\n"+getShared+"\n"+removeListener);                        
               for(int i=0; i<m.length;i++){ 
                  methodName = m[i].getName();
                  // suche Funktionen 
                  if (getFunc.equals(methodName)) {
                       get = m[i];
                       //owner.client.message("info",1,"Param: " + m[i].getReturnType().getName() + " aus " + 
                       //m[i].getName());
                 } else if (setFunc.equals(methodName)) {
                       set = m[i];
                       //owner.client.message("info",1,"Param: " + m[i].getReturnType().getName() + " aus " + 
                       //m[i].getName());
                 }
               }
            } else { // if x,y,width or height
              weiter = false;
            }
            klasse = klasse.getSuperclass() ;
            if (klasse==null) weiter = false;
            if (klasse.getName().equals("java.lang.Object")) weiter = false;
            if (get!=null && set!=null) weiter = false;
        }
       if (beanParam && ! name.equals("X-Pos") && ! name.equals("Y-Pos") 
           && ! name.equals("Height") && ! name.equals("Width") ) {
           if (get == null || set == null)  {
               owner.client.message("error",1,"Param " + name + " nicht gefunden...");
               get = null;
               set = null;
               return false;
           }
       }
     } catch (Exception ex) {
            owner.client.message("error",1,"ParamProxy.ReadFunction: " + ex);
    }
    return true;
  } // END readFuncions


  public String toTxt() {
     if (hasValue())    return name + ":" + type + "=" + value.toString();
     return name + ":" + type + "= null";
  }

  /** Gibt verbundene ParamProxies zurck.
  */
  public Vector getUpParams() { return up; }

  /** Gibt verbundene ParamProxies zurck.
  */
  public Vector getDownParams() { return down; }

 
 /** Einen ParamProxy fr eine vertikale Verbindung anmelden.
     Der anzumeldende Param gehrt der Vaterkomponente.
     @return false, falls ParamProxy schon angemeldet ist.
  */
  public boolean addUp(ParamProxy pp) {
      if (up.contains(pp)) return false;
      up.add(pp);
      return true;
  }

 /** Einen ParamProxy fr eine vertikale Verbindung anmelden.
     Der anzumeldende Param gehrt der Subkomponente.
     @return false, falls ParamProxy schon angemeldet ist.
  */
  public boolean addDown(ParamProxy pp) {
      if (down.contains(pp)) return false;
      down.add(pp);
      return true;
  }

 /** Einen ParamProxy fr eine vertikale Verbindung abmelden.
     Der abzumeldende Param gehrt der Vaterkomponente.
     @return false, falls ParamProxy nicht angemeldet ist.
  */
  public boolean removeUp(ParamProxy pp) {
      if (!up.contains(pp)) return false;
      up.remove(pp);
      return true;
  }

 /** Einen ParamProxy fr eine vertikale Verbindung abmelden.
     Der abzumeldende Param gehrt der Subkomponente.
     @return false, falls ParamProxy nicht angemeldet ist.
  */
  public boolean removeDown(ParamProxy pp) {
      if (!down.contains(pp)) return false;
      down.remove(pp);
      return true;
  }
   

   /** Sucht alle Params, die dieser Param ber Verbindungen sammelt.
     <br>
     Sucht rekursiv alle nach unten verbundene Params. Also alle Params die mittels 
     "Hochreichen" an unseren Param angeschlossen sind.<br>
     Der Vector enthlt, falls pParam ein SharedObject-Provider-Param ist, nur einen Param.
 */
 public Vector getLeafParams() {
    Vector result = new Vector();
    if (this.isBeanParam()) {
        //owner.client.message("proxy",4,"is BeanParam " +this.toTxt());
        result.add(this);
        return result;
    }
    Vector down = this.getDownParams();
    Enumeration e = down.elements();
    ParamProxy p=null;
    while (e.hasMoreElements()) {
       p = (ParamProxy) e.nextElement();
       //owner.client.message("proxy",4,"is nach unten " +p.toTxt());
       result.addAll(p.getLeafParams());
    }
    return result;
}

      /** Gehrt der Param einem Proxy einer FlexiBean?
  */
  public boolean isBeanParam() { return beanParam; }

  /** Sucht ber Verbindungen zu ParamProxy Objekten der Vaterkomponente einen
      Parameter, der gesetzt ist.
  */
  public ParamProxy getValueParam() {
    if (this.hasValue()) {
        //owner.client.message("proxy",4,"is Param whith value " +this.toTxt());
        return this;
    }
    Vector up = this.getUpParams();
    Enumeration e = up.elements();
    ParamProxy p = null;
    while (e.hasMoreElements()) {
       p = (ParamProxy) e.nextElement();
       //owner.client.message("proxy",4,"is nach oben " +p.toTxt());
       if (p.hasValue()) return p;
    }
    return null;
  }

  
}  // END CLASS ParamProxy