// The "freEvolve System" is a platform supporting the reconfiguration of 
// component-based applications'
// Copyright (C) 1999  Ralph Hinken and Dr. Oliver Stiemerling
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Version: 1.0
// Date:    09.10.2001
//
// History:
//

package org.freevolve.server;
import java.util.*;
import java.rmi.*;
import java.rmi.server.*;

/** Verbindet zwei Ports und merkt sich deren Komponenten, sowie deren Instanznamen in der Komponentenhierarchie.
  Der Instanzname ist der Name unter dem eine Komponente bei ihrem Vorgnger angemeldet ist.<br>
  Weiterhin wird der Besitzer/Owner der Verbindung gemerkt, der diese erzeugt hat. Eine
  Verbindung reicht somit einen Port von einer Subkomponente zur Owner-Komponente hoch oder verbindet
  die Ports zweier Subkomponenten (die sind im owner unter den Instanznamen instanceNameOne/-Two bekannt)
  @see Component
  @see Port
  */
public class PortBinding  extends UnicastRemoteObject implements IR_PortBinding {
  IR_Port      portOne;
  IR_Port      portTwo;
  Component    owner;
  boolean   isSymm;

   public PortBinding(IR_Port p_one,  IR_Port p_two, Component owner) throws RemoteException {
     portOne = p_one;
     portTwo = p_two;
    
     this.owner = owner;
     isSymm = false; // noch nicht benutzt;
  }
  
  /** Testet, ob dieses Objekt Ports beider Komponenten verbindet.
  */
  public boolean connects(IR_Component cOne, IR_Component cTwo) {
    try {
       if (portOne.getOwner().equals(cOne) && portTwo.getOwner().equals(cTwo)) return true;
       if (cTwo.equals(portOne.getOwner()) && cOne.equals(portTwo.getOwner())) return true;
    } catch (RemoteException e) {
      owner.message("error", 1, "PortBinding.connects: "+e.toString());
    }
    return false;
  }


  /** Testet, ob dieses Objekt die Ports verbindet.
  */
  public boolean isPortBinding(IR_Port onePort,  IR_Port twoPort) {
      if (onePort.equals(portOne) && twoPort.equals(portTwo) ) return true;
      if (onePort.equals(portTwo) && twoPort.equals(portOne) ) return true;
      return false;
  }

 /** Testet, ob dieses Objekt die Ports verbindet.
  */
  public boolean isPortBinding(String oneComp, String onePort, String twoComp, String twoPort) {
      try {
         if (oneComp.equals(getInstanceNameOne()) && onePort.equals(portOne.getName()) 	&&
             twoComp.equals(getInstanceNameTwo()) && twoPort.equals(portTwo.getName())) return true;
         if (twoComp.equals(getInstanceNameOne()) && twoPort.equals(portOne.getName()) 	&&
             oneComp.equals(getInstanceNameTwo()) && onePort.equals(portTwo.getName())) return true;
      } catch (RemoteException e) {
         try { owner.getServer().message("error", 1, "PortBinding.isPortBinding: " + e.toString() ); }
         catch (RemoteException d) {;};
      }
      return false;
  }

   /** Testet, ob dieses PortBinding Objekt eine Subkomponente des owners mit dem Namen local_name anbindet.
   */
   public boolean connectsComp(String local_name) {
      //System.out.println( getInstanceNameOne() + getInstanceNameTwo());
      if (getInstanceNameOne().equals(local_name) || getInstanceNameTwo().equals(local_name)) return true;
      return false;
   }
 
  /** Gibt einen Port der Verbidung zurck. */
  public IR_Port getPortOne() { return portOne; }

  /** Gibt einen Port der Verbidung zurck. */
  public IR_Port getPortTwo() { return  portTwo; }

  /** Gibt die Komponente von Port One zurck. */
  public IR_Component getComponentOne() { 
      try {
         return portOne.getOwner(); 
      } catch (RemoteException e) {
         owner.message("error", 1, "PortBinding.getComp: " + e.toString() );
      }   
      return null;
  }

  /** Gibt die Komponente von Port Two zurck. */
  public IR_Component getComponentTwo() { 
      try {
        return portTwo.getOwner(); 
      } catch (RemoteException e) {
         owner.message("error", 1, "PortBinding.isPortBinding: " + e.toString() ); 
      }
      return null;
  }

 /** Gibt den Namen des Ports One zurck, unter dem der Port bei seiner Komponente angemeldet ist.
 */
  public String getPortNameOne() { 
   try {
     return portOne.getName();
    } catch (Exception e) {
      owner.message("error", 1, "PortBinding.getPortNameOne() " + e.toString());
      return null;
   }
}

 /** Gibt den Namen des Ports Two zurck, unter dem der Port bei seiner Komponente angemeldet ist.
 */
  public String getPortNameTwo() { 
    try {
      return portTwo.getName();
    } catch (Exception e) {
      owner.message("error", 1, "PortBinding.getPortNameTwo() " + e.toString());
      return null;
   }
}


 /** Gibt von der Komponente, der Port Eins gehrt, den Namen unter dem diese
    Komponente bei seiner Vorgngerkomponente angemeldet ist, zurck. Falls Port Eins
    zur selben Komponente gehrt, die auch dem Verbidungsobjekt gehrt, wird "." zurckgegeben.*/
  public String getInstanceNameOne() { 
    try {
      IR_Component c = portOne.getOwner();
      if (c.equals(this.owner)) return "."; 
      if (c==null) return null;
      else return c.getName();
   } catch (RemoteException e) {
      owner.message("error", 1, "PortBinding.getInstanceNameOne() " + e.toString());
      return null;
   }
}

 /** Gibt von der Komponente, der Port Zwei gehrt, den Namen unter dem diese
    Komponente bei seiner Vorgngerkomponente angemeldet ist, zurck. Falls Port Zwei
    zur selben Komponente gehrt, die auch dem Verbidungsobjekt gehrt, wird "." zurckgegeben.*/
  public String getInstanceNameTwo() { 
    try {
      IR_Component c = portTwo.getOwner();
      if (c.equals(this.owner)) return "."; 
      if (c==null) return null;
      else return c.getName();
   } catch (RemoteException e) {
      owner.message("error", 1, "PortBinding.getInstanceNameTwo() " + e.toString());
      return null;
   }
}

  /** Gibt die Komponente zurck, die diese Port Verbindung erzeugt hat. */
  public IR_Component getOwner() { return owner; }

  /** Hat noch keinen Sinn...*/
  public boolean isSymmetric() { return isSymm; }
  

  // ComponentName.PortName-ComponentName.PortName 
/** Gibt einen String zurck, der einer bind-Zeile (ohne bind Schlsselwort) 
   in der CAT Datei entspricht.
   <b>    ComponentOneName.PortName - ComponentTwoName.PortName </b><br>
   <b>                     PortName - ComponentTwoName.PortName </b><br>
   <b>    ComponentOneName.PortName -                  PortName </b><br>
 */

  public String   toTxt() {
    try {
      String i1 = getInstanceNameOne();
      String i2 = getInstanceNameTwo();
      String result = "";
      if (i1.equals(".")) result += portOne.getName();
      else result += i1 + "." + portOne.getName();
      result += " - ";

      if (i2.equals(".")) result += portTwo.getName();
      else result += i2 + "." + portTwo.getName();

      return result;

    } catch (RemoteException e) {
      owner.message("error", 1, "PortBinding.toTxt: " + e.toString() ); 
      return "error";
    }
  }

/** Gibt einen String zurck, diese Objekt darstellt
   <b>    ComponentOneName PortName   ComponentTwoName PortName </b><br>
   <b>    null             PortName   ComponentTwoName PortName </b><br>
   <b>    ComponentOneName PortName   null             PortName </b><br>
 */

  public String   printShort() {
    try {
      String i1 = getInstanceNameOne();
      String i2 = getInstanceNameTwo();
      String result = "";
      if (i1.equals(".")) result += "null " + portOne.getName();
      else result += i1 + " " + portOne.getName();
      result += " ";

      if (i2.equals(".")) result += "null " + portTwo.getName();
      else result += i2 + " " + portTwo.getName();

      return result;

    } catch (RemoteException e) {
      owner.message("error", 1, "PortBinding.printShort: " + e.toString() ); 
      return "error";
    }
 }

}
