// The "freEvolve System" is a platform supporting the reconfiguration of 
// component-based applications'
// Copyright (C) 1999  Ralph Hinken and Dr. Oliver Stiemerling
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Version: 1.0
// Date:    09.10.2001
//
// History:
//

package org.freevolve.server;
import java.util.*;
import java.lang.reflect.*;
import java.rmi.server.*;
import java.rmi.*;

public class PortProxy  extends UnicastRemoteObject  implements IR_PortProxy {

  private String       type;            
  private String       name;            
  private String       polarity;
  private Proxy        owner;
  private Vector       horizontal;
  private Vector       up;
  private Vector       down;
  private Vector       remote;
  private Method       get;
  private Method       add; // or set
  private Method       remove; // or forget
  private boolean      BeanPort;
  private boolean      isEvent;
//  private Class        porttyp;

  public PortProxy (IR_Port port, Proxy owner) throws RemoteException {
      this.name = port.getName();
      String type = port.getType();
      if (type.indexOf(".")!=-1) type = type.substring(type.indexOf(".")+1);
      this.type = type;
      this.polarity = port.getPolarity();
      this.owner = owner;
      get = remove = add = null;
      horizontal = new Vector();
      up = new Vector();
      down  = new Vector();
      remote = new Vector();
      BeanPort = owner.isFlexiBean();

  }

  public PortProxy (String n, String t, String p, Proxy owner) throws RemoteException {
      this.name = n;
      String type = t;
      if (type.indexOf(".")!=-1) type = type.substring(type.indexOf(".")+1);
      this.type = type;
      this.polarity = p;
      this.owner = owner;
      get = remove = add = null;
      horizontal = new Vector();
      up = new Vector();
      down  = new Vector();
      remote = new Vector();
      BeanPort = owner.isFlexiBean();
  }

  public String getName() { return name; }
  public void   setName(String n) { name = n; }

  public String getType() { return type; }

  public IR_Proxy getOwner() { return owner; }

  public String getPolarity() { return polarity;}

  /** Reprsentiert der Port einen Event?
  */
  public void setEvent(boolean b) { isEvent = b; }

  /** Reprsentiert der Port einen Event?
  */
  public boolean isEvent() { return isEvent; }

  /** Gehrt der Port einem Proxy einer FlexiBean?
  */
  public boolean isBeanPort() { return BeanPort; }

  /** Gibt Zugriffsmethode zurck. Nur bei BeanPorts.
      @return get-Methode (Events + Shared Objects)
  */
  public Method getGetMethod() { return get;}

  /** Gibt Zugriffsmethode zurck. Nur bei BeanPorts.
      @return add-Methode bei Events, set-Methode bei Shared Objects
  */
  public Method getAddMethod() { return add;}

  /** Gibt Zugriffsmethode zurck. Nur bei BeanPorts.
      @return remove-Methode bei Events, forget-Methode bei Shared Objects
  */

  public Method getRemoveMethod() { return remove;}

  /** Gibt verbundene PortProxies zurck.
  */
  public Vector getHorizontalPorts() { return horizontal; }

  /** Gibt verbundene PortProxies zurck.
  */
  public Vector getUpPorts() { return up; }

  /** Gibt verbundene PortProxies zurck.
  */
  public Vector getDownPorts() { return down; }

  /** Einen PortProxy fr eine horizontale Verbindung hinzufgen.
      @return false, falls PortProxy schon angemeldet ist.
  */
  public boolean addHorizontal(IR_PortProxy pp) {
      if (horizontal.contains(pp)) return false;
      horizontal.add(pp);
      return true;
  }

 /** Einen PortProxy fr eine vertikale Verbindung anmelden.
     Der anzumeldende Port gehrt der Vaterkomponente.
     @return false, falls PortProxy schon angemeldet ist.
  */
  public boolean addUp(IR_PortProxy pp) {
      if (up.contains(pp)) return false;
      up.add(pp);
      return true;
  }

 /** Einen PortProxy fr eine vertikale Verbindung anmelden.
     Der anzumeldende Port gehrt der Subkomponente.
     @return false, falls PortProxy schon angemeldet ist.
  */
  public boolean addDown(IR_PortProxy pp) {
      if (down.contains(pp)) return false;
      down.add(pp);
      return true;
  }


  /** Einen PortProxy fr eine horizontale Verbindung abmelden.
      @return false, falls PortProxy nicht angemeldet ist.
  */
  public boolean removeHorizontal(IR_PortProxy pp) {
      if (!horizontal.contains(pp)) return false;
      horizontal.remove(pp);
      return true;
  }

 /** Einen PortProxy fr eine vertikale Verbindung abmelden.
     Der abzumeldende Port gehrt der Vaterkomponente.
     @return false, falls PortProxy nicht angemeldet ist.
  */
  public boolean removeUp(IR_PortProxy pp) {
      if (!up.contains(pp)) return false;
      up.remove(pp);
      return true;
  }

 /** Einen PortProxy fr eine vertikale Verbindung abmelden.
     Der abzumeldende Port gehrt der Subkomponente.
     @return false, falls PortProxy nicht angemeldet ist.
  */
  public boolean removeDown(IR_PortProxy pp) {
      if (!down.contains(pp)) return false;
      down.remove(pp);
      return true;
  }
   
  public String toTxt() {
      String result = "Name="+name + " Typ=" + type + " IS " + polarity;
      if (get!=null) result += " get=" + get.getName();
      if (add!=null) result += " add=" + add.getName();
      if (remove!=null) result += " remove= " + remove.getName();
      result += " von " + owner.getName() + " Typ " + owner.getType();
      return  result;
  }

  public boolean readFunctions(Class klasse) {
  get = add = remove = null;
  try {
    boolean weiter = true;
    while (weiter) {
        Method m[] = klasse.getDeclaredMethods();
        //owner.client.message("i",1,toTxt());                                                
        if (polarity.equals("provided")) {
           String postfix = type + "_" + name;
           String addListener = "add" + postfix;
           String getShared = "getShared" + postfix;
           String removeListener ="remove" + postfix;
           String methodName = null;
       //    owner.client.message("i",1,addListener+"\n"+getShared+"\n"+removeListener);                           
           for(int i=0; i<m.length;i++){ 
              methodName = m[i].getName();
              //owner.client.message("info",1,m[i].getReturnType().getName() + " aus " + m[i].getName());
              if (addListener.equals(methodName)) {
                   add = m[i];
                   setEvent(true);
                   //owner.client.message("info",1,m[i].getReturnType().getName() + " aus " + m[i].getName());
                   //Class pt[] = m[i].getParameterTypes();
                   //if (pt.length>0) owner.client.message("i",1,pt[0].getName());
                   /*
                   try{ 
                      if (owner.ucl != null)  porttyp = Class.forName(pt[0].getName(), true, owner.ucl);
                      else                    porttyp = Class.forName(pt[0].getName());
                   }
                   catch (Exception ex) { owner.client.message("error",1,"portproxy: "+ex.toString()); }
                   */
             } else if (getShared.equals(methodName)) {
                   get = m[i];
                   setEvent(false);
                   //owner.client.message("info",1,m[i].getReturnType().getName() + " aus " + m[i].getName());
                   //Class pt[] = m[i].getParameterTypes();
                   //if (pt.length>0) owner.client.message("i",1,pt[0].getName());
             } else if (removeListener.equals(methodName)) {
                   remove = m[i];
                   setEvent(true);
                   //owner.client.message("info",1,m[i].getReturnType().getName() + " aus " + m[i].getName());
                   //Class pt[] = m[i].getParameterTypes();
                   //if (pt.length>0) owner.client.message("i",1,pt[0].getName());
             }
          }
          //" von " + (m[i].getParameterTypes())[0].getName());
        } else if (polarity.equals("required")) {
           String postfix = type + "_" + name;
           String getListener = "getPort" + postfix;
           String setShared = "setShared" + postfix;
           String forgetShared ="forgetShared" + postfix;
           String methodName = null;
       //    owner.client.message("i",1,addListener+"\n"+getShared+"\n"+removeListener);
           for(int i=0; i<m.length;i++){ 
              methodName = m[i].getName();
              //owner.client.message("info",1,m[i].getReturnType().getName() + " aus " + m[i].getName());
              if (getListener.equals(methodName)) {
                   get = m[i];
                   setEvent(true);
                   //owner.client.message("info",1,m[i].getReturnType().getName() + " aus " + m[i].getName());
                   //Class pt[] = m[i].getParameterTypes();
                   //if (pt.length>0) owner.client.message("i",1,pt[0].getName());
             } else if (setShared.equals(methodName)) {
                   add = m[i];
                   setEvent(false);
                   //owner.client.message("info",1,m[i].getReturnType().getName() + " aus " + m[i].getName());
                   //Class pt[] = m[i].getParameterTypes();
                   //if (pt.length>0) owner.client.message("i",1,pt[0].getName());
                   /*
                   try{ 
                      if (owner.ucl != null)  porttyp = Class.forName(pt[0].getName(), true, owner.ucl);
                      else                    porttyp = Class.forName(pt[0].getName());
                   }
                   catch (Exception ex) { owner.client.message("error",1,"portproxy: "+ex.toString()); }
                   */
             } else if (forgetShared.equals(methodName)) {
                   remove = m[i];
                   setEvent(false);
                   //owner.client.message("info",1,m[i].getReturnType().getName() + " aus " + m[i].getName());
                   //Class pt[] = m[i].getParameterTypes();
                   //if (pt.length>0) owner.client.message("i",1,pt[0].getName());
             }
           }

        }
        if ((add!=null && remove!=null) || get!=null) weiter=false;
          klasse = klasse.getSuperclass() ;
          if (klasse==null) weiter = false;
          if (klasse.getName().equals("java.lang.Object")) weiter = false;

        }
        } catch (Exception ex) {
            owner.client.message("error",1,"PortProxy.ReadFunction: " + ex);
        }
   return true;
  }

   /** Sucht alle Ports, die dieser Port ber Verbindungen sammelt.
     <br>
     Sucht rekursiv alle nach unten verbundene Ports. Also alle Ports die mittels 
     "Hochreichen" an unseren Port angeschlossen sind.<br>
     Der Vector enthlt, falls pport ein SharedObject-Provider-Port ist, nur einen Port.
 */
 public Vector getLeafPorts() {
    try {
        Vector result = new Vector();
        if (this.isBeanPort()) {
            //owner.client.message("proxy",4,"is BeanPort " +this.toTxt());
            result.add(this);
            return result;
        }
        Vector down = this.getDownPorts();
        Enumeration e = down.elements();
        IR_PortProxy p=null;
        while (e.hasMoreElements()) {
           p = (IR_PortProxy) e.nextElement();
           //owner.client.message("proxy",4,"is nach unten " +p.toTxt());
           result.addAll(p.getLeafPorts());
        }
        return result;
    } catch (RemoteException x) {
        owner.client.message("error",1,"PortProxy.getLeafPorts: " + x.toString());
        return new Vector();
    }
 }

 /** Sucht verbundene Ports.
     Sucht rekuriv Wege ber horizontale Verbindungen und Verbindungen nach oben, also
     Verbindungen, durch die unser Port pport weiter nach "oben" in der Komponentenhierarchie
     durchgereicht wird ( - und dort dann nach horizontalen Verbindungen - ).<br>
     Gibt die ersten PortProxies auf dem Weg zurck, die andere Polaritt haben. 
 */
 public Vector getHorConnected() {
    try {
        Vector result = new Vector();
        Vector up = this.getUpPorts();
        result.addAll(this.getHorizontalPorts());
        Enumeration e = up.elements();
        IR_PortProxy p=null;
        while (e.hasMoreElements()) {
           p = (IR_PortProxy) e.nextElement();
           result.addAll(p.getHorConnected());
        }
        return result;
    } catch (RemoteException x) {
        owner.client.message("error",1,"PortProxy.getHorConnected: " + x.toString());
        return new Vector();
    }
 }


  public Object  invokeGet() {
      Object o = null;
      if (get==null) {
          owner.client.message("error",1, "**PortProxy.invokeGet: Keine get Fkt gefunden in Port " + toTxt());  
          return null;
      }
       try {
          //owner.client.message("proxy", 4, "aufruf get in PortProxy von: " + get );
          o = get.invoke(owner.getObject(), null);
          //owner.client.message("proxy", 4, "aufruf get ok");
      } catch (InvocationTargetException itx) {
            owner.client.message("error",1,  "**PortProxy.invokeGet: " + itx.toString() + "  targetException ist: " + itx.getTargetException().toString() + " von " + toTxt());
       } catch (Exception x) {
          owner.client.message("error",1, "**PortProxy.invokeGet: " + x.toString() +" bei " + toTxt());
      }
      return  o;
  }

  public boolean invokeAdd(Object ob) {
      //if (ob instanceof Lists.RemoteList) owner.client.message("proxy",4,"INVOKEADD instance RemoteList");             
      Object[] o = new Object[1];
      if (ob==null) o = null;
      else o[0] = ob;
      if (add==null) {
          owner.client.message("error",1, "**PortProxy.invokeAdd: Keine add Fkt gefunden in Port " + toTxt());  
          return false;
      }
       try {
          //owner.client.message("proxy", 4, "aufruf add in PortProxy von :" + add );
          add.invoke(owner.getObject(), o);
      } catch (InvocationTargetException itx) {
            owner.client.message("error",1, "**PortProxy.invokeAdd: " + itx.toString() + " targetException ist: " + itx.getTargetException().toString() + " von " + toTxt());
       } catch (Exception x) {
          owner.client.message("error",1, "**PortProxy.invokeAdd: " + x.toString() +" bei " + toTxt());
          x.printStackTrace(); 
          return false;
      }
      return  true;
  }

  public boolean invokeRemove(Object ob) {
      Object[] o = new Object[1];
      if (ob==null) o = null;
      else o[0] = ob;
      if (remove==null) {
          owner.client.message("error",1, "**PortProxy.invokeRemove: Keine remove Fkt gefunden in Port " + toTxt());  
          return false;
      }
       try {
          //owner.client.message("proxy", 4, "aufruf remove in PortProxy\n" + add );
          remove.invoke(owner.getObject(), o);
      } catch (InvocationTargetException itx) {
            owner.client.message("error",1, "**PortProxy.invokeRemove: " + itx.toString()+ " targetException ist: " + itx.getTargetException().toString() + " von " + toTxt());
       } catch (Exception x) {
          owner.client.message("error",1, "**PortProxy.invokeRemove: " + x.toString() +" bei " + toTxt());
          return false;
      }
      return  true;
  }

}  