// The "freEvolve System" is a platform supporting the reconfiguration of 
// component-based applications'
// Copyright (C) 1999  Ralph Hinken and Dr. Oliver Stiemerling
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Version: 1.0
// Date:    09.10.2001
//
// History:
//

package org.freevolve.server;
import java.io.*;
import java.util.*;
import java.rmi.*;
import java.rmi.server.*;

/** Verwaltet eine Eigenschaftendatei (z.B. ini-Dateien).
   Ein Eintrag einer Eigenschaftendatei sieht folgendermaen aus: <br>
   Schlssel=Wert<br><br>
   
   Dabei kann der Schlssel hierarchisch aufgebaut sein mit Punkt oder Slash als
   Trennsymbol (z.B. user.Lieblingsverzeichnis=d:\)<br><br>

   Ein Prop Objekt wird mittels Angabe von Verzeichnis und Dateiname direkt
   an eine Eigenschaftendatei gekoppelt. Alle Schlssel werden dort herausgelesen
   und hereingeschrieben. <b> Die Klasse untersttzt keinen konkurriernden Zugriff!
   Mglichst sollte pro Datei nur ein Prop Objekt erzeugt werden!</b>
*/
public class Prop extends UnicastRemoteObject implements IR_Prop {

  private String propDir;
  private String propFileName;
  private Hashtable properties;

   /** Erzeugt ein Prop Objekt, das an die Datei pDir/pFileName gebunden wird.<br>
   <b> Die Klasse untersttzt keinen konkurriernden Zugriff!
   Mglichst sollte pro Datei nur ein Prop Objekt erzeugt werden!</b>
   */
  public Prop(String pDir, String pFileName) throws RemoteException {
    super();
    propDir      = pDir;
    propFileName = pFileName;
    properties   = new Hashtable();

    firstReadAllProperties();
  }
  
  /** Gibt den Wert fr einen Schlssel zurck.
  */
  public String getProp(String key)  {
    String value;

    if (properties.containsKey(key)) {
      value = (String) properties.get(key);
    } else {
      value = getMoreProp(key);
      if (value != null) properties.put(key, value);
    }
    return value;
  }
  
  String toTxt() {
      String s="";
      String key;
      for (Enumeration e= properties.keys(); e.hasMoreElements(); ) {
         key= (String) e.nextElement();
         s += key + " = " + properties.get(key) +"\n";
      }  
      return s;
  }

  /** Schreibt fr einen Schlssel einen neuen Wert in die Datei.
      Der Schlssel wird angehngt, falls er neu ist.
  */
  public synchronized void  setProp(String key, String value)   {
    properties.put(key, value); // Erst in lokale Hashtable schreiben, dann in Datei sichern 
       try {
         MyFile file = new MyFile();
         file.writeFileWithBak(propDir, propFileName, "bak", this.toTxt()); 
       } catch (Exception e) { 
         System.out.println("Prop.setProp: write to file: "+ 
               e.getClass().getName() + ": " + e.getMessage());
       }
  }


  /** Lscht einen Schlssel in der Datei.
  */
  public synchronized void  forgetProp(String key)   {
      properties.remove(key); // Erst in lokale Hashtable schreiben, dann in Datei sichern 
       try {
         MyFile file = new MyFile();
         file.writeFileWithBak(propDir, propFileName, "bak", this.toTxt()); 
       } catch (Exception e) { 
         System.out.println("Prop.removeProp: write to file: "+ 
               e.getClass().getName() + ": " + e.getMessage());
       }
  }
  
  String getMoreProp(String key) {  // sieht in Datei nach
    File f;
    FileInputStream fis = null;
    try {
      f = new File(propDir, propFileName);
      fis = new FileInputStream(f);
      PropertyResourceBundle props = new PropertyResourceBundle(fis);
      Enumeration allProps = props.getKeys();
      for(String entry=null ; allProps.hasMoreElements() ; ) {
	entry =  (String) allProps.nextElement();
	if (entry.compareTo(key) == 0) 
	  return props.getString(entry);
      }
    } catch (IOException e) {
      System.err.println("IOFehler: " + e.getClass().getName() + ": " + e.getMessage());
    }
    finally { 
      try { 
	if (fis != null) fis.close(); 
      } catch (IOException e) {} 
    }
    return null;
  }

  private void firstReadAllProperties() {
    File f;
    FileInputStream fis = null;
    try {
      f = new File(propDir, propFileName);
      fis = new FileInputStream(f);
      PropertyResourceBundle props = new PropertyResourceBundle(fis);
      Enumeration allProps = props.getKeys();
      for(String entry=null ; allProps.hasMoreElements() ; ) {
	entry =  (String) allProps.nextElement();
	properties.put(entry, props.getString(entry));
      }
    } catch (IOException e) {
      System.err.println("IOFehler: " + e.getClass().getName() + ": " + e.getMessage());
    }
    finally { try { if (fis != null) fis.close(); } catch (IOException e) {}  }
  }
  
  /*
  public static void main(String[] args) {
    System.out.println("start");
    if (args.length != 3)
      System.err.println("Aufruf: java prop <Property Directory> <Property-Datei> <Property>");
    else {
      Prop x = new Prop(args[0], args[1]);
      System.out.println("Schlssel ist " + x.getProp(args[2]));
      x.setProp("key", "value");
    }
    System.out.println("end");
  }
  */
   
}
