// The "freEvolve System" is a platform supporting the reconfiguration of 
// component-based applications'
// Copyright (C) 1999  Ralph Hinken and Dr. Oliver Stiemerling
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Version: 1.0
// Date:    09.10.2001
//
// History:
//

package org.freevolve.server;

//package Server;
import java.util.*;
import java.io.*;
import java.rmi.server.*;
import java.rmi.*;



/**
  Ein RBD-Objekt steht fr eine RBD-Datei, die Verbindungen zwischen zwei  lokalen Komponenten definiert.
  Diese lokalen Komponenten sind in den DistComponent-Objekten deklariert.
<BR>  
  Fr je zwei DistComponents muss ein neues RBD-Objekt definiert werden, da sich eine 
  RBD-Datei immer nur auf zwei lokale Komponenten bezieht. Die einzelnen Verbindungen 
  zwischen zwei Ports werden durch RBind Objekte definiert.<BR><BR>
  
  Ein RBD-Objekt wird von einem DistComponent Objekt erzeugt, dass dem Benutzer gehrt, 
  der ber dieses RBD Objekt seine Komponente an eine andere entfernte Komponente 
  anbinden mchte. Das DistComponent Objekt sorgt auch dafr, dass das RBD-Objekt in 
  beiden DistComponents referenziert wird.
  <BR><BR>
  Es wird nicht an den ROOT-Ports der Komponenten angemeldet, sodass
  Vernderungen  dort im RBD-Objekt nicht mitbekommen werden. 
  <BR><BR>
  Eine RBD-Datei kann von vielen Benutzern benutzt werden, als
  Beschreibung, wie die eigene Komponente an eine andere entfernte Komponente
  angeschlossen wird. Wenn die Datei allerdings von mehreren Benutzern benutzt
  wird, wird aber fr jeden Benutzer ein neues RBD Objekt erzeugt und eine
  Vernderung eines RBD Objektes verndert nicht die laufende Verbindungen
  der anderen Benutzer. Das vernderte Objekt kann aber wieder als Datei abgespeichert
  werden, und bei den anderen Benutzen im laufenden System fr das alte ersetzt werden.
  <BR><BR>
  Wichtig ist, dass im Konstruktor alles korrekt bergeben wird!
<BR><BR>
  Auf dieser Ebene der Remote Binds geht es nur um die Beschreibung!
  Proxy-Informationen werden in DistComponent verarbeitet. Hier wird
  nur die Integritt bzgl. der Component- und Port-Objekte sichergestellt.
  Dabei mssen die Ports vorhanden und von identischem Typ sein!

  @see DistComponent
  @see RBind
  */
public class RBD extends UnicastRemoteObject implements IR_RBD {
  // Variables ***************************************************************;
  private IR_DistComponent distCompOne;  // der Besitzer!
  private IR_DistComponent distCompTwo;
  private String   name;    // RBD Typ, z.B. Name der RBD-Datei;
  private Vector   rbinds;  // Vector of RBind!!;
  private Server   server;

  // Constants ***************************************************************;
  /** Trennt zwei Portnamen in einer rbind-Anweisung.
    rbind PortName1 PORTTRENNER PortName2
    */
  public final static String  PORTTRENNER = "-";
  /** Definiert den remote Bind Befehl einer rbind-Anweisung
    BINDKEY PortName1 PORTTRENNER PortName2
    */
  public final static String  BINDKEY     = "rbind";


  // Constructors ************************************************************;
  // Zum Deserialisieren - eventuell Vektoren reservieren?;
  public RBD() throws RemoteException {
    super();
  }

  /** Neues leeres RBD Objekt erzeugen, das "c1" mit "c2" verbindet und "name" heisst.
      Aufgerufen von DistComponent.addRBD, wobei dies c1 ist!
    */
  RBD(String name, IR_DistComponent c1, IR_DistComponent c2, Server s) throws RemoteException {
    super();
    this.name = name;
    distCompOne = c1;  // owner;
    distCompTwo = c2;
    rbinds = new Vector();
    server = s;
    // Testen, ob Component vorhanden? ;
  }

  /** RBD Objekt erzeugen, dass im String "rbd" definiert ist und "c1" mit "c2" verbindet.
      Die Funktion wird von c1 aufgerufen!
    Der String "rbd" enthlt durch Semikolon getrennte rbind Zeilen, wobei sich der erste
    Port immer auf DistComponent "c1" bezieht.<br>
    <b>Beispiel: rbind PortName1 - PortName2 ; </b>
    */
  RBD(String name, IR_DistComponent c1, IR_DistComponent c2, String rbd, Server s)
    throws RBD_Exception, RemoteException {
      super();
    this.name = name;
    server = s;
    distCompOne = c1;   // owner;
    distCompTwo = c2;
    rbinds = new Vector();
    readRBD(rbd);

  }

  /** Gibt den Erzeuger oder Besitzer des RBD Objektes zurck.
    @return Erzeuger des Objektes.
    */
  public IR_DistComponent getOwner() { return distCompOne; }
  /** Gibt die Komponente zurck, auf die sich die erste Portangabe bezieht.
    @return Eine der beiden Komponenten, die ber dieses Objekt verbunden werden.
    */
  public IR_DistComponent getDistComponentOne() { return distCompOne; }
  /** Gibt die Komponente zurck, auf die sich die zweite Portangabe bezieht.
    @return Eine der beiden Komponenten, die ber dieses Objekt verbunden werden.
    */
  public IR_DistComponent getDistComponentTwo() { return distCompTwo; }
  /** Der Name des RBD Objektes.
    Dieser Name sollte bezglich des Erzeuger DistComponent Objektes eindeutig sein.<BR><BR>
    Da ein RBD Objekt meisst einer RBD-Datei entspricht, sind beide Namen
    eventuell gleich. Eine RBD-Datei kann von vielen Benutzern benutzt werden, als
    Beschreibung, wie die eigene Komponente an eine andere entfernte Komponente
    angeschlossen wird. Wenn die Datei allerdings von mehreren Benutzern benutzt
    wird, wird aber fr jeden Benutzer ein neues RBD Objekt erzeugt und eine
    Vernderung eines RBD Objektes verndert nicht die laufende Verbindungen
    der anderen Benutzer. Das vernderte Objekt kann aber wieder als Datei abgespeichert
    werden, und bei den anderen Benutzen im laufenden System fr das alte ersetzt werden.
    
    @return Der Name des RBD-Objektes.
    */
  public String           getName() {return name; }
  
  // I/O Handling ************************************************************;
  
  /** Liest einen String, der eine RBD-Datei enthlt und baut dieses Objekt auf.
    Erzeugt Aufrufe von readRbind()
    @see readBind
    */
  private void readRBD(String rbd)  throws RBD_Exception {
    RBind r = null;
    StringTokenizer token = new StringTokenizer(rbd, ";");
    String          value = null;
    while (token.hasMoreTokens()) {
      r = readRBind(token.nextToken().trim());
      if (r!=null) addRBind(r);
    }
  }

  /** liest eine Zeile einer RBD-Datei: "rbind portOne - portTwo ; ";
    Aufgerufen von readRBD.
    @see RBD#PORTTRENNER
    @see RBD#BINDKEY
    @see readRBD
    */
  private RBind readRBind(String rbd) {
    try {
       StringTokenizer token = new StringTokenizer(rbd, " \n\t" + PORTTRENNER);
       if (!token.hasMoreTokens()) return null;
       String          value = token.nextToken();
       RBind           result = new RBind();
   
       try {
         if (value.equalsIgnoreCase(BINDKEY)) { value = token.nextToken(); }
         result.setPortNameOne(value);
         result.setPortNameTwo(token.nextToken());
         if (token.hasMoreTokens()) {
            	server.message("error", 1, "Dies ist keine rbind Anweisung: " +
   		       rbd+ " (" + token + ")");
   	        return null;
         }
       } catch (NoSuchElementException e) {
         server.message("error", 1, "Dies ist keine rbind Anweisung: " + rbd);
         return null;
       }

       return result;
    } catch (RemoteException e) {
      server.message("error", 1, "readRBind: RMI-Fehler " + e);
      return null;
    }
  }
  
  /** Erzeugt aus dem RBD-Objekt einen String, der eine RBD Datei reprsentiert.
    Aufruf von RBind.write()
    @see RBind#write
    */
  public String write  () {
    String result="";
    //if (this.name != null) result += "// name = " + this.name + "\n";
    for (Enumeration e = rbinds.elements() ; e.hasMoreElements() ;) {
      result += ((RBind) e.nextElement()).write() + " ;\n";
    }
    return result;
  }
  
  // hinzufuegen und entfernen von RBinds ********************************;

  /** Anmelden von einem RBind Objekt, also eine weitere Verbindung zwischen beiden Komponenten.
    Hier folgt der Test, ob RBind ok ist. Existieren beide Ports ?<BR><BR>
    Haben beide denselben Typ ?<BR><BR>
    @param r Ein weiteres RBind Objekt steht fuer eine Verbindung.
    @see RBind
    */
  private void addRBind(RBind r) throws RBD_Exception {
    if (r==null) return;
    try {
      if ((r.getPortNameOne() != null) && (r.getPortNameTwo() != null)) {
	      // Component muss in distComp angemeldet sein!;
         IR_Component c1 = distCompOne.getComponent();
         IR_Component c2 = distCompTwo.getComponent();
         if (c1==null || c2==null) {
            server.message("error", 1, "RBD.addRBind: Kann Komponente nicht finden.");
            return;
          }
      	r.setPortOne(c1.getIR_Port(r.getPortNameOne()));
         r.setPortTwo(c2.getIR_Port(r.getPortNameTwo()));
      	if (r.getPortTwo() == null) {
      	  server.message("error", 1, "RBD.addRBind: Port " + r.getPortNameTwo() + " existiert nicht.");
      	  throw new RBD_Exception("wrong Name?");
      	}
         if (r.getPortOne() == null) { 
      	  server.message("error", 1, "RBD.addRBind: Port " + r.getPortNameOne() + " existiert nicht.");
      	  throw new RBD_Exception("wrong Name?");
      	}
      	if (!r.getPortOne().getType().equals(r.getPortTwo().getType())) {
      	  server.message("error", 1, "RBD.addRBind: Ports sind nicht vom selben Typ.");
      	  throw new RBD_Exception("wrong Type");
      	}
      	rbinds.addElement(r); // everything is ok.;
      }
    } catch (RemoteException e) {
      server.message("error", 1, "RBD.addRbind: RMI-Fehler" +e);
    }
  }

  /** Abmelden eines RBind Objektes, also einer Verbindung zwischen unseren zwei Komponenten
    @param r Ein weiteres RBind Objekt steht fr eine Verbindung.
    @see RBind 
   */
  private void removeRBind(RBind r) {
    rbinds.removeElement(r);
  }

  /** Verbindet zwei Ports miteinander, die den lokalen Komponenten unserer DistComponent Objekten gehren.
    Erzeugt ein neues RBind Objekt und meldet es bei uns an.
    @param p1 Der Name des Ports der ersten lokalen Komponente, die in DistComponent owner ist.
    @param p2 Der Name des Ports der zweiten lokalen Komponente, die in DistComponent owner ist.
    @see DistComponent
    @see Component
    @see Port
    @see RBind
    */
   public void rbind(String p1, String p2) throws RBD_Exception {
     try {
       RBind r = new RBind(p1, p2);
       addRBind(r);
     } catch (RemoteException e) {
       server.message("error", 1, "RBD.rbind: RMI-Fehler: " + e);
       return;
    }
  }

  // asymmetrisch;
  /** Entfernt eine Verbindung zweier Ports, die den lokalen Komponenten unserer DistComponent Objekten gehren.
    @param p1 Der Name des Ports der ersten lokalen Komponente, die in DistComponent owner ist.
    @param p2 Der Name des Ports der zweiten lokalen Komponente, die in DistComponent owner ist.
    @see DistComponent
    @see Component
    @see Port
    @see RBind
    */
   public void unbind(String p1, String p2)  {
    RBind r = null;
    if (p1==null || p2==null) return;
    for (Enumeration e = this.rbinds.elements(); e.hasMoreElements(); ) {
         r = (RBind) e.nextElement();
         if (p1.equals(r.getPortNameOne()) && p2.equals(r.getPortNameTwo()) ) {
         	removeRBind(r);
         }
    }
  }

  /** Gibt alle Verbindungen an c.portName zurck.
    @param c Eine unserer beiden DistComponents, fr die wir die Verbindungen wissen mchten.
    @param portName Der Name des Ports der entfernten Komponente.
    @return Vector von RBind Objekten, die sich auf den bergebenden Port beziehen.
    */
  public Vector getRBinds(IR_DistComponent c, String portName) {
    Vector result = new Vector() ;
    RBind r = null;
    if (c.equals(distCompOne)) {
      for (Enumeration e = this.rbinds.elements(); e.hasMoreElements();) {
      	r = (RBind) e.nextElement() ;
      	if (portName.equals(r.getPortNameOne())) { result.addElement(r);} 
      }
    }
    if (c.equals(distCompTwo)) {
      for (Enumeration e = this.rbinds.elements(); e.hasMoreElements(); ) {
      	r = (RBind) e.nextElement();
      	if (portName.equals(r.getPortNameTwo())) { result.addElement(r);}
      }
    }
    return result;
  }

  /** Testet ob dieses RBD Objekt c1 mit c2 verbindet.
    HIER IST DIE REIHENFOLGE WICHTIG!
    Da ein RBD-Objekt von einem DistComponent Objekt erzeugt wird, dass sich
    als DistComponentOne anmeldet wird erwartet, dass c1 der Besitzer des
    RBD Objektes ist!
    */
  public boolean links(IR_DistComponent c1, IR_DistComponent c2) {
//    if (rbinds.isEmpty()) { return false; }
    if ((c1.equals(distCompOne)) && (c2.equals(distCompTwo))) return true;
    return false;
  }

  /** Testet ob dieses RBD Objekt die Komponente c verbindet.
   */
  public boolean links(IR_DistComponent c) {
    if (c.equals(distCompOne) || c.equals(distCompTwo)) return true;
    return false;
  }

   /** Testet, ob dieses RBD Objekt zwei Ports verbindet.
       Test, ob RBinds angemeldet sind.
       @see RBind
   */
  public boolean isEmpty() { return rbinds.isEmpty(); }
  
  /** Testet ob dieses RBD Objekt c gehrt.

    Da ein RBD-Objekt von einem DistComponent Objekt erzeugt wird, dass sich
    als DistComponentOne anmeldet, wird erwartet, dass diese Komponente
    der Besitzer des  RBD Objektes ist!
    */
  public boolean owns(IR_DistComponent c) {
    return (c.equals(distCompOne));
  }

  /** Alle Verbindungen zwischen beiden Komponenten, die in diesem Objekt definiert sind.
    @return Vector von allen RBind Objekten, die hier angemeldet sind.
    */
  public Vector getRBinds() {
    Vector v = (Vector) rbinds.clone();
    return v;
  }


}









