// The "freEvolve System" is a platform supporting the reconfiguration of 
// component-based applications'
// Copyright (C) 1999  Ralph Hinken and Dr. Oliver Stiemerling
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Version: 1.0
// Date:    09.10.2001
//
// History:
//

package org.freevolve.server;

import java.util.*;
import java.rmi.*;
import java.rmi.server.*;

/**
  Die Klasse RBind stellt eine remote Verbindung zwischen zwei Ports dar.
  Ein Objekt dieser Klasse wird durch ein Objekt der Klasse RBD erzeugt.
  Dieses Objekt referenziert beide Ports, die verbunden werden sollen und
  deren Namen. Die Klasse ist aber nur ein Container fr diese Informationen,
  die ber Zugriffsfunktionen eingetragen werden mssen. Die Klasse enthlt
  keine weitere Integrittsprfungen.

  @see RBD
  @see IR_RBind
  @see Port
  */
public class RBind extends UnicastRemoteObject implements IR_RBind {
  private boolean isSymmetric;
  private IR_Port    portOne;    // Port of Component Structure on Server
  private IR_Port    portTwo;
  private String     portNameOne;
  private String     portNameTwo;

  
  RBind() throws RemoteException {
    super();
    portOne = null;
    portTwo = null;
    portNameOne = null;
    portNameTwo = null;
    isSymmetric = false;
  }

  RBind(String pO, String pT) throws RemoteException {
    super();
    portOne = null;
    portTwo = null;
    portNameOne = pO;
    portNameTwo = pT;
    isSymmetric = false;
  }

  /** Gibt einen String zurueck, der eine rbind-Zeile einer RBD Datei darstellt.
    In der Regel in der Form <BR> "rbind PortName1 - PortName2",<BR> wobei in der RBD-Datei
    noch ein Semikolon folgt.

    @see RBD#BINDKEY
    @see RBD#PORTTRENNER
  */
  public String write() {
    return RBD.BINDKEY + " " + portNameOne + " " + RBD.PORTTRENNER +
      " " + portNameTwo;
  }

  public void setPortNameOne(String s) { portNameOne = s; }
  public void setPortNameTwo(String s) { portNameTwo = s; }

  public void setPortOne(IR_Port s) { portOne = s;  }
  public void setPortTwo(IR_Port s) { portTwo = s;  }
  
  public String getPortNameOne() { return portNameOne; }
  public String getPortNameTwo() { return portNameTwo; }

  public IR_Port getPortOne() { return portOne; }
  public IR_Port getPortTwo() { return portTwo; }
  
 
}

