// The "freEvolve System" is a platform supporting the reconfiguration of 
// component-based applications'
// Copyright (C) 1999  Ralph Hinken and Dr. Oliver Stiemerling
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Version: 1.0
// Date:    09.10.2001
//
// History:
//

package org.freevolve.server;
import java.rmi.*;
import java.rmi.server.*;
import java.util.*;
import java.net.*;
import java.io.*;
import java.lang.reflect.*;

class SERVER_Exception extends Exception {
  public SERVER_Exception () {super(); }
  public SERVER_Exception (String s) {super(s); }
}

/** Der Evolve Server. Wird nur einmal instanziiert.
   Hier erhlt man alle oder bestimmte DistComponent Objekte,
   kann neue erzeugen lassen oder alte beende,
   erhlt Dateien und Verzeichnisse sowie Properties.
   ber die DistComponent Objekte knnen sie alles mit den Anwendungen
   machen. <br><br>

   Hierber knnen DistComponent Objekte (incl RBD) und Component Objekte 
   persistent gemacht werden.
*/
public class Server extends UnicastRemoteObject implements IR_Server, I_Info, IR_URLgetter {

  // Name eindeutig;
  // Hier nur lokale Komponenten, da hier erzeugt;
  private Hashtable components; // ueber TypName der Root_Komponente;
  private Vector    distComps;  // Typ DistComponent
  private Vector    name_id;    // Liste der name+id aller Clients
  private Hashtable user;       // Liste aller User und Usergruppen
  private Vector    tailor_listeners;

  private static int ID;
  public EvolveMsg evolveMsg;
  private String host;
  public  Prop evolveIni;
  private static final String PROPDIR=".";
  private int nr=0;
  private MyFile f;
  private static URLClassLoader ucl = null;
  private static URL[] evolveURLs = null;
  static final String iniFileName = "Evolve.ini";
  static final String URLTag      = "EvolveURLClassPath";
  static final String startTag    = "EvolveStartDCATFile";
  static final String EvolveSuffix = "ini";
  static String ROOTDIR;
  static String USERDIR;
  static String DCATDIR;
  static boolean debug = false;
  
  
  public Server() throws RemoteException, Exception {
    super();
    evolveIni = new Prop(PROPDIR, iniFileName);
    ROOTDIR   = evolveIni.getProp("Evolve.RootDir");
    USERDIR   = ROOTDIR + evolveIni.getProp("Evolve.UserDir");
    DCATDIR   = ROOTDIR + evolveIni.getProp("Evolve.DCATDir");
    String d  = evolveIni.getProp("Evolve.Debug");
    if (d!=null && d.equalsIgnoreCase("true")) {
        debug = true;
    }
    components = new Hashtable();
    distComps = new Vector();
    name_id = new Vector();
    user    = new Hashtable();
    tailor_listeners = new Vector();

    f = new MyFile();
    if (ucl == null) {
      String url = evolveIni.getProp(URLTag);
      if (url==null) {
                System.out.println("In Datei " + iniFileName + " fehlt Eintrag "+URLTag+ ". Nehme hkm.cs.uni-bonn.de");
                url="hkm.cs.uni-bonn.de";
      }
      evolveURLs = readClassPath(url);
      host = evolveURLs[0].getHost();
      ucl = new URLClassLoader(evolveURLs);
      System.out.println("URLClassloaader gestartet.");
   }
   Hashtable eIni = getProperties(iniFileName);
   for (Enumeration e = eIni.keys(); e.hasMoreElements(); ) {
      String key = (String) e.nextElement();
      if (key.startsWith("user.")) {
          User u = new User(key.substring(5), USERDIR, (String) eIni.get(key), false, this);
          if (user.put(key.substring(5), u) != null) 
              message("error", 1, "Server: In der Datei " + iniFileName + " existieren zwei Eintrge " + key);
      } else if (key.startsWith("group.")) {
          User u = new User(key.substring(6), USERDIR, (String) eIni.get(key), true, this);
          if (user.put(key.substring(6), u) != null) 
             message("error", 1, "Server: In der Datei " + iniFileName + " existieren zwei Eintrge " + key);
      }
   }

   //ucl=null;
 }

  public void showMsg() {
      evolveMsg = new EvolveMsg(this, debug);
      evolveMsg.show();
   }

   /** Ein Property Objekt, das mit Evolve.ini verbunden ist. 
       Hiermit lassen sich dann Eintrge auslesen und verndern, sowie
       neue erzeugen.<br><br>
       Benutzerspezifische Eintrge mssen immer mit dem Benutzernamen und
       einem Punkt anfangen.
       @see Prop
   */
  public IR_Prop giveMeProp() {
    return evolveIni;
  }

  /** Ein MyFile Objekt zum Verzeichnis lesen, Dateien lesen und schreiben.
      @see MyFile
  */
  public IR_File giveMeFile() {
    try {
      return new MyFile();
      
    } catch (RemoteException e) {
      message("error", 1, "MyFile() : " + e.toString());
    }
    
    return null;
  }

  /** Gibt ein Component Objekt zurck, falls es schon vorhanden ist oder erzeugt es selbst.
      Falls diese Komponente nicht am Server angemeldet ist (und von anderen Usern
      benutzt wird), wird im Verzeichnis des Benutzers nach der CAT-Datei typName
      gesucht und nach der Beschreibung die Komponente erstellt.
  */
  Component giveComp (String typName) {
    Component c = null;
    if (user == null || typName == null ) return null;
    //c = (Component) components.get(typName);
    c = this.getComponent(typName);
    if (c == null) {
      // noch nicht geladen, also jetzt aber;
      String dir = DCATDIR;
      try {
         // Im Konstruktor wird die Datei "typName".CAT gesucht ! Der Typname
         // entspricht aber dann dem Namen der s_component...
      	c = new Component(dir, typName, this);
      } catch (Exception e) {
   	   message("error", 1, "Server.giveComp: " +e.toString());
         return null;
      }
      if (c == null || c.getTypeName().equals("##")) {
      	message("error", 1, "Komponente " + typName +  " existiert noch nicht.");
      	return null;
      }
      //c.setName(instanzName);
      addComponent(c);
      message("comp", 2, "Komponente  vom Typ " + typName + " wurde  angelegt.");
    } else
            message("comp", 2, "Bestehende Komponente vom Typ " + typName + " wurde gefunden.");

    return c;
  }

  /** Erzeugt neues Component Objekt. 
      Der Typ kann eine CAT Datei oder eine FlexiBean (im EvolveServerURL-Pfad) 
      referenzieren. 
  */
  public IR_Component getNewComponent(String Typ) {
      boolean weiter = true;
      try {
        if (is_cat(Typ)) {
            return new Component(DCATDIR, Typ, this);
        } else {
            Component c = new Component(Typ, this);
            Class klasse = null;
            try {
              klasse  = Class.forName(Typ, true, ucl);
            } catch (Exception ex) {
               return c;
            }
            c.setAtomic(true);
            // now read all Ports and Params
            if (checkVisibility(klasse)) {
                  c.addParam("X-Pos", "int", null);
                  c.addParam("Y-Pos", "int", null);
                  c.addParam("Width", "int", null);
                  c.addParam("Height", "int", null);
            }
            while (weiter) { // for all Superclasses

                Method m[] = klasse.getDeclaredMethods();
                for(int i=0; i<m.length;i++){ 
                  String methodName = m[i].getName();
                  
                  if (methodName.startsWith("getPort")  && methodName.length() > 7) {
                      String name = methodName.substring(7);
                      if (name.indexOf("_")!=-1) {
                          String type = name.substring(0, name.indexOf("_"));
                          name = name.substring(name.indexOf("_")+1);
                          c.addPort(name, type, "required", "event");
                      }
                  } else if (methodName.startsWith("add") && methodName.length() > 3) {
                      String name =  methodName.substring(3);
                      if (hasFunction(klasse, "remove"+name) ) {
                          if (name.indexOf("_")!=-1) {
                              String type = name.substring(0, name.indexOf("_"));
                              name = name.substring(name.indexOf("_")+1);
                              c.addPort(name, type, "provided", "event");
                          }
                      }
                  } else if (methodName.startsWith("setShared")  && methodName.length() > 9) {
                      String name =  methodName.substring(9);
                      if (hasFunction(klasse, "forgetShared" + name) ) {
                          if (name.indexOf("_")!=-1) {
                                String type = name.substring(0, name.indexOf("_"));
                                name = name.substring(name.indexOf("_")+1);
                                c.addPort(name, type, "required", "sharedObject");
                          }
                      }
                  } else if (methodName.startsWith("getShared")  && methodName.length() > 9) {
                      String name = methodName.substring(9);
                      if (name.indexOf("_")!=-1) {
                            String type = name.substring(0, name.indexOf("_"));
                            name = name.substring(name.indexOf("_")+1);
                            c.addPort(name, type, "provided", "sharedObject");
                      }
                  } else if (methodName.startsWith("set")) {
                      String name =  methodName.substring(3);
                      if (hasFunction(klasse, "get"+name) ) {
                            // Port gefunden
                            String typ = m[i].getParameterTypes()[0].getName();
                            if (typ.indexOf('.')!=-1) {
                                typ = typ.substring(typ.lastIndexOf('.')+1);
                            }
                            if (typ!= null) {
                            
                               if (typ.equalsIgnoreCase("String") || 
                                   typ.equalsIgnoreCase("int") ||                                
                                   typ.equalsIgnoreCase("boolean") || 
                                   typ.equalsIgnoreCase("long") || 
                                   typ.equalsIgnoreCase("char") || 
                                   typ.equalsIgnoreCase("float") || 
                                   typ.equalsIgnoreCase("null") || 
                                   typ.equalsIgnoreCase("Integer") || 
                                   typ.equalsIgnoreCase("char") || 
                                   typ.equalsIgnoreCase("byte") || 
                                   typ.equalsIgnoreCase("short") || 
                                   typ.equalsIgnoreCase("double")  ){
                                  c.addParam(name, typ, null);
                               }
                            }
                      }
                  } // end if parameter (set)
                }

                if (klasse.getName().equals("java.lang.Object")) weiter = false;
                klasse = klasse.getSuperclass() ;
                if (klasse==null) weiter = false;
                
            }
            
            return c;
        }
      } catch (Exception e) {
        message("error", 1, "Server.getNewComponent " + e.toString());
        return null;
    }
  }

  /** Untersuche Klasse auf sichtbarkeit. Ist die Klasse von java.awt.Component abgeleitet?
   */
   private boolean checkVisibility(Class c) {
      java.awt.Component comp = (java.awt.Component) new java.awt.Label();
      //message("proxy", 4, "Class = " + c.getName());
      if (c==null) return false;
      if (c.isInstance(comp)) return true;
      Class superclass = c.getSuperclass();
      while (superclass != null) {
         //message("proxy", 4, "Class = " + superclass.getName());
         if (superclass.getName().equals("java.lang.Object")) return false;
         if (superclass.isInstance(comp)) return true;
         superclass = superclass.getSuperclass();
      }
      return false;
   }

  boolean hasFunction(Class klasse, String s) {
      if (s==null) return false;
      Method m[] = klasse.getDeclaredMethods();
      for(int i=0; i<m.length;i++){ 
          String methodName = m[i].getName();
          if (s.equals(methodName)) return true;
      }
      return false;
 };

  /** Gibt ein DistComponent Objekt zurck, das die Komponente mit dem Exemplarnamen und TypNamen 
      enthlt und von dem Benutzer 'user' am Client 'id' instanziiert wurde. Dabei kann der Client
      null sein und der Benutzer kann  "server" sein.
      Falls diese Komponente nicht am Server angemeldet ist, wird  sie erstellt.
      @param user Der Benutzername
      @param ID  Ein String um verschiedene Exemplare der Anwendung des Benutzers unterscheiden zu knnen. Ist ID null, wird das erste Exemplar genommen.
  */
  DistComponent giveDistComp(String instanzName, String typName, String user, String ID) {
    if (typName==null || instanzName == null || user == null) return null;
    DistComponent c = null;
    String distCompName = getDistCompName(instanzName, typName, user, ID);
    //message("dcomp",6,"Finde gerade "+distCompName+ " mit Komp. " +instanzName+
    //         " vom Typ " + typName+" fr user "+user);
    c = getLocalDistComponent(user, ID, distCompName);
    if (c!=null) {
          message("dcomp", 2, "Bestehende Komponente " + instanzName+ " vom Typ " + 
                     typName + " fr "  + user + ":" + ID + " wurde gefunden. DCAT-Name = "
                     + distCompName );
          return c;
    }
    if (!user.equals("server")) return null;
    c = startServerComp(instanzName, typName); // starten und anmelden..; 
    return c; // null bei Fehler;
  }

   /** Sucht das DistComponent Objekt, das (user,ID) gehrt 
       und die Komponente InstanzName::typName als lokale Komponente enthlt.
      @param user Der Benutzername
      @param ID  Ein String um verschiedene Exemplare der Anwendung des Benutzers unterscheiden zu knnen. Ist ID null, wird das erste Exemplar genommen.
   */
   String getDistCompName(String instanzName, String typName, String user, String ID) {
    DistComponent c = null;
    try {
       for(Enumeration e = distComps.elements() ; e.hasMoreElements(); ) {
         c = (DistComponent) e.nextElement();
         if ( (c.getUserName().equals(user)) &&
      	   (c.getLocalName().equals(instanzName)) &&
        	   (c.getComponent().getTypeName().equals(typName)) )
            	if (ID==null || c.getId().equals(ID)) return c.getName();
       }
    } catch (RemoteException e) {
      message("error", 1, "Server.getDistCompName " + e.toString());
    }
    return null;
}

  synchronized DistComponent startServerComp(String instanzName, String typName) {
    if (instanzName==null || typName==null) return null;
    DistComponent dc = null;
    Component     c  = null;
    IR_Component  rc = null;
    Proxy      proxy = null;
    try {
       c = giveComp(typName);
       if (c==null) {
           message("error", 1 , "CAT-Datei " + typName + ".CAT fr Server konnte nicht erzeugt werden.");
           return null;
       }
       dc = new DistComponent("server", null, instanzName+"_"+typName, instanzName, c, this);
    } catch (Exception e) {
      message("error", 1, "startServerComp: Fehler beim Bearbeiten von " + instanzName+"::"+typName + ": "
               + e.toString());
      return null;
    }
    if ((rc = dc.getComponent()) == null) {
      message("error", 1 , "DCAT " + instanzName+"::"+typName + " konnte nicht angelegt werden.");
      //return null;
    } else {
        try {
          proxy = new Proxy(rc, this, false);
          dc.setProxy(proxy);
       } catch (Exception ex) {
         message("error", 1 , "Proxy fr " + instanzName+"::"+typName + " konnte nicht angelegt werden." +ex.toString());
         return null;      
       }
    }
    distComps.addElement(dc);
    message("dcomp", 2 , "DCAT " + instanzName+"::"+typName + " fr server wurde angelegt.");
    return dc;
  }
   
      
  // DistComponent HANDLING *******************************************************++

    public String getDCATName(String user, String Anwendungsname) {
        if (user==null || Anwendungsname==null  ) return null;        
        User u = getUser(user);
        if (u==null) {
              message("error", 1, "Server.getDCATName: Benutzername "+user+" ist nicht in Evolve.ini eingetragen.");
              return null;
        }
        Hashtable appl = u.getApplications();
        return  (String) appl.get(Anwendungsname);
    }


  /** Hier meldet der Benutzer seine DistComponent an, indem er den Namen der
      Anwendung bergibt, dass ist der Name, der in der Benutzerverwaltung dem
      Anwendungstyp (DCAT-Dateiname) zugeordnet ist.(incl. Benutzername und Rechnername).<br>
      Dann wird DistComponent incl. lokaler Komponentenbeschreibung aufgebaut.
      @param user Der Benutzername
      @param ID  Ein String um verschiedene Exemplare der Anwendung des Benutzers unterscheiden zu knnen. Ist ID null, wird das erste Exemplar genommen.
  */
  synchronized public IR_DistComponent getNewApplication(String user, String ID, String Anwendungsname) {
    if (user==null || Anwendungsname==null || ID == null ) return null;        
    User u = getUser(user);
    if (u==null) {
          message("error", 1, "Server.getNewApplication: Benutzername "+user+" ist nicht in Evolve.ini eingetragen.");
          return null;
    }
    Hashtable appl = u.getApplications();
    String DCATName = (String) appl.get(Anwendungsname);
    if (DCATName == null) {
          message("error", 1, "Server.getNewApplication: Anwendung "+Anwendungsname+" ist nicht in dem Benutzer "+user+" zugeordent.");
          return null;
    }
    return getNewDistComp(user, ID, DCATName);
  }


  /** Hier meldet der Benutzer seine DistComponent an, indem er den Namen der
      DCAT Datei bergibt (incl. Benutzername und Rechnername).<br>
      Dann wird DistComponent incl. lokaler Komponentenbeschreibung aufgebaut.
      @param user Der Benutzername
      @param ID  Ein String um verschiedene Exemplare der Anwendung des Benutzers unterscheiden zu knnen. Ist ID null, wird das erste Exemplar genommen.
  */
  synchronized public IR_DistComponent getNewDistComp(String user, String ID, String DCATname) {
    if (user==null || DCATname==null || ID == null ) return null;

    DistComponent c = getLocalDistComponent(user, ID, DCATname);
    String DCAT = null;
    if (c!=null) {
         message("error", 2, "DCAT " + DCATname + " fr " + user + ":" + ID + 
                  " existiert bereits.");
         return c;
    }
    try {
       DCAT = getFile(DCATname+".DCAT");
       if (DCAT==null) {
         message("error", 1 , "DCAT-Datei " + DCATname + " konnte nicht geladen werden.");
         return null;
       }
       c = new DistComponent(user, ID, DCATname, DCAT, this);
    } catch (Exception e) {
      message("error", 1, "getNewDistComp: " + e.toString());
      return null;
    }
    if (c.getComponent()==null) {
      message("error", 1 , "DCAT " + DCATname + " konnte nicht angelegt werden.");
      return null;
    }
    distComps.addElement(c);
    message("dcomp", 2 , "DCAT " + DCATname + " fr " + user +  " wurde angelegt.");
    showServerObjects();
    return c;
  }
   
    
  // findet nur erstes Vorkommen!;
  /** Gibt am Server angemeldetes DistComponent Objekt zurck.
      @param user Der Benutzername
      @param ID  Ein String um verschiedene Exemplare der Anwendung des Benutzers unterscheiden zu knnen. Ist ID null, wird das erste Exemplar genommen.
      
      @param compName Der Name des DCAT Files.
  */
  public IR_DistComponent getDistComponent(String user, String ID, String compName)  {
    DistComponent c = null;
    for(Enumeration e = distComps.elements() ; e.hasMoreElements(); ) {
         c = (DistComponent) e.nextElement();
         if ( (c.getUserName().equals(user)) &&
   	      (c.getName().equals(compName)) )
         	   if (ID==null || ID.equals(c.getId())) return c;
    }
    return null;
  }

    // findet nur erstes Vorkommen!;
   /** Gibt am Server angemeldetes DistComponent Objekt zurck.
      @param user Der Benutzername
      @param ID  Ein String um verschiedene Exemplare der Anwendung des Benutzers unterscheiden zu knnen. Ist ID null, wird das erste Exemplar genommen.
      
      @param compName Name des DistComponent Objektes (DCAT-Filename)
  */
  DistComponent getLocalDistComponent(String user, String ID, String compName)  {
    DistComponent c = null;
    for(Enumeration e = distComps.elements() ; e.hasMoreElements(); ) {
         c = (DistComponent) e.nextElement();
         if ( (c.getUserName().equals(user)) &&
      	   (c.getName().equals(compName)) )
         	   if (ID==null || ID.equals(c.getId())) return c;
    }
    return null;
  }

  public void removeApplication(String user, String ID, String Anwendungsname) {
    if (user==null || Anwendungsname==null || ID == null ) return ;        
    User u = getUser(user);
    if (u==null) {
          message("error", 1, "Server.removeApplication: Benutzername "+user+" ist nicht in Evolve.ini eingetragen.");
          return ;
    }
    Hashtable appl = u.getApplications();
    String DCATName = (String) appl.get(Anwendungsname);
    if (DCATName == null) {
          message("error", 1, "Server.removeApplication: Anwendung "+Anwendungsname+" ist nicht in dem Benutzer "+user+" zugeordent.");
          return ;
    }
     removeDistComponent(user, ID, DCATName);
  }

  /** Entfernt am Server angemeldetes DistComponent Objekt.
      @param user Der Benutzername
      @param ID  Ein String um verschiedene Exemplare der Anwendung des Benutzers unterscheiden zu knnen. Ist ID null, wird das erste Exemplar genommen.
  */
  public void removeDistComponent(String user, String ID, String compName) {
    DistComponent c = getLocalDistComponent(user, ID, compName);
    if (c != null) {
      c.kill();
    } else
        message("error", 3, "DistComponent mit " + compName + " fr " + user + ":" + ID +
                            " zum Entfernen nicht gefunden.");
    distComps.removeElement(c);
    // hier muss man sich vielleicht noch um die Komponente kuemmern, die vielleicht;
    // nicht mehr gebraucht wird. Bisher bleibt sie im Speicher !!!;
    // Wird sie in Datei geaendert, merkt der Server dies nicht... -> reload...;
    
    message("dcomp", 3, "DistComponent " + compName + " fr " + user + ":" + ID+ 
                        " wurde entfernt.");
    showServerObjects();
  }

  /**  Gibt alle am Server angemeldeten DistComponent Objekte zurck.
  */
  public Vector getAllDistComps() {
    return distComps;
  } 
      

  // END DistComponent Handling ******************************************************************************************;

//  public String getFile(String user, String file, String ext) {
  /** Gibt Datei file in einm String zurck.
      Die Datei mu sich im Verzeichnis Evolve.DCATDir (in Evolve.ini
      angegeben) befinden.
      @param file Dateiname mit Endung
      @return Der Inhalt der Datei in einem String.
  */
  public String getFile(String file) {
    if (user==null || file==null) return null;
    String dir = DCATDIR;
    String string = null;
    try {
      string = (new MyFile()).getFile(dir, file);
    } catch (Exception e) {
      message("error", 1, "Konnte Datei " + dir + file +" nicht lesen." +e);
      return null;
    }
    if (string == null) {
      message("error", 1, "Konnte Datei " + dir + file + " nicht lesen.");
      return null;
    }
    return string;
  } // END getFile;
    

  // Component HANDLING *****************************************************************************
  /** Neue Komponente am Server unter ihrem Namen anmelden. 
      Ggf. wird alte mit dem Namen ersetzt --> <b>Component Objekte sollen
      eindeutigen Namen haben!</b> Vom Benutzer ihmspezifisch vernderte
      Komponenten sollten vielleicht im Namen einen Hinweis auf den Benutzer
      haben. (CAT-File)
  */
  void addComponent(Component c) {
    if (c==null) return;
    components.put(c.getTypeName(), c);
    message("comp", 4, "Komponente  vom Typ " + c.getTypeName() +
            " wurde angemeldet.");
  }

   /** Erhalte am Server angemeldete Komponente zurck. Name ist eindeutig.
   */
  Component getComponent(String typName) {
    return (Component) components.get(typName);
  }

   /** Entferne am Server angemeldete Komponente
   */
  void removeComponent(Component c) {
    if (c==null) return;
    components.remove(c.getTypeName()); //Pfui, wenn Componente ihren Namen aendert...;
    message("comp", 4, "Komponente vom Typ " + c.getName() + " wurde abgemeldet.");
  }
  // END Component HANDLING *****************************************************************************  
  
    
    /** Fr Benutzer und Schlssel den Wert aus der Evolve.ini zurck.
        In Evolve.ini steht dann user.TAG=xxx
    */
  protected String getProp(String user, String TAG) {
    // hier fehlt noch ein resourcen schonende Property-Verwaltung;
    return evolveIni.getProp(user + "." + TAG);
  }
  
  
  /** Empfngt von allen Serverklassen Nachrichten und zeigt sie am Server an.
      Noch zu implementieren sind Messages, falls Operationen fehlschlagen.
      Bisher wird eine Meldung an den Server gegeben und null oder nichts
      zurckgegeben.
  */
  public void message(String type, int importance, String text) {
    //    System.out.println(type + ": " + text);
    if (type.equals("server")) {
      evolveMsg.add(type, importance, text);
    } else {
       nr++;
       evolveMsg.add(type, importance, text+"("+nr+")");
    }
  }

  /** Speichert die Komponente im Verzeichnis directory.
      Die Datei wird unter dem Typnamen der Komponente gespeichert. Einer bestehende 
      Datei wird die Endung ".bak" angehngt.
      @return true, falls das Speichern erfolgreich war.
  */
  public boolean saveComponent(IR_Component c, String directory) {
      if (c == null || directory == null) return false;
      try {
         String CAT = c.toText();
         MyFile f = new MyFile();
         f.writeFileWithBak(directory, c.getTypeName()+".CAT", "bak", CAT);
         message("dcomp", 4, "CAT Datei " + c.getTypeName() + " wurde in " + directory + 
                  " gespeichert.");
      } catch (RemoteException e) {
         message("error", 1, "Server.saveComponent: Netzwerk unten. "+e.toString());
         return false;
      } catch (Exception e) {
         message("error", 1, "Server.saveComponent: "+e.toString());
         return false;
      } 
      return true;
   }

  /** Speichert die DCAT Datei der verteilten Komponente im Verzeichnis directory.
      Die Datei wird unter dem Namen der Komponente gespeichert. Einer bestehende 
      Datei wird die Endung ".bak" angehngt.
      @param directory darf null sein, dann wird in Evolve.DCATDir gespeichert
      @return true, falls das Speichern erfolgreich war.
  */
  public boolean saveDCAT(IR_DistComponent c, String directory) {
      if (c == null ) return false;
      if (directory == null) { directory = DCATDIR;
      }
      try {
         String DCAT = c.writeDistComponent();
         MyFile f = new MyFile();
         f.writeFileWithBak(directory, c.getName()+".DCAT", "bak", DCAT);
         message("dcomp", 4, "DCAT Datei " + c.getName() + " wurde in " + directory + 
                  " gespeichert.");
      } catch (RemoteException e) {
         message("error", 1, "Server.saveDCAT: Netzwerk unten. "+e.toString());
         return false;
      } catch (Exception e) {
         message("error", 1, "Server.saveDCAT: "+e.toString());
         return false;
      } 
      return true;
   }


  /** Speichert die RBD Datei im Verzeichnis directory.
      Die Datei wird unter dem Namen des Remote Bind Datei gespeichert. Einer bestehende 
      Datei wird die Endung ".bak" angehngt.
      @param directory darf null sein, dann wird in Evolve.DCATDir gespeichert
      @return true, falls das Speichern erfolgreich war.
  */
  public boolean saveRBD(IR_RBD rbd, String directory) {
      if (rbd == null ) return false;
      if (directory == null) { directory = DCATDIR;
      }     try {
         String RBD = rbd.write();
         MyFile f = new MyFile();
         f.writeFileWithBak(directory, rbd.getName()+".RBD", "bak", RBD);
         message("dcomp", 4, "RBD Datei " + rbd.getName() + " wurde in " + directory + 
                  " gespeichert.");
      } catch (RemoteException e) {
         message("error", 1, "Server.saveRBD: Netzwerk unten. "+e.toString());
         return false;
      } catch (Exception e) {
         message("error", 1, "Server.saveRBD: "+e.toString());
         return false;
      } 
      return true;
   }
    
 /** Speichert die verteilte Komponente im Verzeichnis directory.
      Dazu gehrt die DCAT Datei, die CAT Datei der lokalen Komponente und alle
      assoziierten RBD Dateien.
      @param directory darf null sein, dann wird in Evolve.DCATDir gespeichert
      @see #saveDCAT
      @see #saveComponent
      @see #saveRBD
      @return true, falls das Speichern erfolgreich war.
  */
  public boolean saveDistComponent(IR_DistComponent c, String directory) {
      if (c == null ) return false;
      if (directory == null) { directory = DCATDIR;
      }     try {
         IR_RBD rbd = null;
         for (Enumeration e = c.getMyRBDs().elements(); e.hasMoreElements(); ) {
            rbd = (IR_RBD) e.nextElement();
            if (saveRBD(rbd, directory) == false) 
               message("error", 1, "RBD Datei " + rbd.getName() + " konnte nicht gespeichert werden.");
         }
         if (saveDCAT(c, directory) == false) 
               message("error", 1, "DCAT Datei " + c.getName() + " konnte nicht gespeichert werden.");
         if (saveComponent(c.getComponent(), directory) == false) 
               message("error", 1, "CAT Datei " + c.getComponent().getName() + " konnte nicht gespeichert werden.");
     } catch (RemoteException e) {
         message("error", 1, "Server.saveRBD: Netzwerk unten. "+e.toString());
         return false;
     }
     return true;
   }
  
  public java.awt.Container getPanel() { return null; }

  private void showServerObjects() {
      String s = "";
      DistComponent dc = null;
      Component c = null;
      String id = "";
      s += "Root Component-Exemplare (DCAT): \n";
      for (Enumeration e = distComps.elements(); e.hasMoreElements() ; ) {
         dc = (DistComponent) e.nextElement();
         try {
            if ((id=dc.getId())==null) id="";
            s += dc.getUserName()+":" + id  + " " +  dc.getComponent().getTypeName() + 
                 " (" + dc.getLocalName() + ")\n";
         } catch (RemoteException ex) {
            message("error", 1, "Server.showServerObjects: " + ex);
         }
      }
      s += "\nRoot Component-Objekte (CAT): \n";
      for (Enumeration e = components.elements(); e.hasMoreElements() ; ) {
         c = (Component) e.nextElement();
            s += c.getTypeName() +"\n";
      }
      s += "\nClients: \n";
      for (Enumeration e = name_id.elements(); e.hasMoreElements() ; ) {
         s += (String) e.nextElement() + "\n";
      }
      message("server", 1, s);
  }

 public void reset() throws RemoteException {
    components = new Hashtable();
    distComps = new Vector();
    user      = new Hashtable();
    evolveIni = new Prop(PROPDIR, iniFileName);
    ROOTDIR   = evolveIni.getProp("Evolve.RootDir");
    USERDIR   = ROOTDIR + evolveIni.getProp("Evolve.UserDir");
    DCATDIR   = ROOTDIR + evolveIni.getProp("Evolve.DCATDir");
    String d  = evolveIni.getProp("Evolve.Debug");
    if (d!=null && d.equalsIgnoreCase("true")) {
        debug = true;
    }
   Hashtable eIni = getProperties(iniFileName);
   for (Enumeration e = eIni.keys(); e.hasMoreElements(); ) {
      String key = (String) e.nextElement();
      if (key.startsWith("user.")) {
          User u = new User(key.substring(5), USERDIR, (String) eIni.get(key), false, this);
          if (user.put(key.substring(5), u) != null) 
              message("error", 1, "Server: In der Datei " + iniFileName + " existieren zwei Eintrge " + key);
      } else if (key.startsWith("group.")) {
          User u = new User(key.substring(6), USERDIR, (String) eIni.get(key), true, this);
          if (user.put(key.substring(6), u) != null) 
             message("error", 1, "Server: In der Datei " + iniFileName + " existieren zwei Eintrge " + key);
      }
   }

    //name_id = new Vector();
    nr = -1;
    message("dcomp",1,"Alle Anwendungen laufen noch, werden aber bei der nchsten Anfrage neu geladen.");
    showServerObjects();
 }    

  synchronized public String  getID(String user, String ID) {
      int idi = 0;
      boolean ok = false;
      String newID = ID;

      if (ID==null) {
         ID = "";
         newID = "0";
      } else newID = ID;
      if (user == null) return null;
      String dc = null;
      
      while (!ok) {
         ok = true;             
         for(Enumeration e = name_id.elements() ; e.hasMoreElements(); ) {
            dc = (String) e.nextElement();
            if (dc.equals(user + ":"+ newID)) ok = false;
         }
         if (!ok) {
            idi ++;
            newID = ID + idi;
         }
      }
      name_id.add(new String(user + ":" + newID));
      showServerObjects();
      //System.out.println("Vergebe fr " + user + " ID " + newID);
      return newID;
  }
  public URLClassLoader getURLClassLoader(){ return ucl;}

  public void goodByeServer(String name, String ID) {
      if (name_id.remove(new String(name+":"+ID)) == false ) {
        message("error", 1, "Benutzer "+name+":"+ID+" war nicht angemeldet.");
      }
    showServerObjects();
  }

  public String getServerInfo() {
      String s = "";
      DistComponent dc = null;
      Component c = null;
      String id = "";
      s += "Root Component-Exemplare (DCAT): \n";
      for (Enumeration e = distComps.elements(); e.hasMoreElements() ; ) {
         dc = (DistComponent) e.nextElement();
         try {
            if ((id=dc.getId())==null) id="";
            s += dc.getUserName()+":" + id  + " " +  dc.getComponent().getTypeName() + 
                 " (" + dc.getLocalName() + ")\n";
         } catch (RemoteException ex) {
            message("error", 1, "Server.showServerObjects: " + ex);
         }
      }
      s += "\nRoot Component-Objekte (CAT): \n";
      for (Enumeration e = components.elements(); e.hasMoreElements() ; ) {
         c = (Component) e.nextElement();
            s += c.getTypeName() +"\n";
      }
      s += "\nClients: \n";
      for (Enumeration e = name_id.elements(); e.hasMoreElements() ; ) {
         s += (String) e.nextElement() + "\n";
      }
      return s;
  }

 /** Liest alle Eintrge aus der Datei propFileName der Form KEY=VALUE
 */
 private Hashtable getProperties(String propFileName) {
    File f;
    FileInputStream fis = null;
    Hashtable properties = new Hashtable();
    try {
      f = new File(ROOTDIR, propFileName);
      fis = new FileInputStream(f);
      PropertyResourceBundle props = new PropertyResourceBundle(fis);
      Enumeration allProps = props.getKeys();
      for(String entry=null ; allProps.hasMoreElements() ; ) {
          entry =  (String) allProps.nextElement();
          properties.put(entry, props.getString(entry));
      }
      return properties;
    } catch (IOException e) {
      System.err.println("Server.getProperties: IOFehler: " + e);
      return null;
    }
    finally { try { if (fis != null) fis.close(); } catch (IOException e) {}  }
  }

  /** Liest aus einer Ini-Datei die zu startenden DCAT Dateien.
      Dazu werden die Meta-Objekte fr die Anwendung und auch die Proxy-Objekte
      fr die Serverkomponenten aufgebaut und die Serverkomponenten instanziiert.
  */
  private void startFirstDCATs() {
     String  startFile = evolveIni.getProp(startTag);
     if (startFile == null) {
        message("error", 1, "Eintrag in Datei '" + iniFileName + "' fehlt : " + startTag +
                " . Kann initiale DCAT Dateien nicht einlesen.");
        return;
     }
     Enumeration e = getProperties(startFile).keys();
     if (e==null) {
        message("error", 1, "Konnte Datei " + startFile + " nicht lesen" +
                ". Kann initiale DCAT Dateien daher nicht einlesen.");
        return;
     }  
     IR_DistComponent c = null;
     for ( ; e.hasMoreElements() ; ) {
          String dcat = (String) e.nextElement();
          c = getNewDistComp("server", "", dcat);
          distComps.removeElement(c);
          showServerObjects();
     }
  }

  /** Fgt dem Evolve Server einen neuen Benutzer zu. 
      In Evolve.ini wird die Datei eingetragen, in der die Zuordnung Gruppe, DCAT, CAT, CLASS 
      eingetragen wird. Schlsselwort in Evolve.ini ist user.<name>
      @return false Falls der Benutzer schon eingetragen wurde.
  */
  public boolean addUser       (String name) { 
      String  userFile = evolveIni.getProp("user."+name);
      if (userFile == null) return false;
      evolveIni.setProp("user."+name, name + "." + EvolveSuffix);
      userFile = evolveIni.getProp("user."+name);
      if (userFile == null) return false;
      User u = new User(name, USERDIR, userFile, false, this);
      user.put(name, u);
      return true; 
  };

  /** Entfernt einen Benutzer. Dabei wird die Zuordnung in der Evolve.ini entfernt.
  */
  public boolean removeUser    (String name) { 
      String  userFile = evolveIni.getProp("user."+name);
      if (userFile == null) return true;
      evolveIni.forgetProp("user."+name);
      userFile = evolveIni.getProp("user."+name);
      if (userFile == null) return true;
      user.remove(name);
      return false; 
  };

  /** Fgt dem Evolve Server eine neue Benutzergruppe zu. 
      In Evolve.ini wird die Datei eingetragen, in der die Zuordnung DCAT, CAT, CLASS 
      eingetragen wird. Schlsselwort in Evolve.ini ist group.<name>
      @return false Falls die Benutzergruppe schon eingetragen wurde.
  */
  public boolean addUserGroup  (String name) { 
    String  groupFile = evolveIni.getProp("group."+name);
    if (groupFile == null) return false;
    evolveIni.setProp("group."+name, name + "." + EvolveSuffix);
    groupFile = evolveIni.getProp("group."+name);
    if (groupFile == null) return false;
    User u = new User(name, USERDIR, groupFile, true, this);
    user.put(name, u);
    return true; 
  };

  /** Entfernt eine Benutzergruppe. Dabei wird die Zuordnung in der Evolve.ini entfernt.
  */
  public boolean removeUserGroup(String name) { return true; };

  /** Ordnet den Benutzer zur Komponenten-Gruppe zu (CAT). Dabei wird ein Eintrag in die Benutzer.ini Datei hinzfugefgt.
      Sowohl Benutzer als auch die Gruppe mssen in der Evolve.ini eingetragen sein mit einem Verweis
      auf ihre eigenen Ini-Dateien.
  */
  public boolean addToGroupForComp     (String userName, String groupName) { 
      if (!user.containsKey(userName)) {
          message("error", 1, "Server.addToGroupForComp: User " + userName + " ist nicht in Evolve.ini eingetragen.");
          return false;          
      }
      if (!user.containsKey(groupName)) {
          message("error", 1, "Server.addToGroupForComp: Gruppe " + groupName + " ist nicht in Evolve.ini eingetragen.");
          return false;          
      }
      User u = (User) user.get(userName);
      u.addGroupForComp(groupName);
      return true; 
  };

  /** Ordnet den Benutzer zur Anwendungs-Gruppe zu (DCAT). Dabei wird ein Eintrag in die Benutzer.ini Datei hinzfugefgt.
      Sowohl Benutzer als auch die Gruppe mssen in der Evolve.ini eingetragen sein mit einem Verweis
      auf ihre eigenen Ini-Dateien.
  */
  public boolean addToGroupForAppl    (String userName, String groupName) { 
      if (!user.containsKey(userName)) {
          message("error", 1, "Server.addToGroupForApl: User " + userName + " ist nicht in Evolve.ini eingetragen.");
          return false;          
      }
      if (!user.containsKey(groupName)) {
          message("error", 1, "Server.addToGroupForApl: Gruppe " + groupName + " ist nicht in Evolve.ini eingetragen.");
          return false;          
      }
      User u = (User) user.get(userName);
      u.addGroupForApl(groupName);
      return true; 
  };

  /** Lscht die Gruppenzuordung fr den Benutzer (CAT).
  */
  public boolean removeFromGroupForComp(String userName, String groupName) { 
      if (!user.containsKey(userName)) {
          message("error", 1, "Server.removeFromGroupForComp: User " + userName + " ist nicht in Evolve.ini eingetragen.");
          return false;          
      }
      if (!user.containsKey(groupName)) {
          message("error", 1, "Server.removeFromGroupForComp: Gruppe " + groupName + " ist nicht in Evolve.ini eingetragen.");
          return false;          
      }
      User u = (User) user.get(userName);
      u.removeGroupForComp(groupName);
      return true; 
  };

  /** Lscht die Gruppenzuordung fr den Benutzer (DCAT).
  */
  public boolean removeFromGroupForAppl(String userName, String groupName) { 
        if (!user.containsKey(userName)) {
          message("error", 1, "Server.removeFromGroupForApl: User " + userName + " ist nicht in Evolve.ini eingetragen.");
          return false;          
      }
      if (!user.containsKey(groupName)) {
          message("error", 1, "Server.removeFromGroupForApl: Gruppe " + groupName + " ist nicht in Evolve.ini eingetragen.");
          return false;          
      }
      User u = (User) user.get(userName);
      u.removeGroupForApl(groupName);
      return true; 
  };

  /** Gibt die Namen der Gruppen zurck, in denen der Benutzer eingetragen ist.
  */
  public Vector  whichGroupsForComp    (String userName) { 
     if (!user.containsKey(userName)) {
          message("error", 1, "Server.removeFromGroupForApl: User " + userName + " ist nicht in Evolve.ini eingetragen.");
          return null;          
      }
      User u = (User) user.get(userName);
      return u.whichGroups("cat"); 
  };

  /** Gibt die Namen der Gruppen zurck, in denen der Benutzer eingetragen ist.
  */
  public Vector  whichGroupsForAppl    (String userName) { 
     if (!user.containsKey(userName)) {
          message("error", 1, "Server.removeFromGroupForApl: User " + userName + " ist nicht in Evolve.ini eingetragen.");
          return null;          
      }
      User u = (User) user.get(userName);
      return u.whichGroups("dcat"); 
  };

  /** Fgt in die Liste des Benutzers oder der Gruppe eine (erlaubte) Anwendung hinzu.
      Einem Benutzer sind Anwendungen und Gruppen zugeordnet, wobei den Gruppen wiederum 
      Anwendungen und Gruppen zugeordnet sind. Die Eintragungen fr den Benutzer berschreiben allerdings
      die Eintragungen fr eine dem Benutzer zugeordnete Gruppe. <br>
      Wenn ein Gruppeneintrag e also gendert wird, ist die nderung bei allen Benutzern, denen die Gruppe 
      zugeordnet ist und denen nicht e direkt zugeordnet ist aktiv.
  */
  public boolean addApplication(String user_or_group, String applName, String applType) { 
      User u = (User) user.get(user_or_group);
      if (u==null) {
          message("error", 1, "Server.addApplication: User " + user_or_group + " ist nicht in Evolve.ini eingetragen.");
          return false;
      }
      return u.addApplication(applName, applType);
  };

  /** Entfernt aus der Liste des Benutzers eine erlaubte Anwendung.
  */
  public boolean removeApplication(String user_or_group,  String applType) { 
      User u = (User) user.get(user_or_group);
      if (u==null) {
          message("error", 1, "Server.addApplication: User " + user_or_group + " ist nicht in Evolve.ini eingetragen.");
          return false;
      }
      return u.removeApplication(applType);
  };

  /** Gibt eine Hashtable mit Anwendungen fr den Benutzer oder die Gruppe zurck.
      Schlssel ist die Beschreibung (Name) der Anwendung, Wert ist der Typ (DCAT-Filename) der Anwendung.
  */
  public Hashtable getApplications(String user_or_group) { 
      User u = (User) user.get(user_or_group);
      if (u==null) return null;
      return u.getApplications();
  };

  /**  Gibt Vector von Strings mit Benutzernamen zurck, die die Anwendung verwenden drfen.
  */
  public Vector  getUsersOfAppl(String applTyp) { 
    Vector v = new Vector();
    for (Enumeration e = user.elements(); e.hasMoreElements(); ) {
        User u = (User) e.nextElement();
        if (evolveIni.getProp("user." + u.name) != null) {
            if (u.hasAppl(applTyp)) v.add(u.name);
        }
    }
    return v;
  };

  /**  Gibt Vector von Strings mit Gruppennamen zurck, die die Anwendung verwenden drfen.
  */
   public Vector  getGroupsOfAppl(String applTyp) { 
    Vector v = new Vector();
    for (Enumeration e = user.elements(); e.hasMoreElements(); ) {
        User u = (User) e.nextElement();
        if (evolveIni.getProp("group." + u.name) != null) {
            if (u.hasAppl(applTyp)) v.add(u.name);
        }
    }
    return v;
  }

  /** ndert den Typ einer Anwendung fr den Benutzer oder die Gruppe.
  */
  public boolean  changeApplication(String user_or_group, String applType, String newAplType) { 
        User u = (User) user.get(user_or_group);
        if (u==null) {
            return false;
        }
        return u.changeApplication(applType, newAplType);
  };

  /** ndert den Typ einer Komponente fr den Benutzer oder die Gruppe.
  */
  public boolean   changeComponent(String user_or_group, String compType, String newCompType) {
        User u = (User) user.get(user_or_group);
        if (u==null) {
            return false;
        }
        return u.changeComponent(compType, newCompType);
  };

  /** Fgt in die Liste des Benutzers oder der Gruppe eine (erlaubte) Komponente hinzu.
  */
  public boolean addComponent(String user_or_group, String componentName, String componentTyp) { 
      User u = (User) user.get(user_or_group);
      if (u==null) {
          message("error", 1, "Server.addComponent: User " + user_or_group + " ist nicht in Evolve.ini eingetragen.");
          return false;
      }
      return u.addComponent(componentName, componentTyp);
  };

  /** Entfernt aus der Liste des Benutzers eine erlaubte Komponente.
  */
  public boolean removeComponent(String user_or_group, String componentTyp) { 
      User u = (User) user.get(user_or_group);
      if (u==null) {
          message("error", 1, "Server.removeComponent: User " + user_or_group + " ist nicht in Evolve.ini eingetragen.");
          return false;
      }
      return u.removeComponent(componentTyp);
  };

  /** Gibt eine Hashtable mit Komponenten fr den Benutzer oder die Gruppe zurck.
      Schlssel ist die Beschreibung (Name) der Komponente, Wert ist der Typ (CAT-Filename oder CLASSNAME) der Komponente.
  */
  public Hashtable getComponents(String user_or_group) {
      User u = (User) user.get(user_or_group);
      if (u==null) {
          message("error", 1, "Server.getComponents: User " + user_or_group + " ist nicht in Evolve.ini eingetragen.");
          return null;
      }
      return u.getComponents();
  };

  /**  Gibt Vector von Strings mit Benutzernamen zurck, die die Komponente verwenden drfen.
  */
  public Vector  getUsersOfComp(String compTyp) { 
    Vector v = new Vector();
    for (Enumeration e = user.elements(); e.hasMoreElements(); ) {
        User u = (User) e.nextElement();
        if (evolveIni.getProp("user." + u.name) != null) {
            if (u.hasComp(compTyp)) v.add(u.name);
        }
    }
    return v;
  };

  /**  Gibt Vector von Strings mit Gruppennamen zurck, die die Komponente verwenden drfen.
  */
   public Vector  getGroupsOfComp(String compTyp) { 
    Vector v = new Vector();
    for (Enumeration e = user.elements(); e.hasMoreElements(); ) {
        User u = (User) e.nextElement();
        if (evolveIni.getProp("group." + u.name) != null) {
            if (u.hasComp(compTyp)) v.add(u.name);
        }
    }
    return v;
  }

 
 // public void addToEvolveServerURL(String ClassPath) { return; };

  User getUser(String userName) { 
      User u = (User) user.get(userName); 
      return u;
  }

  /** Gibt Infos ber alle Benutzer aus.
  */
  public void showUsers() {
    for (Enumeration e = user.elements(); e.hasMoreElements(); ) {
      User u = (User) e.nextElement();
      if (evolveIni.getProp("user." + u.name) != null) {
            System.out.println("--------------------------------------------");
            System.out.println(u.name + ":\n" + u.toTxt());
            Hashtable appl = getApplications(u.name);
            if (appl!= null) {
                System.out.println("Applications");
                for (Enumeration e2 = appl.keys() ; e2.hasMoreElements(); ) {
                    String name = (String) e2.nextElement();
                    String typ  = (String) appl.get(name);
                    System.out.println("Typ " + typ + " ist " + name);
                }
            }
            Hashtable comps = getComponents(u.name);
            if (comps!= null) {
                System.out.println("\nComponents");
                for (Enumeration e2 = comps.keys() ; e2.hasMoreElements(); ) {
                    String name = (String) e2.nextElement();
                    String typ  = (String) comps.get(name);
                    System.out.println("Typ " + typ + " ist " + name);
                }
            }
            Vector groups = whichGroupsForComp(u.name);
            if (groups!= null) {
                System.out.println("\nCAT-Groups");
                for (Enumeration e3 = groups.elements() ; e3.hasMoreElements(); ) {
                    String name = (String) e3.nextElement();
                    System.out.println(name);
                }
            }
            Vector groups2 = whichGroupsForAppl(u.name);
            if (groups2!= null) {
                System.out.println("\nDCAT-Groups");
                for (Enumeration e4 = groups2.elements() ; e4.hasMoreElements(); ) {
                    String name = (String) e4.nextElement();
                    System.out.println(name);
                }
            }

          }
    } // end if is user
    System.out.println("--------------------------------------------");
  }

  public boolean is_cat (String typ) {
      File f = new File(DCATDIR, typ+".cat");
      if (f.exists()) return true;
      return false;
  }

  public boolean is_dcat (String typ) {
      File f = new File(DCATDIR, typ+".dcat");
      if (f.exists()) return true;
      return false;
  }

  public String getUser() { return "server";}
  public String getID() { return "";}

  /** ndert den Geltungsbereich einer Anpassung.
      Fr die Liste der Benutzer wird die Meta Struktur der Anwendung vom Typ oldDCATName kopiert. Der neue Anwenungstyp
      heit newDCATName, der Typ der neuen lokalen Komponente heit newCATName.
      Dazu wird ein neues Component Objekt angelegt und als CAT Datei gespeichert.
      Dieses Objekt wird bei allen in Frage kommenden DistComponent Objekten angemeldet. Dazu muss allerdings
      mindestens ein DistComponent Objekt instanziiert sein.<br>
      Ausserdem wird die neuen Zuordnung in der Benutzerverwaltung eingetragen.
  */
  synchronized public void  changeScope(String oldDCATName, String newDCATName, String newCATName, Vector users) {
        // existiert newCATName schon?
        if (is_cat(newCATName)) {
            message("error", 1, "Server.changeScope: Der neue Komponententyp " + newCATName +" existiert bereits. Kann Geltungsbereich daher nicht ndern.");
            return;
        } else {
            Class klasse = null;
            try {
              klasse  = Class.forName(newCATName, true, ucl);
            } catch (Exception ex) {;}
            if (klasse!= null) {
                message("error", 1, "Server.changeScope: Der neue Komponententyp " + newCATName +" existiert bereits. Kann Geltungsbereich daher nicht ndern.");
                return;
            }
        }
        // existiert newDCAT schon?
        if (is_dcat(newDCATName)) {
                message("error", 1, "Server.changeScope: Der neue DistComponent-Typ " + newDCATName +" existiert bereits. Kann Geltungsbereich daher nicht ndern.");
                return;
        }
        // DistComponents und Component Objekte auslesen
        DistComponent dc = null;
        Vector dcat_old = new Vector();
        Component c_old = null;
        for(Enumeration e1 = distComps.elements() ; e1.hasMoreElements(); ) {
           dc = (DistComponent) e1.nextElement();
           if  (dc.getName().equals(oldDCATName) && users.contains(dc.getUserName())) {
                dcat_old.addElement(dc);
           }
        }
        if (dc == null) {
                message("error", 1, "Server.changeScope: Es ist kein DistComponent Objekt vom Typ " + oldDCATName +" getartet.  Kann Geltungsbereich daher nicht ndern.");
                return;
        }

        c_old  = dc.local_getComponent();
        // Neues Component Objekt erzeugen
        Component c_new = c_old.local_copy(newCATName);        
        addComponent(c_new);
        // DCATs updaten
        for(Enumeration e2 = dcat_old.elements() ; e2.hasMoreElements(); ) {
            dc = (DistComponent) e2.nextElement();
            dc.setName(newDCATName);
            dc.setComponent(c_new);
        }
        // Dateien speichern (HIER WIRD NOCH DIE ALTE RBD DATEI VERWENDET)
        saveComponent(c_new, DCATDIR);
        saveDCAT(dc, DCATDIR);
        // Proxies umbiegen
        c_old.changeScope(users, c_new);
        // In Benutzerverwaltung umbiegen
        String u = null;
        User   us = null;
        for (Enumeration e = users.elements(); e.hasMoreElements(); ) {
              u = (String) e.nextElement();
              us = getUser(u);
              if (us != null) {
                  if (us.hasAppl(oldDCATName)) {
                      us.changeApplication(oldDCATName, newDCATName);
                  }
              }
              else {
                  message("error", 1, "Server.changeScope: Benutzer " + u + " ist nicht bekannt.");
              }
        }
       // fertig
       showServerObjects();
      
  }

  /** Anmelden eines Eventlisteners fr Anpassungsoperationen.
  */
  synchronized public void addTailorActionListener(TailorActionListenerInterface t) {
      tailor_listeners.add(t);
  }

  /** Abmelden eines Eventlisteners fr Anpassungsoperationen.
  */
  synchronized public void removeTailorActionListener(TailorActionListenerInterface t) {
      tailor_listeners.remove(t);
  }

  /** Den TailorActionListenern wird ein Anpassungsbefehl bergeben.
      @param comp IR_Component 
      @param command Der Befehlsname aus dem Component Interface
      @param argues Die Argumente des Befehls aus dem Component Interface
  */
  synchronized public void TailorActionPerformed(Object comp, String command, Object[] argues) {
      Vector v =  (Vector) tailor_listeners.clone();
      for (Enumeration e = v.elements(); e.hasMoreElements(); ) {
           TailorActionListenerInterface t = (TailorActionListenerInterface) e.nextElement();
           t.TailorActionPerformed(comp, command, argues);
      }
  }

  URL[] readClassPath(String url) {
    Vector v = new Vector();
    StringTokenizer token = new StringTokenizer(url, ";");    
    String value;
    int i = 0;
    // First read all components
    while (token.hasMoreTokens()) {
      if (!token.hasMoreTokens()) continue;
      value = token.nextToken();
      v.add(value);
      i++;
    }
    URL[] u = new URL[i];
    i=0;
    for (Enumeration e = v.elements(); e.hasMoreElements(); ) {
        try {
            u[i] = new URL((String) e.nextElement());
        } catch (java.net.MalformedURLException ex) {
            message("error",1,"Server.readClassPath: URL "+u[i]+" hat falsches Format.");
        }
        i++;        
    }
    return u;  
  };

  public URL[]     getEvolveURLs() {
      return evolveURLs;
  }


  public static void main(String[] args) {
    String serverName = null;  
    if (System.getSecurityManager() == null) {
         System.setSecurityManager(new RMISecurityManager());
    }
    try {
      Server myServer = new Server();
      if (args.length != 1) {
        System.err.println("Aufruf: java Server <Servername>, take " + myServer.host);
      } else myServer.host = args[0];

      Naming.rebind("rmi://" + myServer.host + "/EvolveServer", myServer);
      System.out.println("EvolveServer gestartet");
      myServer.showMsg();                       
      Hashtable h = myServer.getProperties("EvolveStart.ini");
      myServer.startFirstDCATs();
      // myServer.addToGroupForComp("hinken", "muddies");
      // myServer.removeFromGroupForAppl("hinken", "ToDo");
      // myServer.addToGroupForAppl("hinken", "ToDo");
      // myServer.changeApplication("hinken", "ListClientRalph", "ListClientRalph2");
      // for (Enumeration e = myServer.getUsersOfComp("klingel").elements();e.hasMoreElements() ; ) {
      //        System.out.println("klingel = " + (String) e.nextElement());
      // };
      //myServer.showUsers();
      //IR_Component c = myServer.getNewComponent("Test.Textbaustein");
      //System.out.println(c.toText());
      //IR_Component c2 = myServer.getNewComponent("ListClient");
      //System.out.println(c2.toText());


    } catch(Exception e) {
      System.err.println(e);
    }
  }

} ///~:
