// The "freEvolve System" is a platform supporting the reconfiguration of 
// component-based applications'
// Copyright (C) 1999  Ralph Hinken and Dr. Oliver Stiemerling
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Version: 1.0
// Date:    09.10.2001
//
// History:
//

package org.freevolve.server;
import java.util.*;
import java.io.*;

class User {
  public String name ;
  Hashtable appl;
  Hashtable comp;
  String DefFileName;
  String ROOTDIR;
  boolean group;
  Server server;

  public User (String name, String RootDir, String DefFileName, boolean group, Server server) {
    this.server = server;
    this.name = name;
    appl = new Hashtable();
    comp = new Hashtable();
    this.DefFileName =  DefFileName;
    this.ROOTDIR = RootDir;
    getFile();
 };

 public boolean addGroupForComp(String groupName) {
     if (comp.put(groupName, "is_group.cat") != null) {
        server.message("error", 1, "addGroupForComp: Fr den Benutzer "+ this.name + " ist die Gruppe "+groupName+" bereits zugeordnet.");
        return false;
     }
     return save();
 }

 public boolean removeGroupForComp(String groupName) {
    comp.remove(groupName); 
    return save();
 }

 public boolean addGroupForApl(String groupName) {
     if (appl.put(groupName, "is_group.dcat") != null) {
        server.message("error", 1, "User.addGroupForApl: Fr den Benutzer "+ this.name + " ist die Gruppe "+groupName+" bereits zugeordnet.");
        return false;
     }
     return save();
 
 };

 public boolean removeGroupForApl(String groupName) {
    appl.remove(groupName);
    return save();
 }

 public boolean addApplication(String name, String typ) {
     if (appl.contains(name)) {
          server.message("error", 1, "User.addApplication: Fr den Benutzer "+ this.name + " ist die Anwendung "+name+" bereits eingetragen. Verwenden Sie die Funktion changeApplication.");
          return false;
     }
     if (appl.put(typ+".dcat", name) != null) {
          server.message("error", 1, "User.addApplication: Fr den Benutzer "+ this.name + " ist die Anwendung vom Typ "+typ+" bereits eingetragen. Verwenden Sie die Funktion changeApplication.");
        return false;
     }
     return save();
 }

 public boolean removeApplication(String typ) {
    appl.remove(typ + ".dcat");
    return save();
 }

 public boolean changeApplication(String typ, String newTyp) {
     if (appl.containsKey(newTyp+".dcat")) {
          server.message("error", 1, "User.changeApplication: Fr den Benutzer "+ this.name + " ist die Anwendung vom Typ "+newTyp+" bereits eingetragen. Der Typname "+typ + " kann daher nicht gendert werden.");
          return false;
     }
   String name = (String) appl.remove(typ+".dcat");
   /*
   if (name==null) {
       server.message("error", 1, "User.changeApplication: Fr den Benutzer "+ this.name + " ist die Anwendung vom Typ "+typ+" nicht eingetragen und konnte daher nicht entfernt werden.");
       return false;
   }
   */ // Obiges wird zugunsten der berschreibbarkeit von Gruppen-Zuordnungen gestrichen
   if (name==null) {
        Hashtable a = getApplications();
        for (Enumeration e = a.keys(); e.hasMoreElements(); ) {
            String n = (String) e.nextElement();
            String t = (String) a.get(n);
            if (t.equals(typ)) name = n;
        }
   }
   if (name==null) {
       server.message("error", 1, "User.changeApplication: Fr den Benutzer "+ this.name + " ist die Anwendung vom Typ "+typ+" nicht eingetragen und konnte daher nicht entfernt werden.");
       return false;
   }
   appl.put(newTyp+".dcat", name);   
   return save();
 }

 /** Hinweis: Hier muss der neue Typ im Falle einer CAT-Datei schon gespeichert sein.
 */
 public boolean addComponent(String name, String typ) {
     typ += "."+ cat_or_comp(typ);
     if (comp.contains(name)) {
          server.message("error", 1, "User.addComponent: Fr den Benutzer "+ this.name + " ist die Komponente "+name+" bereits eingetragen. ");
          return false;
     }
     if (comp.put(typ, name) != null) {
          server.message("error", 1, "User.addComponent: Fr den Benutzer "+ this.name + " war die Komponente vom Typ "+typ+" bereits eingetragen. Der Eintrag wurde nun berschrieben.");
     }
     return save();
 }

 public boolean removeComponent(String typ)  {
    comp.remove(typ + ".cat");
    return save();
 }

 /** Hinweis: Hier muss der neue Typ im Falle einer CAT-Datei schon gespeichert sein.
 */
 public boolean changeComponent(String typ, String newTyp) {
   String suffix = cat_or_comp(newTyp);
   String o_suffix = cat_or_comp(typ);
   if (comp.containsKey(newTyp)) {
          server.message("error", 1, "changeComponent: Fr den Benutzer "+ this.name + " ist die Komponente vom Typ "+newTyp+" bereits eingetragen. Der Typname "+typ + " kann daher nicht gendert werden.");
          return false;
     }
   newTyp += "."+  suffix;
   String name = (String) comp.remove(typ+o_suffix );
   comp.put(newTyp, name);   
   if (comp.put(typ, name) != null) {
          server.message("error", 1, "User.changeComponent: Fr den Benutzer "+ this.name + " war die Komponente vom Typ "+typ+" bereits eingetragen. Der Eintrag wurde nun berschrieben.");
   }
   return save();
 }

 public boolean save() {
   try {
      MyFile file = new MyFile();
      file.writeFileWithBak(ROOTDIR, DefFileName, "bak", this.toTxt()); 
      return true;
   } catch (Exception e) { 
        System.out.println("User.save: write to file: "+ 
           e.getClass().getName() + ": " + e.getMessage());
        return false;
   }
 }

 public String toTxt() {
  String result = "";
  for (Enumeration e = comp.keys(); e.hasMoreElements(); ) {
      String key = (String) e.nextElement();
      result += key + "=" + (String) comp.get(key) + "\n";
  }
  for (Enumeration e = appl.keys(); e.hasMoreElements(); ) {
      String key = (String) e.nextElement();
      result += key + "=" + (String) appl.get(key) + "\n";
  }
  return result;
}

/** Tests, if component typ is defined in a cat-file.
    @return cat, if there is a file typ.cat
            class, else 
*/
public String cat_or_comp (String typ) {
  //File f = new File(ROOTDIR, typ+".cat");
  //if (f.exists()) return "cat";
  return "cat";
}

private Hashtable getFile() {
    File f;
    FileInputStream fis = null;
    Hashtable properties = new Hashtable();
    try {
      f = new File(ROOTDIR, DefFileName);
      fis = new FileInputStream(f);
      PropertyResourceBundle props = new PropertyResourceBundle(fis);
      Enumeration allProps = props.getKeys();
      String value = null;
      for(String entry=null ; allProps.hasMoreElements() ; ) {
          entry =  (String) allProps.nextElement();
          value = props.getString(entry);
//          System.out.println("Die User-Datei " + DefFileName + " enthlt Eintrag " +
//                                    entry + "=" + value);
          if (entry.endsWith(".dcat")) {
              if (appl.put(entry, value) != null) {
                  System.out.println("Die User-Datei " + DefFileName + " enthlt doppelten Eintrag " +
                                     entry + "=" + value);
              }
          }
          else if (entry.endsWith(".cat")) {
              if (comp.put(entry, value) != null) {
                  System.out.println("Die User-Datei " + DefFileName + " enthlt doppelten Eintrag " +
                                    entry + "=" + value);
              }
          } 
          else if (value.endsWith("is_group.dcat")) {
              if (appl.put(entry, value) != null) {
                  System.out.println("Die User-Datei " + DefFileName + " enthlt doppelten Eintrag " +
                                     entry + "=" + value);
              }
          }
          else if (value.endsWith("is_group.cat")) {
              if (comp.put(entry, value) != null) {
                  System.out.println("Die User-Datei " + DefFileName + " enthlt doppelten Eintrag " +
                                    entry + "=" + value);
              }
          }          
          else {
              System.out.println("Die User-Datei " + DefFileName + " enthlt unbekannten Eintrag " +
                                    entry + "=" + value);
          }
              
      }
      return properties;
    } catch (IOException e) {
      System.err.println("Server.getProperties: IOFehler: " + e);
      return null;
    }
    finally { try { if (fis != null) fis.close(); } catch (IOException e) {}  }
  }

  Hashtable getApplications() {
     Hashtable result = new Hashtable();
     for (Enumeration e = appl.keys(); e.hasMoreElements() ; ) {
          String typ   = (String) e.nextElement();
          String name  = (String) appl.get(typ);
         if (typ.endsWith(".dcat")) {
              String t = new String(typ.substring(0, typ.lastIndexOf(".dcat")));
              result.put(name, t); 
          } else if (name.equals("is_group.dcat")) {
              User u = server.getUser(typ);
              if (u!=null) {
                  result.putAll(u.getApplications());
              }
         }
    }
    return result;
  }

  Hashtable getComponents() {
     Hashtable result = new Hashtable();
     for (Enumeration e = comp.keys(); e.hasMoreElements() ; ) {
          String typ   = (String) e.nextElement();
          String name  = (String) comp.get(typ);
         if (typ.endsWith(".cat")) {
              String t = new String(typ.substring(0, typ.lastIndexOf(".cat")));
              result.put(name, t); 
         } else if (typ.endsWith(".class")) {
              String t = new String(typ.substring(0, typ.lastIndexOf(".class")));
              result.put(name, t); 
         } else if (name.equals("is_group.cat")) {
              User u = server.getUser(typ);
              if (u!=null) {
                  result.putAll(u.getComponents());
              }
         }
    }
    return result;
  }

  Vector whichGroups(String suffix) {
    Vector v = new Vector();
    for (Enumeration e = comp.keys(); e.hasMoreElements() ; ) {
          String typ   = (String) e.nextElement();
          String name  = (String) comp.get(typ);
          if (name.startsWith("is_group."+suffix)) {
              v.add(typ);
          }
    }
    return v;
  }

  boolean hasComp(String typ) {
      Hashtable h = getComponents();
      return h.contains(typ);
  }

  boolean hasAppl(String typ) {
      Hashtable h = getApplications();
      return h.contains(typ);

  }

} /// : ~