// The "freEvolve System" is a platform supporting the reconfiguration of 
// component-based applications'
// Copyright (C) 1999  Ralph Hinken and Dr. Oliver Stiemerling
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Version: 1.0
// Date:    09.10.2001
//
// History:
//

package org.freevolve.server;

import java.util.*;
import java.io.*;
import java.rmi.*;

// requires Component.java 

public class comp_io {
  String compString;
  Component component;
  Vector usedComp;
  Hashtable comps;
  Server   server;
  boolean  found_it;
  
  public comp_io(Server s) {
    super();
    usedComp = new Vector();
    comps    = new Hashtable();
    server   = s;
    found_it = false;
  }


  private  int myNextToken(StreamTokenizer st) {
    try {
      int res = st.nextToken();
      //server.message("comp", 5, st);
      return res;
    } catch (IOException e) {
      server.message("error", 1, "comp_io.myNextToken: " + e.toString());
    }
    return 0;
  }
  
 /**  Liest CAT-String und baut daraus eine Hierarchie vom Component Objekten mit
      Port, Param, PortBinding und ParamBinding Objekten.
      Wenn der letzte Parameter false ist, mu als letzter Eintrag im CAT-String 
      die s_component definiert sein. Diese wird schlielich durch c reprsentiert.<br>
      Wenn der letzte Parameter true ist, reprsentiert c das Objekt, das durch den
      letzten Eintrag in CAT definiert wurde.<br>
      Komponenten, die im CAT-String definiert sind, aber schlielich durch c nicht
      ber subcomponent - Wege erreichbar sind, gehen verloren.
      @param c Das Objekt, das die neue s_component beinhaltet.
     @param copy true, der CAT-String keine s_componente enthlt. 
  */
  public Component getComponent(String CAT, String compName, Component c_return) {
    if (CAT == null || compName==null) return null;
    Component result = null;
    boolean s_comp = false;
    StringReader sr = new StringReader(CAT);
    StreamTokenizer st = new StreamTokenizer(sr);
    st.ordinaryChar('.');
    st.quoteChar('"');
    st.wordChars('_','_');
    int tokenType = myNextToken(st);
    while( tokenType != StreamTokenizer.TT_EOF) {
      if (tokenType != StreamTokenizer.TT_WORD) { scan_error("X_component. Scanne gerade "+ compName, st.lineno());  return null; }
      if (st.sval.equals("i_component")) {
	        if ((result=scan_i_component(st, compName, c_return)) == null) return null;
          if (found_it) {
            return result;
          }
      } else if (st.sval.equals("a_component")) {
          Component comp = null;
       	  if ((result=scan_a_component(st, compName, c_return))  == null) return null;
          if (found_it) {
            return result;
          }
      } else if (st.sval.equals("s_component")) {
         	if (s_comp != false) server.message("comp", 5, "Komponente "+compName+" hat mehr als eine s_component... ;-(");
         	result = scan_a_component(st, compName, c_return); // sollte wie eine a_componente behandelt werden
          if (result == null) return null;
          else if (found_it) {
                  return result;
          } else s_comp = true;
      }
      tokenType = myNextToken(st);
    } // end while not eof;
    server.message("comp", 5, "CAT fertig gelesen.");
    server.message("error", 1, "CAT File enthlt keine Komponente "+compName +".");
    return null;
  }



  private  void scan_error(String e, int line) {
    server.message("error", 5, "Comp_io: Zeile " + line + ": Erwarte " + e);
  }




  private Component scan_a_component(StreamTokenizer st, String compName, Component c) {
    int tokenType;
    String typName = scan_PackageName(st);
    // first token should be the comp_type_name;
    if (typName == null) { scan_error("Typnamen der Komponenten "+compName, st.lineno());  return null; }
    // Falls nicht die erwartete Top-Level Komponente, erzeuge ein neues Component Objekt
    if (!typName.equals(compName)) {
      try { 
         c = new Component(typName, server); 
      } catch (Exception e) { 
            server.message("error", 1, "comp_io.getComponent: Kann Component Objekt "+compName+" nicht erzeugen." + e.getClass().getName() + ": " + e.getMessage()); 
            return null;
      }
    } else found_it = true;

    c.changeTypeName(typName); // comp_type_name;
    if (comps.containsKey(c.getTypeName())) { scan_error("anderen Typ-Namen. Typ-Name "+c.getTypeName()+" wurde in der Komponente "+compName+" wiederholt." , st.lineno()); return null;}
    comps.put(c.getTypeName(), c);
    c.setAtomic(false); // komplexe Komponente
    tokenType = myNextToken(st);  // Opening bracket ;
    if ((char) tokenType != '{' ) { scan_error("{ zur Definition der Komponente " +c.getTypeName()+" in Komponente " + compName, st.lineno());  return null; }
    if (scanParamAndPort(st, c) == -1) return null;   // i_comp_body ;
    if (scanSubcomponents(st, c) == -1) return null;
    // server.message("comp", 5, "eine subcomponente gescannt");
    if (scanBindings(st, c) == -1) return null;
    tokenType = myNextToken(st);  // Closing bracket ;
    if ((char) tokenType != '}' ) { scan_error("} nach Definition der Komponente "+c.getTypeName()+" in Komponente " + compName, st.lineno());  return null; }
    tokenType = myNextToken(st);
    if ((char) tokenType != ';') { scan_error("; nach Definition der Komponente "+c.getTypeName()+" in Komponente " + compName, st.lineno());  return null; }
    // server.message("comp", 5, "Eine abstrakte Komponente gescannt.");
    return c;
  }
  
  private String scan_PackageName(StreamTokenizer st) {
      String s = "";
      int tokenType  = myNextToken(st);
      if (tokenType != StreamTokenizer.TT_WORD) return null;
      s = st.sval;
      while ( (char) (tokenType = myNextToken(st)) == '.') {
         tokenType  = myNextToken(st);
         s += "." + st.sval;
      }
      // I think there are no more ports or parameters ...;
     	st.pushBack();
     	return s;
   }
      

  private   Component scan_i_component(StreamTokenizer st, String compName, Component c) {
    String typ = scan_PackageName(st);
    int tokenType;
    // first token should be the comp_type_name;
    if (typ == null) { scan_error("Komponententypname innerhalb Komponente " + compName , st.lineno());  return null; }
    // Falls nicht die erwartete Top-Level Komponente, erzeuge ein neues Component Objekt
    if (!typ.equals(compName)) {
          try { 
             c = new Component(typ, server); 
          } catch (Exception e) { 
                server.message("error", 1, "comp_io.getComponent: Kann Component Objekt nicht erzeugen." + e.getClass().getName() + ": " + e.getMessage()); 
                return null;
          }
    } else found_it = true;
    c.changeTypeName(typ); // comp_type_name;
    c.setAtomic(true); // atomare Komponente
    if (comps.containsKey(c.getTypeName())) { scan_error("anderer Typname erwartet. Typname "+ c.getTypeName()+" wurde wiederholt (i_comp) innerhalb Komponente "+compName, st.lineno()); return null;}
    comps.put(c.getTypeName(), c);
    tokenType = myNextToken(st);  // Opening bracket ; 
    if ((char) tokenType != '{' ) { scan_error("{ fuer Anfang der i_component Definition der Komponente "+c.getTypeName()+ " innerhalb Komponente "+compName, st.lineno());  return null; }
    if (scanParamAndPort(st, c) == -1) return null;
    tokenType = myNextToken(st); // Closing bracket ;
    if ((char) tokenType != '}') { scan_error("} fr Ende der i_component Definition der Komponente "+c.getTypeName()+ " innerhalb Komponente "+compName, st.lineno());  return null; }
    tokenType = myNextToken(st);
    if ((char) tokenType != ';') { scan_error("Semikolon hinter der i_component Definition der Komponente "+c.getTypeName()+ " innerhalb Komponente "+compName, st.lineno());  return null; }
    // server.message("comp", 5, "Eine implementierte Komponente gescannt.");
    return c;
  }
  



  // scans i_component_body (consists of parameters and ports);
  private int scanParamAndPort(StreamTokenizer st, Component c) {
    int tokenType;
    // jetzt folgen Ports und Parameter in -- beliebigem -- wechsel;
    while ((tokenType = myNextToken(st)) !=  StreamTokenizer.TT_EOF) {
      if (tokenType != StreamTokenizer.TT_WORD) {
	      // I think there are no more ports or parameters ...;
      	st.pushBack();
      	return 0;
      }
      
      if ( (st.sval.equals("provided")) || (st.sval.equals("symmetric")) || (st.sval.equals("required")) ) {
      	String polarity = new String(st.sval);
      	tokenType = myNextToken(st);
      	if (tokenType != StreamTokenizer.TT_WORD) { scan_error("Portname in Komponente "+c.getTypeName(), st.lineno());  return -1; }
      	String name = new String(st.sval);
//      	tokenType = myNextToken(st);
      	if (tokenType != StreamTokenizer.TT_WORD) { scan_error("Porttype in Komponente "+c.getTypeName()+ " fr Port "+name, st.lineno());  return -1; }
      	String type = new String(scan_PackageName(st));
      	tokenType = myNextToken(st);
         String protocoll = "undefined";
         if ((char) tokenType != ';') { // knnte eine Angabe ohne Protokoll sein.
              protocoll = new String(st.sval);
              tokenType = myNextToken(st);
         }
      	if ((char) tokenType != ';') { scan_error("; nach Portdeklaration in Komponente "+c.getTypeName()+ " fr Port "+name, st.lineno());  return -1; }
      	c.addPort(name, type, polarity, protocoll);
     } else if (st.sval.equals("config_parameter")) {
      	tokenType = myNextToken(st);
      	if (tokenType != StreamTokenizer.TT_WORD) { scan_error("Parametername in Komponente "+c.getTypeName(), st.lineno());  return -1; }
      	String name = new String(st.sval);
      	tokenType = myNextToken(st);
      	if (tokenType != StreamTokenizer.TT_WORD) { scan_error("Parametertyp in Komponente "+c.getTypeName(), st.lineno());  return -1; }
      	String type = new String(st.sval);
      	tokenType = myNextToken(st);
      	c.addParam(name, type, null);
      	if ((char) tokenType == ':') {
      	  String svalue = null;
      	  double nvalue = 0.0;
      	  char   cvalue = '?';
      	  type = null;
      	  tokenType = myNextToken(st);
      	  if ((char) tokenType != '=') { scan_error("=", st.lineno());  return -1; }
      	  // Literal: integer, floatingPoint, boolean, char, string, NULL;
          st.quoteChar('"');
      	  tokenType = myNextToken(st);
      	  switch (tokenType) {
      	  case StreamTokenizer.TT_WORD:
      	    if (st.sval.equals("true")) {type = "boolean"; svalue = "true";}
      	    else if (st.sval.equals("false")) { type = "boolean"; svalue = "false";}
      	    else if (st.sval.equals("null")) { type = "null"; svalue = null; }
      	    else { 
                svalue = new String(st.sval); 
                type = "String";
//                System.out.println("P: " + svalue);
            	  tokenType = myNextToken(st);
                while (tokenType != ';') {
                  svalue += " " + new String(st.sval);
              	  tokenType = myNextToken(st);                
                }
                st.pushBack(); // ; zurck
            }
      	    break;
      	  case StreamTokenizer.TT_NUMBER:
      	    type = "double";
      	    nvalue = st.nval;  // NUR REFERENZ KOPIERT !!!!
      	    break;
      	  default:
//            System.out.println("aehm " + st.toString() + " und " + st.sval);
            if (tokenType == '"') {
                type = "string";
                svalue = new String (st.sval);
                
            } else {
          	    type = "char";
          	    cvalue = (char) tokenType;
            }
      	  }

      	  c.setParam(name, null, svalue, nvalue, cvalue);
      	  tokenType = myNextToken(st);
      	}
      	if ((char) tokenType != ';') { scan_error("; nach Parameterdeklaration in Komponente "+c.getTypeName()+" von Parameter "+name, st.lineno());  return -1; }
      } // end else if config_param;
      else {
	      // the next keyword is found that is no param or port ...;
      	st.pushBack();
      	return 0;
      }
    }
    return 0; // wird nie erreicht !;
  }

  private Component getComp(String typeName) {
    if (comps.containsKey(typeName)) {
      Component c = (Component) comps.get(typeName);
      if (usedComp.contains(typeName)) return (Component) c.copy(typeName);
      else {
      	usedComp.addElement(typeName);
      	return c;
      }
    } else {
      return null;
    }
  }

  
  private int scanSubcomponents(StreamTokenizer st, Component c) {
    int tokenType= myNextToken(st);
    if (tokenType != StreamTokenizer.TT_WORD) { st.pushBack(); return 0; } // there are no subcomponents ;
    while ( (tokenType == StreamTokenizer.TT_WORD) && (tokenType != StreamTokenizer.TT_EOF) ) {
      if (!st.sval.equals("subcomponent")) break;
      //      server.message("comp", 5, "START SCAN einer subcomponente");
      tokenType = myNextToken(st);
      if (tokenType != StreamTokenizer.TT_WORD) { scan_error("locale Name der Komponente "+c.getTypeName(), st.lineno()); return -1; }
      String localName = new String(st.sval);
      String typName = new String(scan_PackageName(st));
      if (typName == null) { scan_error("Typname der Komponente", st.lineno()); return -1; }
      Component subComp = getComp(typName);
      if (subComp == null) { scan_error("schon gescannte Subkomponente. "+typName + " nicht gefunden.", st.lineno()); return -1; }
      c.addSubComponent(subComp, localName);
      tokenType = myNextToken(st);
      // Parameter-settings ; ********************************************************************;
      // Test, ob parameter ueberhaupt existiert ... wird beim interpretieren getan;
      if ((char) tokenType == '{') {
      	tokenType = myNextToken(st);
      	while (((char) tokenType != '}') && (tokenType != StreamTokenizer.TT_EOF) ){
      	  if (tokenType != StreamTokenizer.TT_WORD) { scan_error("Parametername der Komponente "+c.getTypeName(), st.lineno()); return -1; }
      	  //Param p = new Param();
      	  String name = new String(st.sval);
      	  String instanceName = new String(localName);
      	  tokenType = myNextToken(st);
      	  if ((char) tokenType != ':') {scan_error(":= fuer Parameterdefinition "+ name +"der Komponente "+c.getTypeName(), st.lineno()); return -1; }
      	  tokenType = myNextToken(st);
      	  if ((char) tokenType != '=') { scan_error(":= fuer Parameterdefinition "+ name +"der Komponente "+c.getTypeName(), st.lineno());  return -1; }
      	  // Literal: integer, floatingPoint, boolean, char, string, NULL;
      	  String svalue = null;
      	  double nvalue = 0.0;
      	  char   cvalue = '?';
      	  String type = null;
      	  tokenType = myNextToken(st);
      	  switch (tokenType) {
      	  case StreamTokenizer.TT_WORD:
        	    if (st.sval.equals("true")) {type = "boolean"; svalue = "true";}
	          else if (st.sval.equals("false")) { type = "boolean"; svalue = "false";}
      	    else if (st.sval.equals("null")) { type = "null"; svalue = null; }
      	    else { 
              svalue = new String(st.sval); 
              type = "String";
          	  tokenType = myNextToken(st);
              while (tokenType != ';') {
                svalue += " " + new String(st.sval);
            	  tokenType = myNextToken(st);                
              }
              st.pushBack(); // ; zurck
            }
      	    break;
      	  case StreamTokenizer.TT_NUMBER:
      	    type = "double";
      	    nvalue = st.nval;  // NUR REFERENZ KOPIERT !!!!;
	          break;
      	  default:
            if (tokenType == '"') {
                type = "string";
                svalue = new String (st.sval);
                
            } else {
          	    type = "char";
          	    cvalue = (char) tokenType;
            }      	  }
           Param psub = subComp.getParam(name, null);	
           if (psub==null)  {
               server.message("error", 1 ,"comp_io " +st.lineno() + ": Parameter " + 
                                             name + " fr Komponente " + c.getTypeName()+ " existiert nicht.");
               return -1;
           }
           if (psub.isValueSet()) {
               scan_error("Parameter " + name + " schon in Komponenete " + instanceName + 
                    " selbst gesetzt. Kein Subkomponenten-Parametersetting", st.lineno()); 
               return -1;
           }
           type = psub.getType();
      	  tokenType = myNextToken(st);
      	  if ((char) tokenType != ';') { scan_error("; nach Parameterdefinition " + name, st.lineno());  return -1; }
           c.addParam(name, type, instanceName);
           c.setParam(name, instanceName, svalue, nvalue, cvalue);
           //c.addParamBinding(instanceName+"#"+name, instanceName, name);

           tokenType = myNextToken(st); // this should be next parameter or closing bracket ;
   	} // end while next parameter setting ;
	if ((char) tokenType != '}') { scan_error("} nach Subkomponentendefinition innerhalb der Komponente " +c.getTypeName(), st.lineno());  return -1; }
	tokenType = myNextToken(st);
      } // end if parameter setting ;
      if ((char) tokenType != ';') { scan_error("; nach Subkomponentendefinition innerhalb der Komponente " +c.getTypeName(), st.lineno());  return -1; }
      // server.message("comp", 5, "eine subcomponente gescannt");
      tokenType = myNextToken(st); // next 'subcomponent' keyword;
    } // end while is subcomponent ;
    st.pushBack(); // when it isn't a next 'subcomponent' keyword';
    return 0; // there are no (more) subcomponents;
  } // End of scanSubcomponents;



  protected int  scanBindings(StreamTokenizer st, Component c) {
    int tokenType;
    boolean paramBinding = false;
    tokenType = myNextToken(st);
    if (tokenType != StreamTokenizer.TT_WORD) { st.pushBack(); return 0; } // there are no Bindings ...;
    while ( (tokenType == StreamTokenizer.TT_WORD)  && (tokenType != StreamTokenizer.TT_EOF) ) {
      paramBinding = false;
      if (!st.sval.equals("bind")) break;
      //server.message("comp", 5, "bind start");
      tokenType = myNextToken(st);
      if (tokenType != StreamTokenizer.TT_WORD) { scan_error("Portname or local Name bei Portverbindung innerhalb Komponente "+c.getTypeName(), st.lineno()); return -1; }
      String compOne = null;
      String nameOne = new String(st.sval); // this is probably the port_name of this component c;
      tokenType = myNextToken(st);
      if ((char) tokenType == '.') {
	      // this is a PortBinding with local_name !;
      	tokenType = myNextToken(st);
      	if (tokenType != StreamTokenizer.TT_WORD) { scan_error("Portname bei Portverbindung innerhalb Komponente "+c.getTypeName(), st.lineno()); return -1; }
        	compOne = nameOne;
      	nameOne =  new String(st.sval);
      	tokenType = myNextToken(st);
      } else if (c.getPort(nameOne) == null) {
   	   // HIER REIN DIE PARAMBINDING !!!!!;
      	paramBinding = true;
         server.message("comp", 6, "Das ist Param Binding");
   	   // then it is a parameter_name;
      } 
      // Trennstrich zwischen den Ports;
      if ((char) tokenType != '-' ) { scan_error("-", st.lineno()); return -1; }
      // der zweite Port ;
      tokenType = myNextToken(st);
      if (tokenType != StreamTokenizer.TT_WORD) { scan_error("Portname or local Name bei Portverbindung innerhalb Komponente "+c.getTypeName(), st.lineno()); return -1; }
      String nameTwo = new String(st.sval);
      String compTwo = null;
      tokenType = myNextToken(st);
      if ((char) tokenType == '.') {
	      // this is a PortBinding with local_name !;
      	tokenType = myNextToken(st);
      	if (tokenType != StreamTokenizer.TT_WORD) { scan_error("Portname bei Portverbindung innerhalb Komponente "+c.getTypeName(), st.lineno()); return -1; }
      	compTwo = nameTwo;
      	nameTwo = new String(st.sval);
      	tokenType = myNextToken(st);
      } 
      if ((char) tokenType != ';' ) { scan_error("; nach bind - Zeile bei Portverbindung innerhalb Komponente "+c.getTypeName(), st.lineno()); return -1; }
      if (paramBinding) { // TEST OB WIRKLICH PARAM ODER PORTS DA SIND...
      	c.addParamBinding(nameOne, compTwo,  nameTwo);
      } else {
      	c.addPortBinding(compOne, nameOne, compTwo, nameTwo);  // WICHTIG, die beteiligten Komponenten und Ports muessen schon da sein! ;
      }
      tokenType = myNextToken(st); // next bind ?;
    } // end while there are bindings ;
    st.pushBack(); // that was no more binding ;
    //    server.message("comp", 5, "Ende add Bindings "); 
    return 0;
  }
      

  /** Gibt CAT-Darstellung der Komponente zurck. 
      Die Komponente, die bergeben wird (i_component, a_component oder s_component)
      kommt am Ende des CAT-Strings.
  */
  public String     getCompString(Component comp) {
    if (comp == null) return null;
    usedComp = new Vector();
    String i_comp = getImpComp(comp);
    String a_comp = getAbsComp(comp);
    return i_comp + a_comp;
  }


  /** Gibt CAT-Darstellung der Komponente ohne die der Subkomponenten zurck. 
  */
  public String     getSingleCompString(Component comp) {
    if (comp == null) return null;
    return  "component " + comp.getTypeName() + " {\n" +
            i_compToString(comp) + a_compToString(comp) + "};\n\n";
  }    


  private   String getImpComp(IR_Component comp) {
    String result = "";
    IR_Component nextComp = null;
    try {
        if (comp.getSubComponents().elements().hasMoreElements() == false) {
          if (!usedComp.contains(comp.getTypeName())) {
            result =  "i_component " + comp.getTypeName() + " {\n" + i_compToString(comp) + "};\n\n";
            usedComp.addElement(comp.getTypeName());
          }
        } else {
            // a_comp;
            for(Enumeration e = comp.getSubComponents().elements(); e.hasMoreElements(); ) {
                nextComp = (IR_Component) e.nextElement();
                result += getImpComp(nextComp);
            }
        }
        return result;
    } catch (RemoteException e) {
       server.message("error", 1, "RMI-ERROR: comp_io.getImpComp: " + e.toString());
       return "RMI-Fehler aufgetreten.";
    }
  }


 private String getAbsComp(IR_Component comp) {
    String result = "";
    try {
        if (!usedComp.contains(comp.getTypeName())) {
          if (!comp.isImplComp()) {
              result = "a_component " + comp.getTypeName() + " {\n" +
                       i_compToString(comp) + a_compToString(comp) + "};\n\n";
          }
          usedComp.addElement(comp.getTypeName());
        } // end if schon geschrieben...
        IR_Component nextComp = null;
        // a_comp;
        for(Enumeration e = comp.getSubComponents().elements(); e.hasMoreElements(); ) {
          nextComp = (IR_Component) e.nextElement();
          result = getAbsComp(nextComp) + result;
        }
        return result;
    } catch (RemoteException e) {
       server.message("error", 1, "RMI-ERROR: comp_io.getAbsComp: " + e.toString());
       return "RMI-Fehler aufgetreten.";
    }
  }    



  private   String i_compToString(IR_Component comp) {
    // ports;
    String req = "";
    String prov = "";
    String symm = "";
    IR_Port p = null;
    String s = "";
    try {
        for(Enumeration e = comp.getPorts().elements() ; e.hasMoreElements() ; ) {
          p  = (IR_Port) e.nextElement();
          if (p.getPolarity().equalsIgnoreCase("provided")) {
            prov += "   provided " + p.getName() + " " + p.getType() + " " + p.getProtocol()+ ";\n";
          } else if (p.getPolarity().equalsIgnoreCase("required")) {
            req += "   required "  + p.getName() + " " + p.getType() +" " + p.getProtocol()+ ";\n";
          } else {
            symm += "   symmetric "  + p.getName() + " " + p.getType() + " " + p.getProtocol()+ ";\n";
          }
        }
        s += req + prov + symm;
        // params;
        IR_Param pa=null;
        for(Enumeration e = comp.getParams().elements() ; e.hasMoreElements() ;) {
          pa  = (IR_Param) e.nextElement();
          if (pa.isSubCompParam()) continue;
          s += "   config_parameter " + pa.getName() + " " + pa.getType();
          if (pa.isValueSet()) {
          if (pa.getType().equals("boolean") || pa.getType().equalsIgnoreCase("String") || pa.getType().equals("null")) {
            s += " := " + pa.getSValue();
          } else if (pa.getType().equals("double")) {
            s += " := " + (double) pa.getNValue();
          } else if (pa.getType().equals("long")) {
            s += " := " + (long) pa.getNValue();
          } else if (pa.getType().equals("short")) {
            s += " := " + (short) pa.getNValue();
          } else if (pa.getType().equals("byte")) {
            s += " := " + (byte) pa.getNValue();
          } else if (pa.getType().equals("int")) {
            s += " := " + (int) pa.getNValue();
          } else if (pa.getType().equals("integer")) {
            s += " := " + (int) pa.getNValue();
          } else if (pa.getType().equals("float")) {
            s += " := " + (float) pa.getNValue();
          } else if (pa.getType().equals("char")) {
            s += " := " + pa.getCValue();
          }
          } // end if isValueSet
          s += ";\n";
        }
        return s;
    } catch (RemoteException e) {
       server.message("error", 1, "RMI-ERROR: comp_io.i_compToString: " + e.toString());
       return "RMI-Fehler aufgetreten.";
    }
  }


  
  private   String a_compToString(IR_Component comp) {
    try {
        String s = "";
        // Subcomponents;
        String subCompName = null;
        // server.message("comp", 5, "anzahl subcomp " + comp.getNumberOfSubComponents());
        for(Enumeration e = comp.getSubComponents().keys() ; e.hasMoreElements() ; ) {
          subCompName  = (String) e.nextElement();
          s += "    subcomponent " + subCompName + " " + ((IR_Component) comp.getSubComponents().get(subCompName)).getTypeName();
          ///**************
          IR_Param pa = null;
          String param_s = ""; // param_settings;
          String l = "         ";
          boolean ps = false;
          for(Enumeration ep = comp.getParams().elements() ; ep.hasMoreElements() ; ) {
            pa  = (IR_Param) ep.nextElement();
            if ((pa.isSubCompParam()) && (pa.getSubCompName().equals(subCompName))) {
                ps = true;
              if (pa.getType().equals("boolean") || pa.getType().equalsIgnoreCase("String") 
                   || pa.getType().equals("null")) {
                param_s += l+pa.getName() + " := " + pa.getSValue() + ";\n";
              } else if (pa.getType().equals("double")) {
                param_s += l+pa.getName() + " := " + (double) pa.getNValue() + ";\n";
             } else if (pa.getType().equals("long")) {
              param_s += l+pa.getName() +" := " + (long) pa.getNValue() + ";\n";
             } else if (pa.getType().equals("short")) {
              param_s += l+pa.getName() +" := " + (short) pa.getNValue() + ";\n";
             } else if (pa.getType().equals("byte")) {
               param_s += l+pa.getName() +" := " + (byte) pa.getNValue() + ";\n";
             } else if (pa.getType().equals("int")) {
              param_s += l+pa.getName() +" := " + (int) pa.getNValue() + ";\n";
             } else if (pa.getType().equals("integer")) {
              param_s += l+pa.getName() +" := " + (int) pa.getNValue() + ";\n";
             } else if (pa.getType().equals("float")) {
               param_s += l+pa.getName() +" := " + (float) pa.getNValue() + ";\n";

             } else if (pa.getType().equals("char")) {
                param_s += l+pa.getName() + " := " + pa.getCValue() + ";\n";
               }
           }
          } // end for enumeration
          if (ps) {
            s += "     {\n" + param_s + "    };\n";
          } else {
            s += ";\n";
          }
          //************************
        //s += ";\n";    // statt paramsetting kram;
        }
        
        // Bindings;
        // component_bindings;
        IR_PortBinding pa = null;
        for(Enumeration e = comp.getPortBindings().elements() ; e.hasMoreElements() ; ) {
          pa  = (IR_PortBinding) e.nextElement();
          String paText = pa.toTxt();
          if (paText != null)  s += "   bind " + paText + ";\n";
        }
        // Parameter_bindings;
        IR_ParamBinding pb = null;
        for(Enumeration e = comp.getParamBindings().elements() ; e.hasMoreElements() ; ) {
          pb = (IR_ParamBinding) e.nextElement();
          String pbText = pb.toTxt();
          if (pbText != null)  s += "   bind " + pbText + ";\n";
        }
        return s;
    } catch (RemoteException e) {
       server.message("error", 1, "RMI-ERROR: comp_io.a_compToString: " + e.toString());
       return "RMI-Fehler aufgetreten.";
    }

  }
  
 
}





